(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL.thy -
 * Copyright (C) 2001,2002,2003,2004 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    

 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)
header {* The \OCL Calculi *}

theory OCL_Calculi 
imports 
  "$HOLOCL_HOME/src/OCL_Library"
begin
text_raw{* \label{thy:OCL_Calculi} *}
(*<*)
ML{* OCL_strict_syntax_checking:=true; *}
(*>*)
text {* 
  This theory develops some more experimental calculi for \OCL such as
  the local judgement tableaux calculus (\acs{lct}) or a particular
  congruence rewriting calculus.  It is the basis of some more
  experimental proof procedures so far. 
*}

(**************************************************************************)
section {* Some core tableaux calculus *}
(**************************************************************************)
text {*
  The conceptual basis of this calculus is drawn
  from~\cite{kerber.ea:mechanization:1994}  where ``labeled formulas''
  where introduced. These are:
   \begin{itemize}
    \item $A^\mathsf{t}$ (for $A\ap s = \isasymMathOclTrue\ap s$),
    \item $A^\mathsf{f}$ (for $A\ap s = \isasymMathOclFalse\ap s$)
    \item $A^\mathsf{u}$ (for $A\ap s = \isasymMathOclUndefined\ap s$)
   \end{itemize}
   A ``labeled clause'' is then a set of labelled formulas
   (considered as disjunctions).
   In their notation, the crucial rules for labeled clauses are:
   \begin{gather}
     \begin{prooftree}
       C, (A \wedge B)^\mathsf{t}
       \justifies
       C, A^\mathsf{t} \quad B^\mathsf{t}
     \end{prooftree}
     \qquad
     \begin{prooftree}
       C, (A \wedge B)^\mathsf{f}
       \justifies
       C, A^\mathsf{f} \quad B^\mathsf{f}
     \end{prooftree}
     \begin{prooftree}
       C, (A \wedge B)^\mathsf{u}
       \justifies
       C, A^\mathsf{u} \quad B^\mathsf{u} \quad C, A^\mathsf{u}, B^\mathsf{u}
     \end{prooftree}
     \prooftreenewline
   \begin{prooftree}
    C, (\lnot A)^\mathsf{t}
    \justifies
    C, A^\mathsf{f}
   \end{prooftree}
   \qquad
   \begin{prooftree}
    C, (\lnot A)^\mathsf{f}
    \justifies
    C, A^\mathsf{t}
   \end{prooftree}
   \qquad
   \begin{prooftree}
    C, (\lnot A)^\mathsf{u}
    \justifies
    C, A^\mathsf{u}
   \end{prooftree}
   \end{gather}
   Tableau-calculi were implemented in Isabelle not on the basis
   of clauses, but on Horn formula
   $\isasymlbrakk A_1; \ldots;  A_n \isasymrbrakk  \isasymLongrightarrow  B$ 
   corresponding to
   $\lnot A_1, \ldots A_n, B$
   in clause representation. The conclusion in the Horn-format is used
   with preference during proof search. 

   As a consequence, it is necessary to duplicate the set of rules into
   introduction- and elimination rules. Further, we use the validity normal
   form to represent negation; thus the handling of the not is different
   and has to be encoded into the rules (leading to further duplications
   of the rule sets), this is necessary in Isabelle's representation 
   of tableaux calculi anyway.
*}

lemma and_TRUE_I: 
  "\<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> B  \<rbrakk> \<Longrightarrow>  \<tau> \<MathOclValid> (A \<MathOclAnd> B)"
  by (erule subst_LJ_TRUE, erule subst_LJ_TRUE, auto)
  (* safe rule *)

lemma and_D1: 
  "\<lbrakk> \<tau> \<MathOclValid> A \<MathOclAnd> B  \<rbrakk> \<Longrightarrow>  \<tau> \<MathOclValid> A"
  apply (simp_all add: OclAnd_def OclLocalValid_def 
                       DEF_def lift2_def split add: split_if)
  apply (case_tac "A \<tau> = down")
  apply (simp_all add: OclTrue_def lift0_def)
  apply (simp add: split_if_eq1)
  apply (case_tac "B \<tau> = down")
  apply (simp_all add: split_if_eq1 not_down_exists_lift)
  apply (erule exE) 
  apply (erule exE)
  apply (simp_all add: drop_lift)
  done
    (* obviously unsafe destruction rule  *)


lemma and_D2: 
  "\<lbrakk> \<tau> \<MathOclValid> A \<MathOclAnd> B  \<rbrakk> \<Longrightarrow>  \<tau> \<MathOclValid> B"
  by (rule and_D1, subst and_commute, assumption)  
    (* obviously unsafe destruction rule  *)


lemma and_TRUE_E: 
  "\<lbrakk> \<tau> \<MathOclValid> (A \<MathOclAnd> B); \<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> B \<rbrakk> \<Longrightarrow> R \<rbrakk> \<Longrightarrow> R"
  by(frule and_D1, drule and_D2, auto)
    (* safe rule *)

(* completeness: MOVE in own section ! *)
lemma and_congr': 
  "(\<tau> \<MathOclValid> (A \<MathOclAnd> B)) = ((\<tau> \<MathOclValid> A) \<and> (\<tau> \<MathOclValid> B))"
by(auto elim: and_TRUE_E and_D1 and_D2 intro: and_TRUE_I)

lemma not_and_undef_congr1[simp]: 
  "(\<tau> \<MathOclValid> \<MathOclNot> (\<MathOclUndefined> \<MathOclAnd> B)) = (\<tau> \<MathOclValid> \<MathOclNot> B)" 
  apply (cut_tac \<tau> = "\<tau>" and A = "B" in non_quatrium_datur)
  apply auto
  apply (drule subst_LJ_TRUE_fw, assumption, auto) 
  apply (erule_tac subst_LJ_FALSE, auto) 
  by (drule_tac subst_LJ_undef_fw,assumption, auto)

lemma or_undef_congr1[simp]: 
  "(\<tau> \<MathOclValid> \<MathOclUndefined> \<MathOclOr> B) = (\<tau> \<MathOclValid> B)" 
  by (simp add: OclOr_def)

lemma not_and_undef_congr2[simp]: 
  "(\<tau> \<MathOclValid> \<MathOclNot> (B \<MathOclAnd> \<MathOclUndefined>)) = (\<tau> \<MathOclValid> \<MathOclNot> B)"
  by(subst and_commute,simp add: not_and_undef_congr1)

lemma or_undef_congr2[simp]: 
  "(\<tau> \<MathOclValid> B \<MathOclOr> \<MathOclUndefined>) = (\<tau> \<MathOclValid> B)" 
  by (simp add: OclOr_def)

lemma and_undef_congr3[simp]: 
  "~(\<tau> \<MathOclValid> \<MathOclUndefined> \<MathOclAnd> B)"
  apply (cut_tac \<tau> = "\<tau>" and A = "B" in non_quatrium_datur)
  apply auto
  apply (drule_tac P = "\<lambda>B. (\<MathOclUndefined> \<MathOclAnd> B)"  in subst_LJ_TRUE_fw_rev)
  apply simp_all
  by (drule_tac subst_LJ_undef_fw,assumption, auto)

lemma and_undef_congr4[simp]: 
  "~(\<tau> \<MathOclValid> B \<MathOclAnd> \<MathOclUndefined>)"
  by(subst and_commute,simp add: not_and_undef_congr1)

lemma and_FALSE_I_core:
  "\<lbrakk> (\<tau> \<MathOclValid> \<MathOclNot> A) \<or> (\<tau> \<MathOclValid> \<MathOclNot> B) \<rbrakk> \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclNot>(A \<MathOclAnd> B)"
  by (auto elim: subst_LJ_FALSE)

lemmas and_FALSE_I = disjCI [THEN and_FALSE_I_core, standard]
  (* saves one resolution step;  safe rule ... *)

lemma and_FALSE_I1: 
  "\<lbrakk> (\<tau> \<MathOclValid> B) \<or> (\<tau> \<MathOclValid> (\<MathOclIsUndefined> B)) \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclNot> A) \<rbrakk> \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclNot>(A \<MathOclAnd> B)"
  by (rule and_FALSE_I, simp only: not_invalid) 

(*<*)
ML {* fun HINT s tac n = EVERY[  subgoals_tac [s] n, tac (n+1)] *}
(*>*)
lemma and_FALSE_E: 
  "\<lbrakk> \<tau> \<MathOclValid>  \<MathOclNot>(A \<MathOclAnd> B); \<tau> \<MathOclValid>  \<MathOclNot> A \<Longrightarrow>  R; \<tau> \<MathOclValid>  \<MathOclNot> B \<Longrightarrow>  R \<rbrakk>  \<Longrightarrow>  R"
  apply (cut_tac \<tau> = "\<tau>" and A = "A" in non_quatrium_datur)
  apply auto
  apply (rotate_tac [1] -1)
  apply (rotate_tac [2] -1)
  apply (drule_tac [2] subst_LJ_undef_fw) 
  apply (drule subst_LJ_TRUE_fw)
  apply assumption 
  prefer 3
  apply assumption 
  apply auto
  done
    (* safe rule *)

(* The following rule has its basis in the SKL theorem:
 
   isDefined_andD0 :  "isdef(X \<MathOclAnd> Y) = (   \<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y
                                         \<MathOclOr> \<MathOclIsDefined> X \<MathOclAnd> \<MathOclNot> X
                                         \<MathOclOr> \<MathOclIsDefined> Y \<MathOclAnd> \<MathOclNot> Y)"
   As equivalence, it can be used as safe introduction rule when used
   from right to left.
*)
lemma or_2_classic_disj: 
   "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> X; \<tau> \<MathOclValid> \<MathOclIsDefined> Y \<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> X \<MathOclOr> Y) = ((\<tau> \<MathOclValid> X) \<or> (\<tau> \<MathOclValid> Y))"
   apply auto 
   defer 1  apply (erule_tac P = "\<lambda> X St. (X \<MathOclOr> Y) St" in subst_LJ_TRUE) 
   prefer 3 apply (erule_tac P = "\<lambda> Y St. (X \<MathOclOr> Y) St" in subst_LJ_TRUE)
   apply (auto simp: not_valid)
   apply (rotate_tac -1)
   apply (drule_tac  P = "\<lambda> Y St. (X \<MathOclOr> Y) St" in subst_LJ_FALSE_fw) 
   apply auto
   done

lemma isdef_and_congr:
   "(\<tau> \<MathOclValid> \<MathOclIsDefined> (X \<MathOclAnd> Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y) \<or> 
                             (\<tau> \<MathOclValid> \<MathOclIsDefined> X \<MathOclAnd> \<MathOclNot> X)   \<or> 
                             (\<tau> \<MathOclValid> \<MathOclIsDefined> Y \<MathOclAnd> \<MathOclNot> Y))" 
   apply (auto simp: isDefined_andD0 or_2_classic_disj or_2_classic_disj) 
   apply (simp_all add: or_2_classic_disj)
   done

lemma solution0: 
  " \<lbrakk> ((\<tau> \<MathOclValid> \<MathOclIsDefined> X)  \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y)) \<or> 
      ((\<tau> \<MathOclValid> \<MathOclIsDefined> X)  \<and> (\<tau> \<MathOclValid> \<MathOclNot> X)) \<or> 
      ((\<tau> \<MathOclValid> \<MathOclIsDefined> Y)  \<and> (\<tau> \<MathOclValid> \<MathOclNot> Y)) \<rbrakk>      
  \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsDefined>(X \<MathOclAnd> Y)"
  by (simp add: isdef_and_congr and_congr')
  (* solution 0 brings it back modestly to HOL *)


lemma solution1: 
  " \<lbrakk> \<lbrakk>(\<MathOclIsDefined> B \<MathOclAnd> \<MathOclNot> B)\<tau> ~= \<MathOclTrue> \<tau>;(\<MathOclIsDefined> A \<MathOclAnd> \<MathOclNot> A)\<tau> ~= \<MathOclTrue> \<tau> \<rbrakk> 
     \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> A) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> B) \<rbrakk>      
  \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsDefined>(A \<MathOclAnd> B)"
  apply (simp add: OclLocalValid_def[symmetric])
  apply (rule solution0)
  apply (cases "\<not> \<tau> \<MathOclValid> \<MathOclIsDefined> B  \<MathOclAnd> \<MathOclNot> B", simp_all)
  apply (cases "\<not> \<tau> \<MathOclValid> \<MathOclIsDefined> A  \<MathOclAnd> \<MathOclNot> A", simp_all)
  apply(auto elim!: and_TRUE_E)
  done
  (* solution 1 brings it closer to Pure, but is not valid_norm. *)

lemma solution2: 
  "\<lbrakk>  \<lbrakk>  \<tau> \<MathOclValid> \<MathOclNot>(\<MathOclIsDefined> B \<MathOclAnd> \<MathOclNot> B); \<tau> \<MathOclValid>  \<MathOclNot>(\<MathOclIsDefined> A \<MathOclAnd> \<MathOclNot> A) \<rbrakk>          
     \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> A) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> B) \<rbrakk>      
  \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsDefined>(A \<MathOclAnd> B)"
  apply (rule solution1)
  apply (simp only: OclLocalValid_def[symmetric] not_valid) 
  apply auto
  done
  (* solution 2 is similar solution 1, but valid_norm. *)

lemma isDefined_and_TRUE_I: 
  "\<lbrakk> \<lbrakk> (\<tau> \<MathOclValid> \<MathOclIsUndefined> B) \<or> (\<tau> \<MathOclValid> B);(\<tau> \<MathOclValid> \<MathOclIsUndefined> A) \<or> (\<tau> \<MathOclValid> A) \<rbrakk>               
     \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> A) \<and>(\<tau> \<MathOclValid> \<MathOclIsDefined> B) \<rbrakk>      
   \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsDefined>(A \<MathOclAnd> B)"
  apply (rule solution2)
  apply (auto elim!: and_FALSE_E and_TRUE_E simp add: isDefined_andD0 or_def)
  done
  (* solution 2 is similar solution 1, but valid_norm. *)

lemma solution3: 
  "\<lbrakk>  \<lbrakk>~(\<tau> \<MathOclValid> \<MathOclNot> A); ~(\<tau> \<MathOclValid> \<MathOclNot> B) \<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> A);  
     \<lbrakk>~(\<tau> \<MathOclValid> \<MathOclNot> A); ~(\<tau> \<MathOclValid> \<MathOclNot> B) \<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> B) \<rbrakk> 
   \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsDefined>(A \<MathOclAnd> B)"
  apply (rule isDefined_and_TRUE_I)
  apply (simp only: not_invalid) 
  apply auto
  done
  (* solution 3 is similar solution 2, emphasises compactnes over 
     valid_normality. *)

lemma isDefined_and_TRUE_E: 
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(A \<MathOclAnd> B);                             
     \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> A; \<tau> \<MathOclValid> \<MathOclIsDefined> B\<rbrakk> \<Longrightarrow> R;      
     \<tau> \<MathOclValid> \<MathOclNot> A \<Longrightarrow> R;                                 
     \<tau> \<MathOclValid> \<MathOclNot> B \<Longrightarrow> R                                  
   \<rbrakk> \<Longrightarrow> R"
  by (auto elim!: and_FALSE_E and_TRUE_E 
           simp add: isDefined_andD0 OclOr_def)


text{* The following rule follows the principal equality:
\begin{equation}
 \isasymtau  \isasymMathOclValid  \isasymMathOclIsUndefined (A \isasymMathOclAnd  B)
 =
 \isasymtau \isasymMathOclValid 
((\isasymMathOclIsUndefined  A \isasymMathOclOr  \isasymMathOclIsUndefined  B) \isasymMathOclAnd  
 (\isasymMathOclIsUndefined  A \isasymMathOclOr  A) 
\isasymMathOclAnd  (\isasymMathOclIsUndefined  B \isasymMathOclOr  B))
\end{equation}
Via @{text "isDefined_andD0"}, @{text "de_morgan1"}, and 
  @{text "de_morgan2"} one gets:
\begin{equation}   
   = 
     \isasymtau \isasymMathOclValid (\isasymMathOclNot  
(\isasymMathOclIsUndefined  A \isasymMathOclAnd  \isasymMathOclIsUndefined  B) \isasymMathOclAnd  
            \isasymMathOclNot  
(\isasymMathOclIsUndefined  A \isasymMathOclAnd  \isasymMathOclNot  A) \isasymMathOclAnd  
    \isasymMathOclNot  (\isasymMathOclIsUndefined  B \isasymMathOclAnd  \isasymMathOclNot  B))
\end{equation}
*}
lemma isDefined_and_FALSE_I: 
  "\<lbrakk>  \<tau> \<MathOclValid> \<MathOclNot> (\<MathOclIsDefined> A \<MathOclAnd> \<MathOclIsDefined> B);     
      \<tau> \<MathOclValid> \<MathOclNot> (\<MathOclIsDefined> A \<MathOclAnd> \<MathOclNot> A);        
      \<tau> \<MathOclValid> \<MathOclNot> (\<MathOclIsDefined> B \<MathOclAnd> \<MathOclNot> B)     \<rbrakk>  
  \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsUndefined>(A \<MathOclAnd> B)"
  by  (auto intro!: and_TRUE_I and_FALSE_I elim!: and_FALSE_E and_TRUE_E 
            simp add: isDefined_andD0 OclOr_def)
    (* Looks like badly optimized, but at least safe.
    Curiously, experiments show that it behaves
    better than subsequent rules: and_assoc needs with isDefined_and_FALSE_I 0,7 sek,
    in the other versiones 
    isDefined_and_FALSE_I1 22 sek,
    isDefined_and_FALSE_I2 19 sek,
    isDefined_and_FALSE_I3 17,sek !!!   

    Interaction with basic HOL tableau not understood.
    *) 

lemma isDefined_and_FALSE_I1: 
  "\<lbrakk>   ((\<tau>  \<MathOclValid> \<MathOclIsUndefined> A) \<and> ~(\<tau> \<MathOclValid> \<MathOclNot> B)) \<or>     
      (~(\<tau> \<MathOclValid> \<MathOclNot> A) \<and>(\<tau> \<MathOclValid> \<MathOclIsUndefined> B))    \<rbrakk>  
  \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsUndefined>(A \<MathOclAnd> B)"
  apply (rule isDefined_and_FALSE_I)
  apply (simp_all add: not_valid)
  apply (auto elim!: subst_LJ_undef subst_LJ_TRUE subst_LJ_FALSE)
  done

lemma isDefined_and_FALSE_I2: 
 "\<lbrakk>   ((\<tau> \<MathOclValid> \<MathOclNot> A) \<or> \<not>(\<tau> \<MathOclValid> \<MathOclIsUndefined> B)) \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsUndefined> A;  
      ((\<tau> \<MathOclValid> \<MathOclNot> A) \<or> \<not>(\<tau> \<MathOclValid> \<MathOclIsUndefined> B)) \<Longrightarrow> \<not>(\<tau> \<MathOclValid> \<MathOclNot> B) \<rbrakk>  
  \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsUndefined>(A \<MathOclAnd> B)"
    apply (rule isDefined_and_FALSE_I1)
    apply(auto simp del: isDefined_if_valid isDefined_if_invalid)
    done


   
lemma isDefined_and_FALSE_I3: 
  assumes A: "((\<tau> \<MathOclValid> \<MathOclIsUndefined> B) \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclNot> A)) \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsUndefined> A"
  and B:     "((\<tau> \<MathOclValid> \<MathOclIsUndefined> B) \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclNot> A)) \<Longrightarrow> ~(\<tau> \<MathOclValid> \<MathOclNot> B)"
  shows "\<tau> \<MathOclValid> \<MathOclIsUndefined>(A \<MathOclAnd> B)"
    apply (rule isDefined_and_FALSE_I2)
    apply (rule A)
    apply (rule_tac [2] B)
    apply auto
    done
    (* push everything into Pure. Gives a unusual type of rule.
       Don't know if Isabelles provers can handle this. *)

lemma isDefined_and_FALSE_I4: 
"\<lbrakk>   \<tau> \<MathOclValid> (\<MathOclIsDefined> B) \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsUndefined> A);  
    \<tau> \<MathOclValid> (\<MathOclIsDefined> A) \<Longrightarrow> (\<tau> \<MathOclValid> A);           
    \<tau> \<MathOclValid> (\<MathOclIsDefined> B) \<Longrightarrow> (\<tau> \<MathOclValid> B) \<rbrakk>  
 \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsUndefined>(A \<MathOclAnd> B)"
    apply (rule isDefined_and_FALSE_I)
    apply (rule and_FALSE_I)
    prefer 2
    apply (subst and_commute )
    apply (rule_tac and_FALSE_I)
    prefer 3
    apply (subst and_commute)
    apply (rule_tac and_FALSE_I)
    apply(auto simp del: isDefined_if_valid isDefined_if_invalid)
    done

lemma isDefined_and_FALSE_I5: 
  "\<lbrakk>  (\<tau> \<MathOclValid> \<MathOclIsUndefined> A) \<or> (\<tau> \<MathOclValid> \<MathOclIsUndefined> B);     
      (\<tau> \<MathOclValid> \<MathOclIsUndefined> A) \<or> (\<tau> \<MathOclValid> A);        
      (\<tau> \<MathOclValid> \<MathOclIsUndefined> B) \<or> (\<tau> \<MathOclValid> B)  \<rbrakk>  
  \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsUndefined>(A \<MathOclAnd> B)"
 apply(simp add: isDefined_andD0 de_morgan1 de_morgan2
                 and_congr' or_2_classic_disj)
 apply(auto elim!: subst_LJ_undef subst_LJ_TRUE subst_LJ_FALSE
            simp del: isDefined_if_valid 
                      not_isUndefined_eq_isDefined)
 done

lemma isDefined_and_FALSE_E: 
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsUndefined>(A \<MathOclAnd> B);                              
     \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsUndefined> A; \<tau> \<MathOclValid> \<MathOclIsUndefined> B\<rbrakk> \<Longrightarrow> R;        
     \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsUndefined> A; \<tau> \<MathOclValid> B \<rbrakk>        \<Longrightarrow> R;        
     \<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> \<MathOclIsUndefined> B \<rbrakk>        \<Longrightarrow> R \<rbrakk>    
   \<Longrightarrow>  R"
  by (auto elim!: and_TRUE_E and_FALSE_E simp add: isDefined_andD0 OclOr_def)

(* Basic Simplification: We ommit the isDefined_and category and
introduce directly rules for isDefined. This leads to the following
six cases: *)

  lemmas isDefined_TRUE_E = non_tertium_datur_if_isDefined

  lemma isDefined_TRUE_I:  
    "\<lbrakk> \<tau> \<MathOclValid> A \<or> (\<tau> \<MathOclValid> \<MathOclNot> A) \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> A"
    apply (erule disjE)
    apply (drule isDefined_if_valid)
    apply (drule_tac [2] isDefined_if_invalid)
    apply auto
    done

lemma isDefined_TRUE_I1: 
" \<lbrakk> (\<tau> \<MathOclValid> A) \<or> (\<tau> \<MathOclValid> (\<MathOclIsUndefined>(A))) \<Longrightarrow> \<tau> \<MathOclValid> A \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> (\<MathOclIsDefined> A)"
  apply (rule isDefined_TRUE_I)
  apply (auto simp del: isDefined_if_valid isDefined_if_invalid)
  apply (simp only: not_valid not_not isDefined_notD0) 
  done

  lemma not_not_E: 
    "\<lbrakk> \<tau> \<MathOclValid> \<MathOclNot>(\<MathOclNot> A); \<tau> \<MathOclValid> A \<Longrightarrow> R\<rbrakk> \<Longrightarrow> R"
    by auto

  lemma not_not_D: 
    "\<tau> \<MathOclValid> \<MathOclNot>(\<MathOclNot> A) \<Longrightarrow> \<tau> \<MathOclValid> A"
    by auto

  lemma not_not_I: 
    "\<lbrakk> \<tau> \<MathOclValid> A \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclNot>(\<MathOclNot> A)"
    by auto

  lemma isDefined_not_TRUE_E: 
    "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclNot> A)\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> A"
    by auto

  lemma isDefined_not_TRUE_I: 
    "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> A \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclNot> A)"
    by auto

  lemma isDefined_not_FALSE_E: 
    "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsUndefined>(\<MathOclNot> A)\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsUndefined>(A)"
    by auto

  lemma isDefined_not_FALSE_I: 
    "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsUndefined> A \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsUndefined>(\<MathOclNot> A)"
    by auto

(* ******************************** *)
section {* Drived rules for the or  *}
(* ******************************** *)
  lemma or_TRUE_I_core:
    "\<lbrakk> (\<tau> \<MathOclValid> A) \<or> (\<tau> \<MathOclValid> B) \<rbrakk> \<Longrightarrow>  \<tau> \<MathOclValid> A \<MathOclOr> B"
    apply (unfold OclOr_def)
    apply (rule and_FALSE_I_core)
    apply auto
    done

  lemmas or_TRUE_I = disjCI [THEN or_TRUE_I_core, standard]
    (* safe rule *)
    (* such a rule leads to an disjunction in the conclusion
    and then a negated label - thus not a purely positive
    calculus. *)

  lemma or_FALSE_I: 
    "\<lbrakk> \<tau> \<MathOclValid> \<MathOclNot> A; \<tau> \<MathOclValid> \<MathOclNot> B \<rbrakk> \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclNot>(A \<MathOclOr> B)"

    apply (unfold OclOr_def)
    apply (rule not_not_I)
    apply (rule and_TRUE_I)
    apply auto
    done

  lemma or_TRUE_E: 
    "\<lbrakk>  \<tau> \<MathOclValid> A \<MathOclOr> B;         
        \<tau> \<MathOclValid> A \<Longrightarrow> R;        
        \<tau> \<MathOclValid> B \<Longrightarrow> R \<rbrakk>      
    \<Longrightarrow>  R"
    apply (unfold OclOr_def)
    apply (auto elim!: and_TRUE_E and_FALSE_E 
                simp add: isDefined_andD0 OclOr_def)
    done

  lemma or_FALSE_E: 
    "\<lbrakk>  \<tau> \<MathOclValid> \<MathOclNot>(A \<MathOclOr> B);    
        \<lbrakk> \<tau> \<MathOclValid> \<MathOclNot> A; \<tau> \<MathOclValid> \<MathOclNot> B \<rbrakk> \<Longrightarrow> R \<rbrakk>      
    \<Longrightarrow>  R"
    apply (unfold OclOr_def)
    apply (auto elim!: and_TRUE_E and_FALSE_E 
                simp add: isDefined_andD0 OclOr_def)
    done

  lemma isDefined_or_TRUE_I: 
    "(\<lbrakk> (\<tau> \<MathOclValid> \<MathOclIsUndefined> B) \<or> (\<tau> \<MathOclValid> \<MathOclNot> B);         
       (\<tau> \<MathOclValid> \<MathOclIsUndefined> A) \<or> (\<tau> \<MathOclValid> \<MathOclNot> A) \<rbrakk>      
      \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> A) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> B) )      
     \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> (A \<MathOclOr> B)"
    apply (unfold OclOr_def)
    apply (rule isDefined_not_TRUE_I)
    apply (rule isDefined_and_TRUE_I)
    apply auto
    done

lemma isDefined_or_FALSE_I1: 
"\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> B \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsUndefined> A;      
   \<tau> \<MathOclValid> \<MathOclIsDefined> A \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclNot> A;           
   \<tau> \<MathOclValid> \<MathOclIsDefined> B \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclNot> B \<rbrakk>         
 \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsUndefined> (A \<MathOclOr> B))"
   apply (unfold OclOr_def)
   apply (rule isDefined_not_FALSE_I)
   apply (rule isDefined_and_FALSE_I4)
   apply auto
   done

 lemma isDefined_or_FALSE_I: 
 "\<lbrakk> (\<tau> \<MathOclValid> A) \<or> ~(\<tau> \<MathOclValid> \<MathOclIsUndefined> B) \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsUndefined> A);  
    (\<tau> \<MathOclValid> A) \<or> ~(\<tau> \<MathOclValid> \<MathOclIsUndefined> B) \<Longrightarrow> ~(\<tau> \<MathOclValid>  B) \<rbrakk>        
  \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsUndefined> (A \<MathOclOr> B))"

    apply (unfold OclOr_def)
    apply (rule isDefined_not_FALSE_I)
    apply (rule isDefined_and_FALSE_I)
    
    apply (rule and_FALSE_I)
    apply (rule isDefined_not_FALSE_I)
    apply (rule_tac [2] and_FALSE_I)
    apply (rule_tac [2] isDefined_not_FALSE_I)
    apply (rule_tac [3] and_FALSE_I)
    apply (rule_tac [3] isDefined_not_FALSE_I)
    
    apply auto  
    apply (simp_all only: not_valid)
    apply (auto simp: isDefined_if_valid)
    done

  lemma isDefined_or_FALSE_E: 
    "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsUndefined>(A \<MathOclOr> B);                               
       \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsUndefined> A; \<tau> \<MathOclValid> \<MathOclIsUndefined> B\<rbrakk> \<Longrightarrow> R;        
       \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsUndefined> A; \<tau> \<MathOclValid> \<MathOclNot> B \<rbrakk>     \<Longrightarrow> R;        
       \<lbrakk> \<tau> \<MathOclValid> \<MathOclNot> A; \<tau> \<MathOclValid> \<MathOclIsUndefined> B \<rbrakk>     \<Longrightarrow> R   \<rbrakk>    
    \<Longrightarrow>  R"
    apply (unfold OclOr_def)
    apply (drule isDefined_not_FALSE_E)
    apply (auto elim!: and_TRUE_E and_FALSE_E isDefined_and_FALSE_E 
                simp add: isDefined_andD0 OclOr_def)
    done

  lemma isDefined_or_TRUE_E: 
    "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(A \<MathOclOr> B);                         
       \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> A; \<tau> \<MathOclValid> \<MathOclIsDefined> B\<rbrakk> \<Longrightarrow> R;    
       \<tau> \<MathOclValid> B    \<Longrightarrow> R; 
       \<tau> \<MathOclValid> A   \<Longrightarrow> R   \<rbrakk>        
    \<Longrightarrow>  R"
    apply (unfold OclOr_def)
    apply (drule isDefined_not_TRUE_E)
    apply (erule isDefined_and_TRUE_E)
    apply auto
    done

(**************************************************************************)
(** Some basic laws for the if then else  operator                       **)
(**************************************************************************)
lemma strongEq_distrib_if1:
"\<MathOclValid> \<MathOclIsDefined> X \<Longrightarrow>  
 (A \<MathOclStrictEq> (\<MathOclIf> X \<MathOclThen> (Y::'a=>('b::bot)) \<MathOclElse> Z \<MathOclEndif>)) =  
 (\<MathOclIf> X \<MathOclThen> A \<MathOclStrictEq> Y \<MathOclElse> A \<MathOclStrictEq> Z \<MathOclEndif>)"
by(rule if_distrib2, auto simp: OclValid_def OclLocalValid_def)

lemma strongEq_distrib_if2:
"\<MathOclValid> \<MathOclIsDefined> X \<Longrightarrow> 
   ((\<MathOclIf> X \<MathOclThen> (Y::'a=>('b::bot)) \<MathOclElse> Z \<MathOclEndif>) \<MathOclStrictEq>  A) = 
    (\<MathOclIf> X \<MathOclThen> Y \<MathOclStrictEq> A \<MathOclElse> Z \<MathOclStrictEq> A \<MathOclEndif>)"
by(rule if_distrib2, auto simp: OclValid_def OclLocalValid_def)


(**************************************************************************)
(** Fundamental Laws for Congruence Rewriting in OCL                     **)
(**************************************************************************)
lemma cp_local_valid [simp,intro!]:
"cp P \<Longrightarrow> cp (\<lambda> X \<tau>. (\<tau> \<MathOclValid> P X))"
by(auto simp: OclValid_def OclLocalValid_def)

lemma and_congr: 
      "\<lbrakk> (\<tau> \<MathOclValid> A) \<or> (\<tau> \<MathOclValid> \<MathOclIsUndefined> A) \<Longrightarrow> (\<tau> \<MathOclValid> B) = (\<tau> \<MathOclValid> B') \<rbrakk> 
       \<Longrightarrow> (\<tau> \<MathOclValid> A \<MathOclAnd> B) = (\<tau> \<MathOclValid> A \<MathOclAnd> B')"
      apply (cut_tac \<tau> = "\<tau>" and A = "A" in non_quatrium_datur)
      apply (rule_tac X = A in cp_distinct_core_P)
      apply (auto simp: OclLocalValid_def[symmetric])
      done

lemma and_congr_a: 
      "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> A; \<tau> \<MathOclValid> A \<Longrightarrow> (\<tau> \<MathOclValid> B) = (\<tau> \<MathOclValid> B') \<rbrakk> 
       \<Longrightarrow>  (\<tau>  \<MathOclValid> A \<MathOclAnd> B) = (\<tau> \<MathOclValid> A \<MathOclAnd> B')"
      apply (rule and_congr)
      apply auto
      done

lemma and_congr_b: 
      "\<lbrakk> \<tau> \<MathOclValid> \<MathOclNot> A; \<tau> \<MathOclValid> \<MathOclIsUndefined> A \<Longrightarrow> (\<tau> \<MathOclValid> B) = (\<tau> \<MathOclValid> B')\<rbrakk>
       \<Longrightarrow>  ( \<tau>  \<MathOclValid> A \<MathOclAnd> B) = (\<tau> \<MathOclValid> A \<MathOclAnd> B')"
      apply (rule and_congr)
      apply auto
      done

lemma not_congr: 
      "\<lbrakk> A \<tau> = A' \<tau>\<rbrakk>  \<Longrightarrow>  (\<tau> \<MathOclValid> \<MathOclNot> A) = (\<tau> \<MathOclValid> \<MathOclNot> A')"
      apply (rule sym)
      apply (erule_tac P = "\<lambda> A \<tau>. (\<tau> \<MathOclValid> \<MathOclNot> A)" in ocl_cp_subst)
      apply auto
      done

text {* 
  The relevance for automated deduction becomes apparent with the next
  rewrite rule, which boils down reductions in local congruences to
  check if the definedness is equivalent. This can be decomposed into
  two meta-implications; the advantage of this is presentation comes
  from the fact that such derivations are particularly easy and can be
  supported automatically by a forward-closure tactic for literals
  contained in valid definedness formulas. It has to be checked if the
  recomputation of these literals can be avoided. 
*}
lemma not_congr0: 
      "\<lbrakk> (\<tau> \<MathOclValid> \<MathOclIsDefined> A) \<longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> A'); 
         (\<tau> \<MathOclValid> \<MathOclIsDefined> A') \<longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> A);
         (\<tau> \<MathOclValid> A) = (\<tau> \<MathOclValid> A')\<rbrakk> 
       \<Longrightarrow>  (\<tau> \<MathOclValid> \<MathOclNot> A) = (\<tau> \<MathOclValid> \<MathOclNot> A')"
      apply (rule not_congr)
      apply (subst equiv_D_local) back
      apply auto
      done

lemma or_congr: 
      "\<lbrakk> A \<tau> = A' \<tau>; B \<tau> = B' \<tau> \<rbrakk> 
       \<Longrightarrow>  (\<tau> \<MathOclValid> A \<MathOclOr> B) = (\<tau> \<MathOclValid> A' \<MathOclOr> B')"
      apply (rule_tac P = "\<lambda> A \<tau>. (\<tau> \<MathOclValid> A \<MathOclOr> B)" in ocl_cp_subst)
      apply (rule sym, assumption)
      apply (rule_tac P = "\<lambda> B \<tau>. (\<tau> \<MathOclValid> A' \<MathOclOr> B)" in ocl_cp_subst)
      apply (rule sym, assumption)
      apply auto
      done

lemma or_congr0: 
      "\<lbrakk> (\<tau> \<MathOclValid> \<MathOclIsDefined> A)  \<longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> A'); 
         (\<tau> \<MathOclValid> \<MathOclIsDefined> A') \<longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> A);
         (\<tau> \<MathOclValid> \<MathOclIsDefined> B)  \<longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> B'); 
         (\<tau> \<MathOclValid> \<MathOclIsDefined> B') \<longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> B);
         A \<tau> = A' \<tau> ; B \<tau> = B' \<tau> \<rbrakk> 
       \<Longrightarrow>  (\<tau> \<MathOclValid> A \<MathOclOr> B) = (\<tau> \<MathOclValid> A' \<MathOclOr> B')"
      apply (rule or_congr)
      apply (subst equiv_D_local) back
      prefer 2
      apply (subst equiv_D_local) back
      apply (auto simp: OclLocalValid_def) 
      apply (auto simp: OclLocalValid_def[symmetric]) (* UGLY !!! *)
      done

lemma if_congr: 
      "\<lbrakk> \<tau> \<MathOclValid> C \<Longrightarrow> (\<tau> \<MathOclValid> A) = (\<tau> \<MathOclValid> A');
         \<tau> \<MathOclValid> \<MathOclNot> C \<Longrightarrow> (\<tau> \<MathOclValid> B) = (\<tau> \<MathOclValid> B')  \<rbrakk>  
      \<Longrightarrow>  ((\<tau> \<MathOclValid> \<MathOclIf> C \<MathOclThen> A \<MathOclElse> B \<MathOclEndif>) = 
           (\<tau> \<MathOclValid> \<MathOclIf> C \<MathOclThen> A'\<MathOclElse> B'\<MathOclEndif>))"
      apply (cut_tac \<tau> = "\<tau>" and A = "C" in non_quatrium_datur)
      apply (rule_tac X = C in cp_distinct_core_P)
      apply (auto simp: is_FALSE_charn_local OclLocalValid_def[symmetric])
      done

lemma implies_undef_congr[simp]: 
      "(\<tau> \<MathOclValid> \<MathOclUndefined> \<MathOclImplies> B) = (\<tau> \<MathOclValid> B)"
      by (simp add:OclImplies_def)

lemma implies_congr: 
      "\<lbrakk> (\<tau> \<MathOclValid> A) \<or> (\<tau> \<MathOclValid> \<MathOclIsUndefined> A) \<Longrightarrow> (\<tau> \<MathOclValid> B) = (\<tau> \<MathOclValid> B') \<rbrakk>  
      \<Longrightarrow> ((\<tau> \<MathOclValid> A \<MathOclVVImplies> B) = (\<tau> \<MathOclValid> A \<MathOclVVImplies> B'))"
      apply (cut_tac \<tau> = "\<tau>" and A = "A" in non_quatrium_datur)
      apply (rule_tac X = A in cp_distinct_core_P) 
      apply (auto simp: isUndefined_charn_local OclLocalValid_def[symmetric])
      done

lemmas LEC = and_congr not_congr or_congr if_congr
             not_and_undef_congr1
             not_and_undef_congr2
             or_undef_congr1
             or_undef_congr2
             and_undef_congr3
             and_undef_congr4
             implies_undef_congr
(*
(* TODO: similar rules for xor *)

thm LEC

(**************************************************************************)
(**  OCL - calculi                                                       **)
(**************************************************************************)


section {* LTC *}

section {* GTC *}

section {* LEC *}
*)

(*<*)
ML{* OCL_strict_syntax_checking:=false; *}
(*>*)
end
 
