(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_logic.thy - The Core Logics of OCL.
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* \OCL Logic Core *}
 
theory OCL_Logic_core 
  imports "$HOLOCL_HOME/src/library/basic/OCL_Boolean"
  uses    "$HOLOCL_HOME/src/kernel_ext/isabelle_kernel_patch.ML" 
begin
 text_raw{* \label{thy:OCL_Logic_core} *}
(*<*)
ML{* OCL_strict_syntax_checking:=true; *}
(*>*)

section{* Introduction *}
text {*
  The purpose of this theory is to introduce the concept of
  \OCL judgement (written $\tau \models \phi$) over a context $\tau$
  and an \OCL formula $\phi$ and to derive a number of mechanizeable
  calculi for it. Here, a \emph{context} is
  usually a system transition in \OCL, i.e. a pair of two states
  $\tau = (\tau_\text{pre},\tau_\text{post})$;
  however, since our theory on \OCL does make perfect sense for
  generalizations such as "sequences of states" of Kripke structures,
  we prefer the more general notion of context to simple state pairs.
  A formula $\phi$ may contain paths such as ``\inlineocl{self.a.b}'' which
  were interpreted in $\tau$. The concept of a judgement occurs
  already in the standard in definition A.33 (Semantics of 
  Operation Specifications). We turn it to the heart of deductive
  calculi for \OCL: It will be the basis for 
  equational calculi (both strict and strong equalities), 
  congruences on formulas and local and global equivalences
  (terminology explained below)
  over judgements, and tableaux calculi over judgements. 

  All of them are integrated into tactic procedures that allow
  for automated proof state transformations even automated
  proofs over \OCL formula. Finally, these \ac{sml}-programmed tactics
  were bound to a specific \acs{isar} syntax.
  *}


subsection{* Terminolgy *} 
text{*  
  An \OCL formula is \emph{globally valid} (or just \emph{valid} if
  its evaluation yields for all contexts (e.g., state transitions
  $(s,s')$, Kripke-structures, \ldots) true; an \OCL formula is
  \emph{locally valid} if this is the case for a specific context.

  We have three different equational calculi for \OCL forming
  the bare bones of automated deduction in \OCL:
  \begin{enumerate}
  \item an \emph{universal congruence} (on \OCL formulae) (\acs{uc}) 
        is a (conditional)
        formula with an equality $E = E'$ as conclusion where
        $E$ and $E'$ has type $V_{\alpha}(\beta)$.
        These equalities hold for all contexts $\tau$ and can be
        processed directly by Isabelle's rewriter.
  \item a \emph{local equivalence} (\acs{le}) is a (conditional)
        formula with an equality $E~\tau = E'~\tau$ as conclusion where
        $E$ and $E'$ has type $V_{\alpha}(\beta)$.
        These equalities hold only in some context $\tau$ of type $\tau$ (?) 
        and can only be processed by the tactic \tactic{OCL\_rewriter} tactic 
        in subterms occurring in judgements and in the $E's$ in 
        $E~\tau = E'~\tau$.
  \item a \emph{local judgement equivalence} (\acs{lje}) which states that
        two judgments over \OCL formulae must be equivalent in a particular
        context $\tau$, so $(\tau \models E) = (\tau \models E')$.
        Recall that judgements require that the formula \emph{holds}, i.e.,
        evalutates to \inlineocl{true}. In  contrast to \acs{le}, the formulae 
        must not agree on all three truth values. 
\end{enumerate}
*}


section {* The Theory of \OCL Judgements *}
text {* In the following, the notation for validity judgements is
        introduced: *}
constdefs 
       OclValid  :: "'a Boolean \<Rightarrow> bool" 
       "OclValid P \<equiv> (P = \<MathOclTrue>)"
       OclLocalValid  :: "['a,'a Boolean] \<Rightarrow> bool"
       "OclLocalValid \<tau> P \<equiv> (P \<tau> = \<MathOclTrue> \<tau>)"

syntax  (* standard syntax *)
   "_OclValid_std"     :: "'a Boolean \<Rightarrow> bool"
                         ("(\<OclValid> (_))" 50) 
   "_OclLocalValid_std":: "['a,'a Boolean] \<Rightarrow> bool" 
                         ("(1(_)/\<OclLocalValid>(_))" 50)

syntax  (* standard ascii syntax *)
   "_OclValid_ascii"     :: "'a Boolean \<Rightarrow> bool"
                         ("|= (_)" 50) 
   "_OclLocalValid_ascii":: "['a,'a Boolean] \<Rightarrow> bool" 
                         ("(1(_)/|=(_))" 50)

syntax  (* standard ascii syntax *)
   "_OclValid_math"     :: "'a Boolean \<Rightarrow> bool"
                           ("\<MathOclValid> (_)" 50) 
   "_OclLocalValid_math":: "['a,'a Boolean] \<Rightarrow> bool" 
                           ("(1(_)/ \<MathOclValid> (_))" 50)

(*<*)
parse_translation {* flat(map (emb trans_const) 
                              ["OclValid","OclLocalValid"]) *}

print_translation {* map emb_print 
                              ["OclValid","OclLocalValid"] *}
(*>*)

subsection {* $\HolOclCp$ \ldots *}

lemma cp_OclLocalValid[simp, intro!] : 
  "cp P \<Longrightarrow> cp(\<lambda> X \<tau>. \<tau> \<MathOclValid> P X)"
  by(simp add: OclLocalValid_def)

text{* The rules needed by @{text "cp_unfold"}. *}

lemma OCL_cp0_OclLocalValid: "cp0 (\<lambda> x \<tau>. \<tau> \<MathOclValid> x)"
  by(simp add: OclLocalValid_def OclTrue_def ss_cp_defs ss_lifting')

lemma OCL_cp0_fold_OclLocalValid: "(\<tau> \<MathOclValid> (lift0 (x \<tau>))) = (\<tau> \<MathOclValid> x)"
  by(simp add: OclLocalValid_def OclTrue_def ss_cp_defs ss_lifting')

ML {*
cpR_simpset := (!cpR_simpset) addsimps [thm "OCL_cp0_OclLocalValid"];
cpR_fold_simpset := (!cpR_fold_simpset) addsimps 
                    [thm "OCL_cp0_fold_OclLocalValid"];
*}


subsection{* Generic Definedness Rules on Local Validity *}

lemma lift1_strictify_implies_LocalValid_defined:
   "f \<equiv> lift\<^sub>1 (strictify (\<lambda>x. \<HolOclLiftLeft>g x\<HolOclLiftRight>)) 
    \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X)) = (\<tau> \<MathOclValid> \<MathOclIsDefined> X)"
  apply(drule lift1_strictify_is_isdef_fw)
  apply(rule_tac t = " \<MathOclIsDefined>(f X)" in ssubst)
  apply simp_all
  done


lemma lift2_strictify_implies_LocalValid_defined:
   "f \<equiv> lift\<^sub>2 (strictify (\<lambda>x. strictify (\<lambda>y. \<HolOclLiftLeft>g x y\<HolOclLiftRight>))) 
    \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and>  (\<tau> \<MathOclValid> \<MathOclIsDefined> Y))"
  apply(drule lift2_strictify_is_isdef_fw)
  apply(rule_tac t = " \<MathOclIsDefined>(f X Y)" in ssubst,assumption)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done


subsection{* Semantic Representations of Validity *}

lemma localValid2sem:
"(\<tau> \<MathOclValid> P) = (P \<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>)"
  by (simp add: OclLocalValid_def OclTrue_def lift0_def)

lemma localValidNot2sem:
"(\<tau> \<MathOclValid> \<MathOclNot> P) = (P \<tau> = \<HolOclLiftLeft>False\<HolOclLiftRight>)"
  by (simp add: OclLocalValid_def OclNot_def OclTrue_def 
             strictify_def lift0_def lift1_def o_def
             not_down_exists_lift, auto)

lemma localValidDefined2sem:
"(\<tau> \<MathOclValid> \<MathOclIsDefined> S) = DEF(S \<tau>)"
  by (simp add: OclLocalValid_def OclIsDefined_def strictify_def
             o_def lift0_def lift1_def lift2_def
             OclTrue_def OclFalse_def  OclUndefined_def DEF_def) 

lemma localValidUnDefined2sem:
"(\<tau> \<MathOclValid> \<MathOclIsUndefined> S) = (S \<tau> = \<bottom>)"
  by (simp add: OclLocalValid_def OclIsDefined_def strictify_def
             o_def lift0_def lift1_def lift2_def OclNot_def
             OclTrue_def OclFalse_def  OclUndefined_def DEF_def) 

lemma localValidToNotFalse2sem:
"(\<tau> \<MathOclValid> P (x::'b::type) \<MathOclOr> \<MathOclIsUndefined>(P x)) = (P x \<tau> \<noteq> \<HolOclLiftLeft>False\<HolOclLiftRight>)"
  by(simp add: OclLocalValid_def OclIsDefined_def strictify_def
             OclNot_def OclOr_def OclAnd_def 
             o_def lift0_def lift1_def lift2_def
             OclTrue_def DEF_lift not_down_exists_lift, auto)

lemma localValidToNotTrue2sem:
"(\<tau> \<MathOclValid> \<MathOclNot> P (x::'b::type) \<MathOclOr> \<MathOclIsUndefined>(P x)) = (P x \<tau> \<noteq> \<HolOclLiftLeft>True\<HolOclLiftRight>)"
  by(simp add: OclLocalValid_def OclIsDefined_def strictify_def
             OclNot_def OclOr_def OclAnd_def
             o_def lift0_def lift1_def lift2_def
             OclTrue_def DEF_lift not_down_exists_lift, auto)

lemma isUndefined_charn: 
"(X = \<MathOclUndefined>) = (\<MathOclValid> \<MathOclIsUndefined> X)"
  apply (auto simp: OclValid_def)
  apply (simp add: OclUndefined_def OclTrue_def 
                   lift0_def lift1_def o_def  DEF_def OclNot_def 
                   OclIsDefined_def)
  apply (rule ext)
  apply (drule_tac x = "s" in fun_cong)
  apply auto
  done

lemma is_FALSE_charn: 
"(X = \<MathOclFalse>) = ( \<MathOclValid> \<MathOclNot> X)"
  apply (auto simp: OclValid_def)
  apply (simp add: OclUndefined_def OclFalse_def OclTrue_def 
                   lift0_def lift1_def o_def OclIsDefined_def 
                   DEF_def OclNot_def strictify_def)
  apply (rule ext)
  apply (drule_tac x = "s" in fun_cong)
  apply (case_tac "X s = down")
  apply auto
  apply (simp add: not_down_exists_lift)
  apply auto
  done

lemma is_TRUE_charn: 
"(X = \<MathOclTrue>) = ( \<MathOclValid> X)"
  by (auto simp: OclValid_def)

text{* 
  Local and global validity statements are in fact complete in
  the sense that they can inherently express all three basic values
  of booleans. This foundational fact gives rise to the idea to use
  these judgements both for tableaux and congruence-rewrite
  calculi. 
  *}

text{* 
  Instead of representations in terms of semantic primitives, the
  local judgements can also be expressed on a more abstract level,
  i.e., in terms of global equivalences.  
  *}

text{* 
  The following property reveals that strong equality is just mirrored
  by the \HOL equality. Note that the strong equalty is implicitly
  use also in the construction of sets---its choice is therefore highly
  critical.
  Depending on the layout of objects, i.e., the concrete
  object universe, this results in different properties. For the case
  of the referential universe and states only containing objects with reference
  to themselves in the state, logical equality coincides with referential
  equalty. 
*}

lemma isUndefined_charn_local: 
"(X \<tau> = \<MathOclUndefined> \<tau>) = (\<tau> \<MathOclValid> \<MathOclIsUndefined> X)"
  by (simp add: OclLocalValid_def
                OclUndefined_def OclTrue_def lift0_def lift1_def 
                o_def OclIsDefined_def DEF_def OclNot_def)

lemma is_FALSE_charn_local: 
"(X \<tau> = \<MathOclFalse> \<tau>) = (\<tau> \<MathOclValid> \<MathOclNot> X)"
  apply (simp add: OclLocalValid_def
                   OclUndefined_def OclTrue_def OclFalse_def 
                   lift0_def lift1_def o_def OclIsDefined_def DEF_def 
                   strictify_def OclNot_def)
  apply auto
  apply (case_tac "X St = down")
  apply (simp_all add: not_down_exists_lift)
  apply auto
  done

lemma is_TRUE_charn_local: 
"(X \<tau> = \<MathOclTrue> \<tau>) = (\<tau> \<MathOclValid> X)"
  by (simp add: OclLocalValid_def) (* just unfold syntax *)


lemma strongEq_charn: "(\<tau> \<MathOclValid> x \<MathOclStrongEq> y) = (x \<tau> = y \<tau>) "
apply (simp add: OclStrongEq_def OclLocalValid_def lift2_def
   OclTrue_def  lift0_def)
done


subsection{*The Standardization Rules of (Global) Validity Statements.*}

lemma OclTrue_is_Valid[simp]: "\<MathOclValid> \<MathOclTrue>"
  by (simp add: OclValid_def)

lemma OclFalse_is_Invalid[simp]: "\<not>(\<MathOclValid> \<MathOclFalse>)"
  apply (simp add: OclValid_def OclTrue_def OclFalse_def lift0_def)
  by    (unfold not_def, rule impI, drule fun_cong, simp)

lemma OclUndefined_is_Invalid[simp]: "\<not> (\<MathOclValid> \<MathOclUndefined>)"
  apply (simp add: OclValid_def OclTrue_def OclUndefined_def lift0_def)
  by    (unfold not_def, rule impI, drule fun_cong, simp)


text{*The Standardization Rules of Local Validity Statements.*}

lemma OclTrue_is_LocalValid[simp]: "\<tau> \<MathOclValid> \<MathOclTrue>"
  by (simp add: OclLocalValid_def)
lemma OclFalse_is_LocalInvalid[simp]: "\<not>(\<tau> \<MathOclValid> \<MathOclFalse>)"
  by (simp add: OclLocalValid_def OclTrue_def OclFalse_def lift0_def)
lemma OclUndefined_is_LocalInvalid[simp]: "\<not> (\<tau> \<MathOclValid> \<MathOclUndefined>)"
  by (simp add: OclLocalValid_def OclTrue_def OclUndefined_def lift0_def)

text{* On the basis of the standarization rules, the simplifier
  can already rule out a number of absurd validity statements: *}
 
lemma isdef_congr[simp]:   "(\<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclNot> A)) = (\<tau> \<MathOclValid> \<MathOclIsDefined>(A))" by simp (* test *)

lemma "(\<tau> \<MathOclValid> \<MathOclIsUndefined>(\<MathOclNot> A)) = (\<tau> \<MathOclValid> \<MathOclIsUndefined>(A))" by simp (* test *)

lemma "\<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclIsDefined> A)"  by simp (* test *)

lemma "\<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclTrue>)"    by simp (* test *)
 
lemma "\<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclFalse>)"    by simp (* test *)

lemma "\<not>(\<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclUndefined>))" by simp (* test *)

lemma "\<not>(\<tau> \<MathOclValid> \<MathOclIsUndefined>(\<MathOclTrue>))" by simp (* test *)

lemma "\<not>(\<tau> \<MathOclValid> \<MathOclIsUndefined>(\<MathOclFalse>))" by simp (* test *)

lemma "\<tau> \<MathOclValid> \<MathOclIsUndefined>(\<MathOclUndefined>)"    by simp (* test *)

lemma "\<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclTrue>)"    by simp (* test *)

subsection{* Another Foundational Fact: @{text "non_quatrium_datur"} *}

lemma non_quatrium_datur: "(\<tau> \<MathOclValid> A) \<or> (\<tau> \<MathOclValid> \<MathOclNot> A) \<or> (\<tau> \<MathOclValid> \<MathOclIsUndefined> A)"
  apply (simp add: isUndefined_charn_local [symmetric])
  apply (simp add: is_FALSE_charn_local    [symmetric])
  apply (cut_tac base_distinct)
  apply (auto simp: OclLocalValid_def)
  done


ML_setup {*
fun prover4 (goal,name) =
    let val thm = prove_goalw (the_context())  
                              [OclFalse_def,OclTrue_def,OclUndefined_def,lift0_def] 
                  goal (fn _ => [auto_tac(claset(),simpset()),
                                 TRY(dtac fun_cong 1),
                                 auto_tac(claset(),simpset())])
    in  bind_thm(name,thm);
        Addsimps[thm]
    end;
*}

ML_setup {*
map prover4
[("\<MathOclFalse> \<noteq> \<MathOclUndefined>", "not__FALSE__undef"), 
 ("\<MathOclUndefined> \<noteq> \<MathOclFalse>", "not__undef__FALSE"),
 ("\<MathOclTrue> \<noteq> \<MathOclUndefined>", "not__TRUE__undef"),
 ("\<MathOclUndefined> \<noteq> \<MathOclTrue>", "not__undef__TRUE"),
 ("\<MathOclFalse> \<noteq> \<MathOclTrue>", "not__FALSE__TRUE"),
 ("\<MathOclTrue> \<noteq> \<MathOclFalse>", "not__TRUE__FALSE"),

 ("\<MathOclFalse> \<tau> \<noteq> \<MathOclUndefined> \<tau>","not__FALSE__undef_st"),
 ("\<MathOclUndefined> \<tau> \<noteq> \<MathOclFalse> \<tau>","not__undef__FALSE_st"),
 ("\<MathOclTrue> \<tau> \<noteq> \<MathOclUndefined> \<tau>","not__TRUE__undef_st"),
 ("\<MathOclUndefined> \<tau> \<noteq> \<MathOclTrue> \<tau>","not__undef__TRUE_st"),
 ("\<MathOclFalse> \<tau> \<noteq> \<MathOclTrue> \<tau>","not__FALSE__TRUE_st"),
 ("\<MathOclTrue> \<tau> \<noteq> \<MathOclFalse> \<tau>","not__TRUE__FALSE_st")];

*}

lemma absurd1: "\<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> \<MathOclNot> A \<rbrakk> \<Longrightarrow> R"
  by (auto simp add: is_FALSE_charn_local [symmetric],
           simp add: OclLocalValid_def)


lemma absurd1S: "\<tau> \<MathOclValid> A \<Longrightarrow> ~(\<tau> \<MathOclValid> \<MathOclNot> A)"
  apply (rule_tac Pa = False in swap,simp_all)
  by    (erule absurd1,simp)


lemma absurd2: "\<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> \<MathOclIsUndefined> A \<rbrakk> \<Longrightarrow> R"
  apply(simp add: isUndefined_charn_local[symmetric])
  apply(simp add: is_TRUE_charn_local[symmetric])
  done

lemma absurd2S: "\<tau> \<MathOclValid> A \<Longrightarrow> ~(\<tau> \<MathOclValid> \<MathOclIsUndefined> A)"
  apply (rule_tac Pa = False in swap,simp_all)
  by    (erule absurd2,simp)


lemma absurd3: "\<lbrakk> \<tau> \<MathOclValid> \<MathOclNot> A; \<tau> \<MathOclValid> \<MathOclIsUndefined> A \<rbrakk> \<Longrightarrow> R"
  apply(simp add: isUndefined_charn_local[symmetric])
  apply(simp add: is_FALSE_charn_local[symmetric])
  done

lemma absurd3S: "\<tau> \<MathOclValid> \<MathOclNot> A \<Longrightarrow> ~(\<tau> \<MathOclValid> \<MathOclIsUndefined> A)"
  apply (rule_tac Pa = False in swap,simp_all)
  by    (erule absurd3,simp)


lemma absurd4: "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsUndefined> (\<MathOclIsDefined> A) \<rbrakk> \<Longrightarrow> R"
  by (simp add: OclLocalValid_def)

lemma absurd4S[simp]: "~ (\<tau> \<MathOclValid> \<MathOclIsUndefined> (\<MathOclIsDefined> A))"
  by (simp add: OclLocalValid_def)

lemma absurd5: "\<lbrakk> \<tau> \<MathOclValid> \<MathOclUndefined> \<rbrakk> \<Longrightarrow> R"
  by (simp add: OclLocalValid_def)

lemma absurd5S[simp]: "~ (\<tau> \<MathOclValid> \<MathOclUndefined>)"
  by (simp add: OclLocalValid_def)

lemma absurd6: "\<lbrakk> \<tau> \<MathOclValid> \<MathOclFalse> \<rbrakk> \<Longrightarrow> R"
  by (simp add: OclLocalValid_def)

lemma absurd6S[simp]: "~ (\<tau> \<MathOclValid> \<MathOclFalse>)"
  by (simp add: OclLocalValid_def)


lemmas absurdE = absurd1 absurd2 absurd3 absurd4 absurd5 absurd6
lemmas absurdS = absurd1S absurd2S absurd3S absurd4S absurd6S absurd6S

(* Only one of these options should be switched on ! *)
declare absurdE [elim]   (* bang or not bang, that is the question \<dots> 
                            But experiments show severe problems in auto
                            with bang. *)
(* declare absurdS [simp] *)


ML_setup {*
val OCL_absurd_tac = best_tac (HOL_cs addEs 
                     [(thm "absurd1"),(thm "absurd2"),(thm "absurd3")])
*}


lemma not_valid:   "(~(\<tau> \<MathOclValid> A)) = ((\<tau> \<MathOclValid> \<MathOclNot> A) \<or> (\<tau> \<MathOclValid> \<MathOclIsUndefined> A))"
  apply (cut_tac \<tau> = "\<tau>" and A = "A" in non_quatrium_datur)
  apply auto
  done


lemma not_invalid: "(~(\<tau> \<MathOclValid> \<MathOclNot> A)) = ((\<tau> \<MathOclValid> A) \<or> (\<tau> \<MathOclValid> \<MathOclIsUndefined> A))"
  by (cut_tac \<tau> = "\<tau>" and A = "A" in non_quatrium_datur, auto)

lemma not_isUndefined: "(~(\<tau> \<MathOclValid> \<MathOclIsUndefined> A)) = ( (\<tau> \<MathOclValid> A) \<or> (\<tau> \<MathOclValid> \<MathOclNot> A))"
  by (cut_tac \<tau> = "\<tau>" and A = "A" in non_quatrium_datur, auto)


lemma isUndefined_eq_not_isDefined[simp]: 
  "(~(\<tau> \<MathOclValid> \<MathOclIsDefined> A)) = (\<tau> \<MathOclValid> \<MathOclIsUndefined> A)"
  apply auto
  apply (drule not_valid [THEN iffD1], auto simp: OclLocalValid_def)
  done


lemma not_isUndefined_eq_isDefined[simp]:
  "(~(\<tau> \<MathOclValid> \<MathOclIsUndefined> A)) = (\<tau> \<MathOclValid> \<MathOclIsDefined> A)"
  apply (subst isUndefined_eq_not_isDefined[symmetric])
  apply (rule HOL.not_not)
  done

lemma not_isUndefined2: "(\<tau> \<MathOclValid> \<MathOclIsDefined> A) = ( (\<tau> \<MathOclValid> A) \<or> (\<tau> \<MathOclValid> \<MathOclNot> A))"
  by (simp add: not_isUndefined[symmetric])


lemmas  norm_validS =
  not_valid              (* is problematic since subsumes all others *)
  not_invalid            (* overlaps with not_isUndefined_eq_isDefined *)
  not_isUndefined        (* overlaps with not_isUndefined_eq_isDefined *)
  isUndefined_eq_not_isDefined (* default *)
  not_isUndefined_eq_isDefined (* default *)



(**************************************************************************)
subsection {*  Global (Universal) vs. Local Validity *}
(**************************************************************************)
(* TODO: the equivalent formulation in local validity ... *)

lemma global_vs_local_validity: "(\<forall> \<tau>. (\<tau> \<MathOclValid> A)) =  (\<MathOclValid> A) "
  by (auto simp:OclLocalValid_def OclValid_def,
      rule ext,erule_tac x = "x" in allE, assumption)

lemma valid_intro: "\<lbrakk> \<And>\<tau>. \<tau>  \<MathOclValid> A \<rbrakk> \<Longrightarrow>  \<MathOclValid> A "
  by (simp only: OclLocalValid_def OclValid_def,rule ext, auto)

lemma valid_elim:  "\<lbrakk> \<MathOclValid> A \<rbrakk> \<Longrightarrow>       \<tau> \<MathOclValid> A"
  by (simp only: OclLocalValid_def OclValid_def)

(* Should these two rules be added to the classical
   sets ? This would mean that in the classical reasoner,
   any global validity is internally used to the local 
   validity.

   Answer: Any naive approach would loop. In order
   to control the "entry" into the "local" calculus,
   we use rules like equiv_I. *)


(**************************************************************************)
subsection {*  Universal Congruence vs. Judgement Equivalence             *}
(**************************************************************************)

lemma equiv_I:
  assumes a1: "\<And> \<tau>. ((\<tau> \<MathOclValid> A) = (\<tau> \<MathOclValid> B))"
  and a2:     "\<And> \<tau>. ((\<tau> \<MathOclValid> \<MathOclIsUndefined> A) = (\<tau> \<MathOclValid> \<MathOclIsUndefined> B))"
  shows       "A = B "
  apply (rule ext)
  apply (cut_tac \<tau> = "x" and A = "A" in non_quatrium_datur)
  apply (erule disjE)
  apply (erule_tac [2] disjE)
  apply (cut_tac \<tau>1 = "x" in a1 [THEN iffD1])
  apply (cut_tac [4] \<tau>1 = "x" in a2 [THEN iffD1])
  apply (auto)
  prefer 2
  apply (simp_all add: isUndefined_charn_local [symmetric]) 
  apply (cut_tac \<tau> = "x" and A = "B" in non_quatrium_datur)
  apply (erule disjE)
  apply (erule_tac [2] disjE)
  prefer 2
  apply (simp only: is_FALSE_charn_local [symmetric])

  apply (drule a1 [THEN iffD2])

  apply (drule_tac [2] a2 [THEN iffD2])
  apply (auto)
  apply (simp_all add: OclLocalValid_def)
  done


text{* \emph{Absolute equivalence} of \OCL formula can be expressed
       in terms of absolute validity as follows: *} 
lemma equiv_D:
  "(A = B) = ((\<forall> \<tau>. ((\<tau> \<MathOclValid> A) = (\<tau> \<MathOclValid> B))) \<and> (\<forall> \<tau>. ((\<tau> \<MathOclValid> \<MathOclIsUndefined> A) = (\<tau> \<MathOclValid> \<MathOclIsUndefined> B))))"
  by (auto intro!:equiv_I)

text{* \emph{Local equivalence} of \OCL formula can be expressed in
       terms of local validity as follows: *}

lemma equiv_D_local:
  "(A \<tau> = B \<tau>) = (((\<tau> \<MathOclValid> A) = (\<tau> \<MathOclValid> B)) \<and> (((\<tau> \<MathOclValid> \<MathOclIsUndefined> A) = (\<tau> \<MathOclValid> \<MathOclIsUndefined> B))))"
  apply (auto)
  apply (simp_all add: isUndefined_charn_local[symmetric])
  apply (simp_all only: not_valid)
  apply auto
  apply (simp_all add: is_FALSE_charn_local[symmetric])
  apply (simp_all add: OclLocalValid_def)
  done

lemma equiv_I': 
  assumes A: "\<And> \<tau>. ((\<tau> \<MathOclValid> A) = (\<tau> \<MathOclValid> B))"
  and B:     "\<And> \<tau>. ((\<tau> \<MathOclValid> \<MathOclNot> A) = (\<tau> \<MathOclValid> \<MathOclNot> B))"
  shows      "A = B"
  apply (rule ext)
  apply (cut_tac \<tau> = "x" and A = "A" in non_quatrium_datur,safe)
  apply (cut_tac \<tau>1 = "x" in A [THEN iffD1])
  apply (cut_tac [3] \<tau>1 = "x" in B [THEN iffD1])
  apply auto
  apply (simp only: OclLocalValid_def)
  apply (simp only: is_FALSE_charn_local [symmetric])
  apply (cut_tac \<tau> = "x" and A = "B" in non_quatrium_datur)
  apply (safe)
  apply (drule A[THEN iffD2])
  apply (drule_tac [2] B[THEN iffD2])
  apply auto
  apply (simp only: isUndefined_charn_local [symmetric])
  done

lemma equiv_D':
  "(A = B) = ((\<forall> \<tau>. ((\<tau> \<MathOclValid> A) = (\<tau> \<MathOclValid> B))) \<and> (\<forall> \<tau>. ((\<tau> \<MathOclValid> \<MathOclNot> A) = (\<tau> \<MathOclValid> \<MathOclNot> B))))"
  by(auto intro!:equiv_I')


lemma equiv_D'_local:
  "(A \<tau> = B \<tau>) = (((\<tau> \<MathOclValid> A) = (\<tau> \<MathOclValid> B)) \<and> (((\<tau> \<MathOclValid> \<MathOclNot> A) = (\<tau> \<MathOclValid> \<MathOclNot> B))))"
  apply (auto)
  apply (simp_all add:  is_FALSE_charn_local[symmetric])
  apply (simp_all only: not_valid not_invalid is_FALSE_charn_local)
  apply auto
  apply (simp_all only: isUndefined_charn_local[symmetric])
  apply (simp_all only: OclLocalValid_def)
  done




section {* The Theory of Strong and Strict Equality *}

subsection{* The Setup: $\HolOclCp$, strictness, definedness. *}

lemmas cp_strongEq = OclStrongEq_def   [THEN cp_lift2_fw, standard]
lemmas cp_strictEq = OclStrictEq_def   [THEN cp_lift2_fw, standard]

declare cp_strongEq [intro!] cp_strictEq [intro!]

lemma isDefined_strongEq[iff,simp]:"\<MathOclValid> \<MathOclIsDefined> X \<MathOclStrongEq> Y"
  by(auto simp: OclIsDefined_def OclStrongEq_def 
                lift2_def lift1_def lift0_def
                OclUndefined_def o_def OclTrue_def 
                DEF_def OclValid_def)

lemmas  isDefined_strictEq = OclStrictEq_def 
                             [THEN lift2_strictify_is_isdef_fw, 
                              standard]
lemmas  OclUndefined__strictEq = OclStrictEq_def 
                             [THEN lift2_undef1a_fw, 
                              standard]
lemmas  strictEq__undef = OclStrictEq_def 
                             [THEN lift2_undef2a_fw, 
                              standard]
declare isDefined_strictEq [simp] 
        OclUndefined__strictEq [simp] 
        strictEq__undef [simp]


subsection{* Properties of Strong Equality *}


lemma StrongEq_notvalid1[simp]:
" \<lbrakk>\<tau> \<MathOclValid> \<MathOclIsDefined> x; \<tau> \<MathOclValid> \<MathOclIsUndefined> y\<rbrakk> \<Longrightarrow> (x \<MathOclStrongEq> y) \<tau> = \<MathOclFalse> \<tau>"
by(simp add: localValidDefined2sem localValidUnDefined2sem DEF_def
                OclStrongEq_def OclFalse_def lift0_def lift2_def)

lemma StrongEq_notvalid2[simp]:
" \<lbrakk>\<tau> \<MathOclValid> \<MathOclIsUndefined> x; \<tau> \<MathOclValid> \<MathOclIsDefined> y\<rbrakk> \<Longrightarrow> (x \<MathOclStrongEq> y) \<tau> = \<MathOclFalse> \<tau>"
by(auto simp: localValidDefined2sem localValidUnDefined2sem DEF_def
              OclStrongEq_def OclFalse_def lift0_def lift2_def)

lemma StrongEq_valid_undef :
" \<lbrakk>\<tau> \<MathOclValid> \<MathOclIsUndefined> x; \<tau> \<MathOclValid> \<MathOclIsUndefined> y\<rbrakk> \<Longrightarrow> (x \<MathOclStrongEq> y) \<tau> = \<MathOclTrue> \<tau>"
txt{* will be proven in \tactic|OCL\_Logic| by:  \inlineholocl|by(ocl_hypsubst,simp)|
 *}
oops


lemma strictEq_is_strongEq_LE: 
  "\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsDefined> a; \<tau> \<MathOclValid> \<MathOclIsDefined> b\<rbrakk> \<Longrightarrow> (a \<MathOclStrictEq> b)\<tau>  = (a \<MathOclStrongEq> b)\<tau>"
  by(simp add: OclStrongEq_def OclStrictEq_def OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)



text {*
  In the theory files, this section contains the \acs{sml} code that
  implements \tactic{OCL\_normal\_tac}. 
*}
(*<*)
ML{*
val global2localValidity = (thm"global_vs_local_validity") RS sym

fun OCL_normal_tac (cs,ss) thm goal =
    EVERY[rtac ext goal,
          auto_tac (cs addSIs [thm],
                    ss addsimps [global2localValidity])]

fun OCL_normal thy thma = 
    let val thma  = fst(freeze_thaw thma);
        val prems = map HOLogic.dest_Trueprop (prems_of thma);
        val concl = HOLogic.dest_Trueprop(concl_of thma);
        val claset= (claset_of thy, simpset_of thy)
        fun conv_prem (Const("OCL_Logic_core.OclLocalValid",ty) $ _ $ E)=
                    (Const("OCL_Logic_core.OclValid",range_type ty) $ E)
      fun conv_concl(Const("op =",ty) $ (E1 $ c1) $ (E2 $ c2)) =
                let val ty''= (type_of c1) --> (domain_type ty)
                    val ty' = [ty'',ty''] ---> (range_type(range_type ty))
                in  if c1 = c2 
                    then Const("op =",ty') $ E1 $ E2
                    else raise Match
                end
      val cterm = cterm_of 
                      thy
                      (Logic.list_implies 
                          (map (HOLogic.mk_Trueprop o conv_prem) prems, 
                           (HOLogic.mk_Trueprop o conv_concl) concl))
    in prove_goalw_cterm [] cterm 
       (fn prems => [cut_facts_tac prems 1,
                     (* rtac thma 1, *)
                     OCL_normal_tac claset thma 1])
    end
*}

(*Does not work in this theory - but will work outside.
  See example OCL_Logic.thy.
 
ML{* val hhh = OCL_normal
               (the_context()) 
               (get_thm(the_context())(Name"strictEq_is_strongEq_LE")) *}

 *)
(*>*)  

lemma strictEq_is_strongEq[simp]:  (* Really [simp] ? *)
  "\<lbrakk>\<MathOclValid> \<MathOclIsDefined> a; \<MathOclValid> \<MathOclIsDefined> b\<rbrakk> \<Longrightarrow> a \<MathOclStrictEq> b = a \<MathOclStrongEq> b"
  by(tactic "OCL_normal_tac (get_local_clasimpset ctxt)
                            (thm\"strictEq_is_strongEq_LE\") 1")

(* local auxilliary lemma - should be elsewhere, but needed
   here \<dots> *)
lemma isDefined_if_valid: 
  "\<tau> \<MathOclValid> A \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> A"
  by (rule classical, 
      auto simp: OclTrue_def 
                 OclIsDefined_def lift0_def lift1_def lift2_def
                 o_def OclLocalValid_def)


lemma localvalid_strictEq_implies_localvalid_strongEq: 
  "\<lbrakk>\<tau> \<MathOclValid> a \<MathOclStrictEq> b \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> a \<MathOclStrongEq> b"
apply(frule isDefined_if_valid,
      simp only: OclStrictEq_def
                   [THEN lift2_strictify_implies_LocalValid_defined])
apply(auto simp: OclTrue_def  OclStrictEq_def OclStrongEq_def
                   OclIsDefined_def lift0_def lift1_def lift2_def
                   o_def OclLocalValid_def)
done


lemma strongEq_subst: 
  "\<lbrakk>a \<MathOclStrongEq> b = \<MathOclTrue>; P a = \<MathOclTrue>; cp P\<rbrakk> \<Longrightarrow> P b = \<MathOclTrue>"
  apply(unfold OclStrongEq_def OclTrue_def lift0_def lift1_def lift2_def)
  apply(rule ext) 
  apply(drule_tac x = "s" in fun_cong)
  apply(drule_tac x = "s" in fun_cong)
  apply(simp add: cp_def)
  apply(erule exE) 
  apply(rotate_tac -1)
  apply simp
  done

(* Powerful universal congruence *)
lemma strongEq_refl_UC[simp] : "(a \<MathOclStrongEq> a) =  \<MathOclTrue>"
  apply(rule ext)
  by   (simp add: OclStrongEq_def OclTrue_def 
                  lift0_def lift2_def)

lemma strongEq_refl:  "\<MathOclValid> a \<MathOclStrongEq> a" by simp

lemma strictEq_refl_UC[simp] : 
  "\<lbrakk>\<MathOclValid> \<MathOclIsDefined> a\<rbrakk> \<Longrightarrow> a \<MathOclStrictEq> a =  \<MathOclTrue>"
  by simp
 
lemma strongEq_sym:  
  "\<MathOclValid> a \<MathOclStrongEq> b \<Longrightarrow> \<MathOclValid>  b \<MathOclStrongEq> a"
  apply (unfold OclStrongEq_def OclTrue_def lift0_def lift2_def OclValid_def)
  apply (rule ext)
  apply (drule_tac x = "\<tau>" in fun_cong)
  apply simp
  done

lemma strongEq_trans:   
  "\<lbrakk>\<MathOclValid> a \<MathOclStrongEq> b; \<MathOclValid> b \<MathOclStrongEq> c\<rbrakk> \<Longrightarrow> \<MathOclValid> a \<MathOclStrongEq> c"
  apply(unfold OclStrongEq_def OclTrue_def lift0_def lift2_def OclValid_def) 
  apply (rule ext)
  apply (drule_tac x = "\<tau>" in fun_cong)
  apply (drule_tac x = "\<tau>" in fun_cong)
  apply simp
  done


lemma strongEq_refl_local:  "\<tau> \<MathOclValid> a \<MathOclStrongEq> a"
  by (simp add: OclStrongEq_def OclLocalValid_def 
                OclTrue_def lift0_def lift2_def) 

lemma strongEq_sym_local:  "\<tau> \<MathOclValid> a \<MathOclStrongEq> b \<Longrightarrow> \<tau> \<MathOclValid> b \<MathOclStrongEq> a"
  by (simp add: OclLocalValid_def OclStrongEq_def 
                OclTrue_def lift0_def lift2_def) 

lemma strongEq_trans_local:
  "\<lbrakk>\<tau> \<MathOclValid> a \<MathOclStrongEq> b; \<tau> \<MathOclValid> b \<MathOclStrongEq> c \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> a \<MathOclStrongEq> c"
  by (simp add: OclLocalValid_def OclStrongEq_def 
                OclTrue_def lift0_def lift2_def)

text{* 
  Extensionality must swap parameters and therefore
  looks quite inelegant \ldots *}
lemma strongEq_ext_local:
  "(\<And>x. \<tau> \<MathOclValid> f x \<MathOclStrongEq> g x) \<Longrightarrow> \<tau> \<MathOclValid> (\<lambda> X \<tau>. f \<tau> X ) \<MathOclStrongEq> (\<lambda> X \<tau>. g \<tau> X )"
  by (simp add: OclLocalValid_def OclStrongEq_def 
                OclTrue_def lift0_def lift2_def)

thm strongEq_ext_local

(* experiment: the SW combinator *)
(* strictly speaking unnecessary: comparison
   on higher-order objects in OCL is never required. *)
constdefs
  SW :: " ('b \<Rightarrow> 'a \<Rightarrow> 'c) \<Rightarrow> 'a \<Rightarrow> 'b \<Rightarrow> 'c"
  "SW f \<equiv> (\<lambda> X \<tau>. f \<tau> X )"

lemma SW_idem[simp]: "SW(SW f) = f"
  by(simp add: SW_def)

lemma strongEq_ext_local2:
  "(\<And>x. \<tau> \<MathOclValid> f x \<MathOclStrongEq> g x) \<Longrightarrow> \<tau> \<MathOclValid> (SW f) \<MathOclStrongEq> (SW g)"
  by (simp add: OclLocalValid_def OclStrongEq_def 
                OclTrue_def lift0_def lift2_def SW_def)



lemma strongEq_subst_local: 
  "\<lbrakk> \<tau> \<MathOclValid> a \<MathOclStrongEq> b; \<tau> \<MathOclValid> P a; cp P\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> P b"
  apply (simp add: OclStrongEq_def OclLocalValid_def cp_def
                   OclTrue_def lift0_def lift1_def lift2_def)
  apply (erule exE)
  apply (rotate_tac -1)
  apply simp
  done

lemmas strongEq_subst_local_sym = strongEq_sym_local  
                                  [THEN strongEq_subst_local]

subsection{* Technical Substitution Lemmas (for tactics) *}

lemma strongEq1:
"\<lbrakk>\<tau> \<MathOclValid> X\<MathOclStrongEq>Y; \<tau>  \<MathOclValid> P Y; cp P \<rbrakk> \<Longrightarrow> \<tau>  \<MathOclValid> P X"
by(drule strongEq_sym_local,
   erule strongEq_subst_local, simp)

lemmas strongEq1_rev = strongEq1[OF strongEq_sym_local]

lemma strongEq2:
"\<lbrakk>\<tau> \<MathOclValid> X\<MathOclStrongEq>Y; (\<tau> \<MathOclValid> P Y) = (\<tau> \<MathOclValid> P' Y); cp P; cp P'\<rbrakk> 
 \<Longrightarrow> (\<tau> \<MathOclValid> P X) = (\<tau> \<MathOclValid> P' X)"
apply(rule iffI)
apply(rule_tac X=X in strongEq1, 
      simp_all,drule sym, simp)
apply(erule strongEq1_rev, simp_all)
apply(rule_tac X=X in strongEq1, simp_all)
apply(simp, erule strongEq1_rev, simp_all)
done

lemmas strongEq2_rev = strongEq2[OF strongEq_sym_local]
           
lemma strongEq3:
"\<lbrakk>\<tau> \<MathOclValid> X\<MathOclStrongEq>Y; P Y \<tau> = P' Y \<tau>; cp P; cp P' \<rbrakk> \<Longrightarrow> P X \<tau> = P' X \<tau>"
apply(rule_tac P=P and P' = P' in ocl_cp_subst3, simp_all)
apply(simp add: OclLocalValid_def OclStrongEq_def strictify_def
             o_def lift0_def lift1_def lift2_def OclNot_def
             OclTrue_def OclFalse_def  OclUndefined_def DEF_def)
done

lemmas strongEq3_rev = strongEq3[OF strongEq_sym_local]

lemma strongEq1_fw:
"\<lbrakk>\<tau> \<MathOclValid> P X; \<tau> \<MathOclValid> Y \<MathOclStrongEq> X; cp P\<rbrakk>  \<Longrightarrow> \<tau> \<MathOclValid> P Y"
by(erule strongEq1,simp)

lemma strongEq2_fw:
"\<lbrakk>(\<tau> \<MathOclValid> P X) = (\<tau> \<MathOclValid> P' X); \<tau> \<MathOclValid> Y \<MathOclStrongEq> X; cp P; cp P'\<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P X) = (\<tau> \<MathOclValid> P' X)"
by(erule strongEq2,simp_all)

lemma strongEq3_fw:
"\<lbrakk>P X \<tau> = P' X \<tau>; \<tau> \<MathOclValid> Y \<MathOclStrongEq> X; cp P; cp P'\<rbrakk> \<Longrightarrow> P Y \<tau> = P' Y \<tau>"
by(erule strongEq3,simp_all)



lemma strongEq1_fw2:
"\<lbrakk>\<tau> \<MathOclValid> P X; \<tau> \<MathOclValid> X \<MathOclStrongEq> Y; cp P\<rbrakk>  \<Longrightarrow> \<tau> \<MathOclValid> P Y"
by(erule strongEq1_rev,simp)

lemma strongEq2_fw2:
"\<lbrakk>(\<tau> \<MathOclValid> P X) = (\<tau> \<MathOclValid> P' X); \<tau> \<MathOclValid> X \<MathOclStrongEq> Y; cp P; cp P'\<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P X) = (\<tau> \<MathOclValid> P' X)"
by(erule strongEq2_rev,simp_all)

lemma strongEq3_fw2:
"\<lbrakk>P X \<tau> = P' X \<tau>; \<tau> \<MathOclValid> X \<MathOclStrongEq> Y; cp P; cp P'\<rbrakk> \<Longrightarrow> P Y \<tau> = P' Y \<tau>"
by(erule strongEq3_rev,simp_all)


section {* Automated Canonization and Rewriting in \OCL *}
text {* \OCL offers a number of equivalences: From strict and
strong equality, we have three types of congruences/equivalences,
such that reasoning about them is tedious and must be tactically
supported. The ultimate goal of this section is to work out
two tactic procedures for this purpose that pursue this goal.
As basis for these tactic procedures, a number of rules
and rule variants must be derived, whose existence is merely
justified by deduction technical reasons. *}
(**************************************************************************)
subsection {*  validity propagation rules *}
(**************************************************************************)


lemmas strictEq1 = strongEq1[OF localvalid_strictEq_implies_localvalid_strongEq]
lemmas strictEq2 = strongEq2[OF localvalid_strictEq_implies_localvalid_strongEq]
lemmas strictEq3 = strongEq3[OF localvalid_strictEq_implies_localvalid_strongEq]

lemmas strictEq1_rev = strongEq1_rev
                       [OF localvalid_strictEq_implies_localvalid_strongEq]
lemmas strictEq2_rev = strongEq2_rev
                       [OF localvalid_strictEq_implies_localvalid_strongEq]
lemmas strictEq3_rev = strongEq3_rev
                       [OF localvalid_strictEq_implies_localvalid_strongEq]



(* NOTE: can be used comfortably as elimination rules *)
lemma subst_LJ_TRUE: 
  "\<lbrakk>\<tau> \<MathOclValid> X; \<tau> \<MathOclValid> P \<MathOclTrue>; cp P \<rbrakk> \<Longrightarrow> \<tau>  \<MathOclValid> P X"
  apply (simp add: OclLocalValid_def,drule sym)
  apply (erule_tac P = "P" in ocl_cp_subst,auto)
  done


lemma subst_LJ_TRUE_fw: 
  "\<lbrakk>\<tau> \<MathOclValid> X; \<tau> \<MathOclValid> P X; cp P \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> P \<MathOclTrue>"
  apply (simp add: OclLocalValid_def)
  apply (erule_tac P = "P" in ocl_cp_subst)
  apply auto
  done
  (* for interactive use in drule's only *)

lemma subst_LJ_TRUE_fw_rev: 
  "\<lbrakk>\<tau> \<MathOclValid> P X; \<tau> \<MathOclValid> X; cp P \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> P \<MathOclTrue>"
  apply (simp add: OclLocalValid_def)
  apply (erule_tac P = "P" in ocl_cp_subst) back back
  apply auto
  done



lemma subst_LE_TRUE: 
"\<lbrakk>\<tau> \<MathOclValid> X; P \<MathOclTrue> \<tau> = P' \<MathOclTrue> \<tau>; cp P; cp P'\<rbrakk> \<Longrightarrow> P X \<tau> = P' X \<tau>"
apply(rule local_validity_propagation2 [simplified is_TRUE_charn_local]) back
by   auto


lemma subst_LE_TRUE_fw:
"\<lbrakk>\<tau> \<MathOclValid> X; P X \<tau> = P' X \<tau> ; cp P; cp P'\<rbrakk> \<Longrightarrow> P \<MathOclTrue> \<tau> = P' \<MathOclTrue> \<tau>"
apply(simp add: OclLocalValid_def)
apply(rule trans, rule sym)
apply(erule_tac \<tau>=\<tau> and P=P in cp_charn, simp) 
apply(rule trans, assumption)
apply(erule_tac \<tau>=\<tau> and P=P' in cp_charn, simp) 
done


lemma subst_LE_TRUE_fw_rev:
"\<lbrakk>P (X::'a\<Rightarrow>Boolean_0) \<tau> = P' X \<tau> ; \<tau> \<MathOclValid> X; cp P; cp P'\<rbrakk> \<Longrightarrow> P \<MathOclTrue> \<tau> = P' \<MathOclTrue> \<tau>"
by(erule subst_LE_TRUE_fw, auto)


lemma subst_LJE_TRUE: 
  "\<lbrakk>\<tau> \<MathOclValid> X; (\<tau> \<MathOclValid> P \<MathOclTrue>) = (\<tau> \<MathOclValid> P' \<MathOclTrue>); cp P; cp P'\<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P X)=(\<tau> \<MathOclValid> P' X)"
  by(erule subst_LE_TRUE, auto)

lemma subst_LJE_TRUE_fw: 
  "\<lbrakk>\<tau> \<MathOclValid> X; (\<tau> \<MathOclValid> P X)=(\<tau> \<MathOclValid> P' X); cp P; cp P'\<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P \<MathOclTrue>) = (\<tau> \<MathOclValid> P' \<MathOclTrue>)"
  by(erule subst_LE_TRUE_fw, auto)

lemma subst_LJE_TRUE_fw_rev: 
  "\<lbrakk>(\<tau> \<MathOclValid> P X)=(\<tau> \<MathOclValid> P' X); \<tau> \<MathOclValid> X; cp P; cp P'\<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P \<MathOclTrue>) = (\<tau> \<MathOclValid> P' \<MathOclTrue>)"
  by(erule subst_LJE_TRUE_fw, auto)


lemma subst_TRUE_LJE:
  "\<lbrakk>\<tau> \<MathOclValid> X; cp P \<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P X)=(\<tau> \<MathOclValid> P \<MathOclTrue>)"
  apply(auto)
  apply(erule subst_LJ_TRUE_fw, auto)
  apply(erule subst_LJ_TRUE,auto)
  done
  (* can probably not be processed by rewriter \<dots> *)






lemma subst_LJ_FALSE: 
  "\<lbrakk>\<tau> \<MathOclValid> \<MathOclNot> X; \<tau> \<MathOclValid> P \<MathOclFalse>; cp P \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> P X"
  apply (drule is_FALSE_charn_local [THEN iffD2])
  apply (rotate_tac -1)
  apply (drule sym)
  apply (simp add: OclLocalValid_def)
  apply (erule_tac P = "P" in ocl_cp_subst)
  apply auto
  done

lemma subst_LJ_FALSE_fw: 
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclNot> X; \<tau> \<MathOclValid> P X; cp P \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> P \<MathOclFalse>"
  apply (drule is_FALSE_charn_local [THEN iffD2])
  apply (rotate_tac -1)
  apply (simp add: OclLocalValid_def)
  apply (erule_tac P = "P" in ocl_cp_subst)
  apply auto
  done

lemma subst_LJ_FALSE_fw_rev: 
  "\<lbrakk> \<tau> \<MathOclValid> P X; \<tau> \<MathOclValid> \<MathOclNot> X; cp P \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> P \<MathOclFalse>"
  apply (erule subst_LJ_FALSE_fw)
  apply auto
  done

lemma subst_LE_FALSE: 
"\<lbrakk>\<tau> \<MathOclValid> \<MathOclNot> X; P \<MathOclFalse> \<tau> = P' \<MathOclFalse> \<tau>; cp P; cp P'\<rbrakk> \<Longrightarrow> P X \<tau> = P' X \<tau>"
apply(rule local_validity_propagation3 [simplified is_FALSE_charn_local]) back
by   auto

lemma subst_LE_FALSE_fw:
"\<lbrakk>\<tau> \<MathOclValid> \<MathOclNot> X; P X \<tau> = P' X \<tau> ; cp P; cp P'\<rbrakk> \<Longrightarrow> P \<MathOclFalse> \<tau> = P' \<MathOclFalse> \<tau>"
apply(simp only: is_FALSE_charn_local[symmetric])
apply(rule trans, rule sym)
apply(erule_tac \<tau>=\<tau> and P=P in cp_charn, simp) 
apply(rule trans, assumption)
apply(erule_tac \<tau>=\<tau> and P=P' in cp_charn, simp) 
done

lemma subst_LE_FALSE_fw_rev:
"\<lbrakk>P X (\<tau>::'a) = P' X \<tau>; \<tau> \<MathOclValid> \<MathOclNot> X; cp P; cp P'\<rbrakk> \<Longrightarrow> P \<MathOclFalse> \<tau> = P' \<MathOclFalse> \<tau>"
by(erule subst_LE_FALSE_fw, auto)

lemma subst_LJE_FALSE: 
  "\<lbrakk>\<tau> \<MathOclValid> \<MathOclNot> X; (\<tau> \<MathOclValid> P \<MathOclFalse>) = (\<tau> \<MathOclValid> P' \<MathOclFalse>); cp P; cp P'\<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P X)=(\<tau> \<MathOclValid> P' X)"
  by(erule subst_LE_FALSE, auto)

lemma subst_LJE_FALSE_fw: 
  "\<lbrakk>\<tau> \<MathOclValid> \<MathOclNot> X; (\<tau> \<MathOclValid> P X)=(\<tau> \<MathOclValid> P' X); cp P; cp P'\<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P \<MathOclFalse>) = (\<tau> \<MathOclValid> P' \<MathOclFalse>)"
  by(erule subst_LE_FALSE_fw, auto)

lemma subst_LJE_FALSE_fw_rev: 
  "\<lbrakk>(\<tau> \<MathOclValid> P X)=(\<tau> \<MathOclValid> P' X); \<tau> \<MathOclValid> \<MathOclNot> X; cp P; cp P'\<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P \<MathOclFalse>) = (\<tau> \<MathOclValid> P' \<MathOclFalse>)"
  by(erule subst_LJE_FALSE_fw, auto)


lemma subst_FALSE_LJE:
  "\<lbrakk>\<tau> \<MathOclValid> \<MathOclNot> X; cp P \<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P X)=(\<tau> \<MathOclValid> P \<MathOclFalse>)"
  apply(auto)
  apply(erule subst_LJ_FALSE_fw, auto)
  apply(erule subst_LJ_FALSE,auto)
  done




lemma subst_LJ_undef: 
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsUndefined> X; \<tau> \<MathOclValid> P \<MathOclUndefined>; cp P \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> P X"
  apply (drule isUndefined_charn_local [THEN iffD2])
  apply (rotate_tac -1)
  apply (drule sym)
  apply (simp add: OclLocalValid_def)
  apply (erule_tac P = "P" in ocl_cp_subst)
  apply auto
  done


lemma subst_LJ_undef_fw: 
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsUndefined> X; \<tau> \<MathOclValid> P X; cp P \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> P \<MathOclUndefined>"
  apply (drule isUndefined_charn_local [THEN iffD2])
  apply (rotate_tac -1)
  apply (simp add: OclLocalValid_def)
  apply (erule_tac P = "P" in ocl_cp_subst)
  apply auto
  done

lemma subst_LJ_undef_fw_rev: 
  "\<lbrakk> \<tau> \<MathOclValid> P X; \<tau> \<MathOclValid> \<MathOclIsUndefined> X; cp P \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> P \<MathOclUndefined>"
  by (erule subst_LJ_undef_fw, auto)



lemma subst_LE_undef:
"\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsUndefined> X; P \<MathOclUndefined> \<tau> = P' \<MathOclUndefined> \<tau>; cp P; cp P'\<rbrakk> \<Longrightarrow> P X \<tau> = P' X \<tau> "
  apply (rule_tac P=P and \<tau>=\<tau> in ocl_cp_subst3)
  apply (simp_all)
  apply (simp add: OclLocalValid_def OclIsDefined_def strictify_def
                   OclNot_def o_def lift0_def lift1_def lift2_def
                   OclTrue_def DEF_def OclUndefined_def)
  done  


lemma subst_LE_undef_fw:
"\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsUndefined> X;P X \<tau> = P' X \<tau> ; cp P; cp P'\<rbrakk> \<Longrightarrow>  P \<MathOclUndefined> \<tau> = P' \<MathOclUndefined> \<tau>"
  apply (rule_tac P=P and \<tau>=\<tau> in ocl_cp_subst3)
  apply (simp_all)
  apply (simp add: OclLocalValid_def OclIsDefined_def strictify_def
                   OclNot_def o_def lift0_def lift1_def lift2_def
                   OclTrue_def DEF_def OclUndefined_def)
  done  

lemma subst_LE_undef_fw_rev:
"\<lbrakk>P (X::'a\<Rightarrow>'b::bot) \<tau> = P' X \<tau>; \<tau> \<MathOclValid> \<MathOclIsUndefined> X; cp P; cp P'\<rbrakk> \<Longrightarrow>  P \<MathOclUndefined> \<tau> = P' \<MathOclUndefined> \<tau>"
by(erule subst_LE_undef_fw, auto)

lemma subst_LJE_undef: 
  "\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsUndefined> X; (\<tau> \<MathOclValid> P \<MathOclUndefined>) = (\<tau> \<MathOclValid> P' \<MathOclUndefined>); cp P; cp P'\<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P X)=(\<tau> \<MathOclValid> P' X)"
  by(erule subst_LE_undef, auto)

lemma subst_LJE_undef_fw: 
  "\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsUndefined> X; (\<tau> \<MathOclValid> P X)=(\<tau> \<MathOclValid> P' X); cp P; cp P'\<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P \<MathOclUndefined>) = (\<tau> \<MathOclValid> P' \<MathOclUndefined>)"
  by(erule subst_LE_undef_fw, auto)

lemma subst_LJE_undef_fw_rev: 
  "\<lbrakk>(\<tau> \<MathOclValid> P X)=(\<tau> \<MathOclValid> P' X); \<tau> \<MathOclValid> \<MathOclIsUndefined> X; cp P; cp P'\<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P \<MathOclUndefined>) = (\<tau> \<MathOclValid> P' \<MathOclUndefined>)"
  by(erule subst_LJE_undef_fw, auto)

lemma subst_undef_LJE:
  "\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsUndefined> X; cp P \<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> P X)=(\<tau> \<MathOclValid> P \<MathOclUndefined>)"
  apply(auto)
  apply(erule subst_LJ_undef_fw, auto)
  apply(erule subst_LJ_undef,auto)
  done


lemmas strictEq1_fw2 = strongEq1_fw2
                       [OF _ localvalid_strictEq_implies_localvalid_strongEq]
lemmas strictEq2_fw2 = strongEq2_fw2
                       [OF _ localvalid_strictEq_implies_localvalid_strongEq]
lemmas strictEq3_fw2 = strongEq3_fw2
                       [OF _ localvalid_strictEq_implies_localvalid_strongEq]

lemmas strictEq1_fw  = strongEq1_fw
                       [OF _ localvalid_strictEq_implies_localvalid_strongEq]
lemmas strictEq2_fw  = strongEq2_fw
                       [OF _ localvalid_strictEq_implies_localvalid_strongEq]
lemmas strictEq3_fw  = strongEq3_fw
                       [OF _ localvalid_strictEq_implies_localvalid_strongEq]


subsection{* Implementation of \tactic{OCL\_hyp\_subst\_tac} *}
text{* 
  These rules give some ideas for automation:
  \begin{enumerate}
  \item \tactic{OCL\_norm\_tac} computes local norm of
  validities: 
  $\isasymMathOclValid  A \isasymLongrightarrow  \isasymMathOclValid  B$ 
  is converted
  $\isasymtau  \isasymMathOclValid  A \isasymLongrightarrow  \isasymtau  \isasymMathOclValid  B$
  \ldots. This is done forward and backward style.
\item \tactic{OCL\_absurd\_tac} should decide if a subgoal
      is \OCL absurd.
\item \tactic{OCL\_hyp\_subst\_tac} for variables or parameters $A$,
      which uses: 
      \begin{align}
        A = t, t = A           
          &&\text{(standard \tactic{hyp\_subst\_tac})}\\
        A \tau = t \tau, t \tau = A \tau   
          &&\text{(local congruences)}\\  
        \isasymMathOclValid  \isasymMathOclIsUndefined  A, \tau  \isasymMathOclValid  \isasymMathOclIsUndefined  A       
          &&\text{(undefinednesses local and global)}\\
       \isasymMathOclValid  \isasymMathOclNot  A,  \tau \isasymMathOclValid  \isasymMathOclNot  A       
           &&\text{(falsities  local and global)}\\
       \isasymMathOclValid  A \isasymtriangleq  t, \isasymMathOclValid  t \isasymtriangleq  A       
         &&\text{(strong equalities local and global)}\\
       \tau \isasymMathOclValid  A \isasymtriangleq  t, \tau \isasymMathOclValid  A \isasymtriangleq  t 
      \end{align}
 
\item \tactic{OCL\_def\_hyp\_split\_tac}:
      \begin{itemize}
      \item atomizes definedness hypothesises , e.g.,  
                   $\isasymMathOclIsDefined (f\ap A\ap B) = 
             \isasymMathOclIsDefined A \isasymMathOclAnd \isasymMathOclIsDefined B$ 
               \ldots 
      \item case-splits undefinedness hypothesis  
                   $\isasymMathOclIsUndefined(f\ap A\ap B) 
         = \isasymMathOclIsUndefined(A) \vee \isasymMathOclIsUndefined(B)$ \ldots
      \item and performs \tactic{safe\_tac} and \tactic{OCL\_hyp\_subst\_tac}.
            (but no global simplification afterwards
             in order to do not too much in a black box.
      \end{itemize}
\item @{text "OCL_simp"} and @{text "OCL_asm_simp"}. New rewriter
      that can cope with $\_ \isasymMathOclStrongEq \_$ and $\_ \isasymMathOclStrictEq \_$ 
      and its specific format of the subst rule. Future: use standard rewriter
      with a fake subst rule on OCL, transform the proof 
      objects and refeed this into the kernel?
\item integrate into simp-procedure as looper or subgoaler.
\end{enumerate}

*}

lemmas DEF_rules = subst_LE_TRUE [of _ "\<MathOclIsDefined> X",standard]
                   subst_LJE_TRUE[of _ "\<MathOclIsDefined> X",standard]
                   subst_LJ_TRUE [of _ "\<MathOclIsDefined> X",standard]

lemmas DEF_fw_rev_rules = 
                   subst_LJ_TRUE_fw_rev [of _ _ "\<MathOclIsDefined> X",standard]
                   subst_LJE_TRUE_fw_rev[of _ _ "\<MathOclIsDefined> X",standard]
                   subst_LE_TRUE_fw_rev [of _ "\<MathOclIsDefined> X",standard]
(*<*) 
ML{*
(* bindings - for efficiency ... *)
val  TRUE_fw_rev_rules =  
     (map thm ["subst_LJ_TRUE_fw_rev","subst_LJE_TRUE_fw_rev","subst_LE_TRUE_fw_rev"])
val  TRUE_rules = rev
     (map thm ["subst_LE_TRUE","subst_LJE_TRUE","subst_LJ_TRUE"]) 
val  FALSE_fw_rev_rules =  
     (map thm ["subst_LJ_FALSE_fw_rev","subst_LJE_FALSE_fw_rev","subst_LE_FALSE_fw_rev"]) 
val  FALSE_rules = rev
     (map thm ["subst_LE_FALSE","subst_LJE_FALSE","subst_LJ_FALSE"]) 
val  undef_fw_rev_rules = 
     (map thm ["subst_LJ_undef_fw_rev","subst_LJE_undef_fw_rev","subst_LE_undef_fw_rev"]) 
val  undef_rules = rev
     (map thm ["subst_LE_undef","subst_LJE_undef","subst_LJ_undef"]) 
val  DEF_rules= rev(thms "DEF_rules");
val  DEF_fw_rev_rules = thms "DEF_fw_rev_rules";
val  strongEq_rules = 
     (map thm ["strongEq1","strongEq2","strongEq3"])
val  strongEq_rev_rules =
     (map thm ["strongEq1_rev","strongEq2_rev","strongEq3_rev"])
val  strongEq_fw_rules = 
     (map thm ["strongEq1_fw","strongEq2_fw","strongEq3_fw"])
val  strongEq_fw2_rules = 
     (map thm ["strongEq1_fw2","strongEq2_fw2","strongEq3_fw2"])
val  strictEq_rules = 
     (map thm ["strictEq1","strictEq2","strictEq3"])
val  strictEq_rev_rules = 
     (map thm ["strictEq1_rev","strictEq2_rev","strictEq3_rev"])
val  strictEq_fw_rules  =
     (map thm ["strictEq1_fw","strictEq2_fw","strictEq3_fw"])
val  strictEq_fw2_rules = 
     (map thm ["strictEq1_fw2","strictEq2_fw2","strictEq3_fw2"])

*}



ML
{* (* Term Selection and Matching Infrastructure *)

fun isundef_T T (Const("OCL_Boolean.OclNot",_) $
                                   (Const("OCL_Boolean.OclIsDefined",_)$T'))
                  = (T = T')
   |isundef_T _ _ = false;
           

datatype res = UNDEF of term * term * term |
               FALSE of term * term * term |
               TRUE  of term * term * term |
               DEF   of term * term * term | 
                     (* not used at moment, for case-splits *)
               EQLS  of term * term * term | 
               EQRS  of term * term * term | 
               EQLW  of term * term * term | 
               EQRW  of term * term * term |
               NORES

fun get_canon_free (UNDEF X) = X
   |get_canon_free (FALSE X) = X
   |get_canon_free (TRUE  X) = X
   |get_canon_free (DEF   X) = X
   |get_canon_free (EQLS  X) = X
   |get_canon_free (EQRS  X) = X
   |get_canon_free (EQLW  X) = X
   |get_canon_free (EQRW  X) = X
   (* May fail *)


(* find_canon_free finds first canonizable free variable in 
   assumption list (i.e. Free = Const or related pattern), if none,
   it returns first replaceable variable (i.e.Free = Free or 
   related pattern), if none, it returns NORES.
   
   Performs beta-eta-reduction, but avoids uneccessary
   normalizations. 
   - \<tau> \<MathOclValid> E \<MathOclStrictEq> var, \<tau> \<MathOclValid> var \<MathOclStrictEq> E
   - \<tau> \<MathOclValid> E \<MathOclStrongEq> var, \<tau> \<MathOclValid> var \<MathOclStrongEq> E
   - \<tau> \<MathOclValid> \<MathOclIsUndefined> var,\<tau> \<MathOclValid> \<MathOclIsDefined> var,
   - \<tau> \<MathOclValid> var
   - \<tau> \<MathOclValid> \<MathOclNot> var

   Badly missing: 
   - \<not> \<tau> \<MathOclValid> \<MathOclIsDefined> E, and "semantic" versions E \<tau> = E' \<tau>. 
     Work-around: simplify into one of the above formats.
 

 *)
fun find_canon_free S = 
    let val stack = ref(NONE:res option)
        fun H [] = (case !stack of 
                     NONE => NORES
                   | SOME T => T) 
          | H (term::S) =
           (case Pattern.beta_eta_contract(HOLogic.dest_Trueprop term) of 
          (*(* " \<MathOclIsUndefined> A ST = \<MathOclTrue> ST " *)
            EQ as (Const ("op =",_) $
                     (Const("OCL_Boolean.OclNot",_) $
                        (Const("OCL_Boolean.OclIsDefined",_)$
                           Free(name,ty)) $  ST1) $
                    (Const("OCL_Boolean.OclTrue",_) $ ST2))
              => if ST1 = ST2  then UNDEF(EQ,ST1,Free(name,ty)) 
                 else H S
          |*)
             (* " ST \<MathOclValid> \<MathOclIsUndefined> A" *)
            EQ as (Const ("OCL_Logic_core.OclLocalValid",_) $ ST $ 
                      (Const("OCL_Boolean.OclNot",_) $
                         (Const("OCL_Boolean.OclIsDefined",_)$Free(name,ty))))
              => UNDEF(EQ,ST,Free(name,ty))
       (* | (* " ST \<MathOclValid> \<MathOclIsDefined> A" *)
            EQ as (Const ("OCL_Logic_core.OclLocalValid",_) $ ST $ 
                         (Const("OCL_Boolean.OclIsDefined",_)$Free(name,ty)))
              => DEF(EQ,ST,Free(name,ty)) *)
       (* | (* "\<MathOclNot> A ST = \<MathOclTrue> ST"  *) 
                   EQ as (Const ("op =",_) $
                   ((Const("OCL_Boolean.OclNot",_) $ 
                      Free (name,ty)) $ ST1) $
                 (Const ("OCL_Boolean.OclTrue",_) $ ST2))
              => if ST1 = ST2  then FALSE(EQ,ST1,Free(name,ty)) 
                 else H S *)
          | (* "ST \<MathOclValid> \<MathOclNot> A" *)
            EQ as (Const ("OCL_Logic_core.OclLocalValid",_) $ ST $ 
                      (Const("OCL_Boolean.OclNot",_) $ Free (name,ty)))
              => FALSE(EQ,ST,Free(name,ty))
       (* | (* "A ST = \<MathOclTrue> ST"  *) 
            EQ as (Const ("op =",_) $ 
                       (Free (name,ty) $ ST1) $
                           (Const("OCL_Boolean.OclTrue",_)$ST2))
              => if ST1 = ST2  then TRUE(EQ,ST1,Free(name,ty)) 
                 else H S *)
          | (* "ST \<MathOclValid> A"       *)
            EQ as (Const ("OCL_Logic_core.OclLocalValid",_) 
                    $ ST $ Free (name,ty))
              => TRUE(EQ,ST,Free(name,ty))
          | (* "\<tau> \<MathOclValid> Free(_,_) \<MathOclStrongEq> Free(name,ty)"   *)
            EQ as (Const("OCL_Logic_core.OclLocalValid",_) $ ST $
                      (Const ("OCL_Boolean.OclStrongEq",_ ) 
                         $ Free ZZ $ Free (name,ty)))
              => (case !stack of 
                   NONE   => if ZZ = (name,ty) then ()
                             else stack:=SOME(EQRS(EQ,ST,Free(name,ty)))
                 | SOME _ => ();
                 H S)
          | (* "\<tau> \<MathOclValid> _ \<MathOclStrongEq> Free(name,ty)"   *)
            EQ as (Const("OCL_Logic_core.OclLocalValid",_) $ ST $
                    (Const ("OCL_Boolean.OclStrongEq",_ ) 
                       $ C $ Free (name,ty)))
              => EQRS(EQ,ST,Free(name,ty))
          | (* "\<tau> \<MathOclValid> Free(name,ty) \<MathOclStrongEq> _"   *)
            EQ as (Const("OCL_Logic_core.OclLocalValid",_) $ ST $
               (Const ("OCL_Boolean.OclStrongEq",_ ) $ Free(name,ty) $ C))
              => EQLS(EQ,ST,Free(name,ty))
          | (* "\<tau> \<MathOclValid> Free(_,_) \<MathOclStrictEq> Free(name,ty)"   *)
            EQ as (Const("OCL_Logic_core.OclLocalValid",_) $ ST $
                      (Const ("OCL_Boolean.OclStrictEq",_ ) 
                         $ Free ZZ $ Free (name,ty)))
              => (case !stack of 
                   NONE   => if ZZ = (name,ty) then ()
                             else stack:=SOME(EQRS(EQ,ST,Free(name,ty)))
                 | SOME _ => ();
                 H S)
          | (* "\<tau> \<MathOclValid> _ \<MathOclStrictEq> Free(name,ty)"   *)
            EQ as (Const("OCL_Logic_core.OclLocalValid",_) $ ST $
               (Const ("OCL_Boolean.OclStrictEq",_ ) $ C $ Free (name,ty)))
              => EQRW(EQ,ST,Free(name,ty))
          | (* "\<tau> \<MathOclValid> Free(name,ty) \<MathOclStrictEq> _"   *)
            EQ as (Const("OCL_Logic_core.OclLocalValid",_) $ ST $
                (Const ("OCL_Boolean.OclStrictEq",_ ) $ Free(name,ty) $ C))
              => EQLW(EQ,ST,Free(name,ty))
          | _ => H S)
       handle TERM _ => H S (* dest_Trueprop *)
  in H S end;


(* find_rw_rule finds first occurence of one of
   the following rule patterns in a term list
   (may be assumption list, may be user defined):
   - E \<tau> = E' \<tau>
   - \<tau> \<MathOclValid> E \<MathOclStrictEq> E'
   - \<tau> \<MathOclValid> E \<MathOclStrongEq> E'
   - \<tau> \<MathOclValid> \<MathOclIsUndefined> E,\<tau> \<MathOclValid> \<MathOclIsDefined> E,
   - \<tau> \<MathOclValid> E
   - \<tau> \<MathOclValid> \<MathOclNot> E

   Badly missing: 
   - \<not> \<tau> \<MathOclValid> \<MathOclIsDefined> E
   
   Performs beta-eta-reduction, but avoids uneccessary
   normalizations.  *)
fun find_rw_rule S = 
    let val stack = ref(NONE:res option)
        fun H [] = (case !stack of 
                     NONE => NORES
                   | SOME T => T) 
          | H (term::S) =
           (case Pattern.beta_eta_contract(HOLogic.dest_Trueprop term) of 
             (* " ST \<MathOclValid> \<MathOclIsUndefined> A"   *)
             PPAT as (Const ("OCL_Logic_core.OclLocalValid",_) $ ST $ 
                      (Const("OCL_Boolean.OclNot",_) $
                         (Const("OCL_Boolean.OclIsDefined",_)$ LHS)))
              => UNDEF(PPAT,ST,LHS $ Const("OCL_Boolean.OclUndefined",dummyT))
          | (* " ST \<MathOclValid> \<MathOclIsDefined> A"    *)
             PPAT as (Const ("OCL_Logic_core.OclLocalValid",_) $ ST $ 
                         (Const("OCL_Boolean.OclIsDefined",_)$ LHS))
              => DEF(PPAT,ST,LHS $ Const("OCL_Boolean.OclTrue",dummyT))
          | (* "ST \<MathOclValid> \<MathOclNot> A"     *)
            PPAT as (Const ("OCL_Logic_core.OclLocalValid",_) $ ST $ 
                      (Const("OCL_Boolean.OclNot",_) $ LHS))
              => FALSE(PPAT,ST,LHS $ Const("OCL_Boolean.OclFalse",dummyT))
          | (* "ST \<MathOclValid> A"       *)
            PPAT as (Const ("OCL_Logic_core.OclLocalValid",_) 
                    $ ST $ LHS)
              => TRUE(PPAT,ST,LHS $ Const("OCL_Boolean.OclTrue",dummyT))
          | (* "\<tau> \<MathOclValid> LHS \<MathOclStrongEq> _"  *)
            PPAT as (Const("OCL_Logic_core.OclLocalValid",_) $ ST $
               (Const ("OCL_Boolean.OclStrongEq",_ ) $ LHS $ RHS))
              => EQLS(PPAT,ST,LHS $ RHS)
          | (* "\<tau> \<MathOclValid> LHS \<MathOclStrictEq> _"  *)
            PPAT as (Const("OCL_Logic_core.OclLocalValid",_) $ ST $
                (Const ("OCL_Boolean.OclStrictEq",_ ) $ LHS $ RHS))
              => EQLW(PPAT,ST,LHS $ RHS)
          | PPAT as (Const ("op =",ty2) $ (LHS$ST) $ (RHS$ST'))
            => if ST = ST'  
               then EQRW(PPAT,ST,LHS $ RHS)
                    (* By "abus de langue \<dots>"  ;-)
                       EQRW used for encoding global equivalences. *)
               else H S
          | _ => H S)
       handle TERM _ => H S (* dest_Trueprop *)
  in H S end;




datatype asmres = Jdmt       of term * typ
                | JdmtEquiv  of term * term * typ 
                | GlobCong   of term * term * typ 
                | NoAsmRes


fun find_asm_with_canon_free (patt,PPAT)  [] = NoAsmRes
   |find_asm_with_canon_free ((patt as Free(n,ty)),PPAT)(pattern::R) =
    (* PPAT  : equality under consideration
       patt: pattern, i.e. canonizable variable         *)
    let  val redx = Pattern.beta_eta_contract(HOLogic.dest_Trueprop pattern) in
    if   PPAT = redx
    then find_asm_with_canon_free (patt,PPAT) R
    else case redx of 
         term as (Const ("OCL_Logic_core.OclLocalValid",ty2) $ ST $ E) 
         => if patt mem (term_frees E)
            then Jdmt(absfree (n,ty,E), 
                      ty --> domain_type(range_type ty2)) 
            else find_asm_with_canon_free (patt,PPAT) R
        |term as (Const ("op =",_) 
                  $ (Const ("OCL_Logic_core.OclLocalValid",ty2)  $ ST  $ E)
                  $ (Const ("OCL_Logic_core.OclLocalValid",ty2') $ ST' $ E') )
         => if patt mem (term_frees term) 
            then JdmtEquiv(absfree (n,ty,E),
                           absfree (n,ty,E'), 
                           ty --> domain_type(range_type ty2)) 
            else find_asm_with_canon_free (patt,PPAT) R
        |term as (Const ("op =",ty2) $ (E$ST) $ (E'$ST'))
         => if patt mem (term_frees (E$E')) 
            then GlobCong(absfree (n,ty,E),
                          absfree (n,ty,E'),  
                          [ty,type_of ST] ---> (domain_type ty2))
            else find_asm_with_canon_free (patt,PPAT) R 
        |_ =>  find_asm_with_canon_free (patt,PPAT) R
     end
     handle TERM _ => find_asm_with_canon_free (patt,PPAT) R; (* HOLogic.dest_Trueprop *)

*}

ML{* (* small debugging infra structure \<dots> *)
print_depth 20; 
fun the_state () = snd(snd(Proof.get_goal(
                      Toplevel.proof_of(Toplevel.get_state()))));
fun the_claset() = claset_of(the_context())
(* equivalent to toplevel() in isar *)
val OCL_hypsubst_trace = ref(false);
val OCL_hypsubst_full_trace = ref(false);
fun T C = cond_timeit (!OCL_hypsubst_full_trace);

fun prin_tac f = if !OCL_hypsubst_full_trace then print_tac (f()) 
                 else if !OCL_hypsubst_trace then (tracing (f()); all_tac)
                      else all_tac

*}

ML{*

datatype ('a,'b) sum = Inl of 'a | Inr of 'b

fun subst_asm_gen (matchfun:term * term -> term list -> asmres) thm_sum
                  TS S cla (thmS as [A,B,C]) PPAT (ST:term) PAT goal state =
(* val subst_asm_gen: (term * term -> term list -> asmres) 
                      (ctyp*ctyp)list -> (cterm*cterm)list -> claset -> thm list 
                        -> term -> term -> term -> int 

PRE:                        -> tactic
       matchfun     : match-selector funtion; picks and classifies redex
       Tinsts Subst : pre-instantiations
       thm_sum      : if PPAT external rule t, then Inl t,
                      otherwise dummy data for the case that PPAT stems
                      from assumption
       PPAT         : "protected pattern", i.e.
                      equality under consideration
                      (if thm_sum = (Inl t) then Termrepr of conclusion of t)
       PAT          : pattern (e.g.: canonizable variable)        
       ST           : state transition term
       cla          : claset of context
       thmS         : list of subst-like thms 
       goal         : subgoal number
POST:  result value contains classification and substitution
       for a redex for an assumption (not equal to PPAT) 

 *) let val thy = theory_of_thm state;
        fun solve_asm goal = case thm_sum of
                               Inr (x:term,i:int) => atac goal
                             | Inl t => rtac t goal

        fun substasm1 Subst thm1 = EVERY[
                                   dres_terminst_tac TS S thm1 goal,
                                   solve_asm goal, fast_tac cla goal]
        fun substasm2 Subst thm1 = EVERY[
                                   dres_terminst_tac TS S thm1 goal,
                                   solve_asm goal, fast_tac cla goal,
                                   fast_tac cla goal]
        fun Subst2 (P,Pty)    = (Thm.cterm_of thy (Var(("P",0),Pty)),
                                 Thm.cterm_of thy P)::S
        fun Subst2'(P,P',Pty) = (Thm.cterm_of thy (Var(("P",0),Pty)),
                                 Thm.cterm_of thy P)::
                                (Thm.cterm_of thy (Var(("P'",0),Pty)),
                                 Thm.cterm_of thy P')::S  
        fun string_of_subst [] = ""
           |string_of_subst ((c1,c2)::S) = "("^(string_of_cterm c1)^","^
                                               (string_of_cterm c2)^"),"^
                                            string_of_subst(S)
        fun string_of_inst []  = ""
           |string_of_inst ((c1,c2)::S)  = "("^(string_of_ctyp c1)^","^
                                               (string_of_ctyp c2)^"),"^
                                            string_of_inst(S)

        fun s su   = "["^(string_of_subst su)^"]"
        fun t inst = "["^(string_of_inst inst)^"]"
        fun prin tag sbs = prin_tac (fn _=>"..subst_asm_gen:"^tag^":"^(t TS)^(s sbs))
        val prms   = (Logic.prems_of_goal (prop_of state) goal)
        
    in (* Still potential for speed-up here \<dots> thmS can be reduced to
          one element here !!! *)
       (case matchfun (PAT,PPAT) prms of
        Jdmt P      => (prin "Jdmt" (Subst2 P))       THEN TRY(substasm1 (Subst2  P) A)
       |JdmtEquiv P => (prin "JdmtEquiv" (Subst2' P)) THEN TRY(substasm2 (Subst2' P) B)
       |GlobCong P  => (prin "GlobCong" (Subst2' P))  THEN TRY(substasm2 (Subst2' P) C)
       |NoAsmRes    => (prin_tac (fn _ =>"..subst_asm_gen:NoAsmRes") THEN no_tac)) state
    end;   


val hypsubst_asm = subst_asm_gen find_asm_with_canon_free 
                                 (Inr(Bound 0,0)) (* dummy value signaling 
                                                     solve_asm atac. *)
                                                    

fun hypsubst_concl TS Subst cla thmS PPAT ST FV goal state =
    (* hypsubst_concl   : (ctyp * ctyp) list
                          -> (cterm * cterm) list
                          -> claset -> thm list -> term -> int -> tactic
     *)
    let  val thy = theory_of_thm state
         val concl   = Logic.concl_of_goal (prop_of state) goal
                      (* replaces parameters (Bounds) by frees *)
        
         fun body thm =      eres_terminst_tac TS Subst thm goal
                        THEN fast_tac cla (goal+1) 
                        THEN (if length(prems_of thm) = 4 
                              then fast_tac cla (goal+1)
                              else all_tac)
         val clean    = eres_terminst_tac 
                             [] 
                             [(Thm.cterm_of thy (Var(("V",0),propT)),
                               Thm.cterm_of thy (HOLogic.mk_Trueprop PPAT))] 
                             thin_rl goal;
    in  (if FV mem (term_frees concl) 
         then (prin_tac (fn _ => "..hypsubst_concl:A") THEN 
              (FIRST (map body thmS) ORELSE clean)) 
         else (prin_tac (fn _ => "..hypsubst_concl:B") THEN 
              clean))
         (state)
    end    


(* The tactic assumes that in all rules used for substitution,
   "X" is the name of the canonizeable variable, and "X" must
   have type "'a -> x" where thy TVar 'a is the type of the 
   context \<tau>. x can be a ground type or the typevariable 'b::bot. *)
val ATVar = (TVar(("'a",0),HOLogic.typeS));
val BTVar = (TVar(("'b",0),["Lifting.bot"]));
*}
ML{*
fun hypsubst clasimp goal state =
    let val prems   = Logic.prems_of_goal (prop_of state) goal
                      (* replaces parameters (Bounds) by frees *)
        val thy     = theory_of_thm state 
        val cla     = fst clasimp
        val res     = find_canon_free prems
        val (PPAT,ST,FV as Free(X,Xty)) = get_canon_free res
                      (*both may fail, caught to no_tac. see last line.*)
        val TSa     = [(Thm.ctyp_of thy ATVar, 
                        Thm.ctyp_of thy (domain_type Xty))]
        val TSb     = [(Thm.ctyp_of thy BTVar, 
                        Thm.ctyp_of thy (range_type Xty))]
        val Subst   = [(Thm.cterm_of thy (Var(("X",0),Xty)),
                        Thm.cterm_of thy (Free(X,Xty))),
                       (Thm.cterm_of thy (Var(("\<tau>",0),domain_type Xty)),
                        Thm.cterm_of thy ST)]
        fun hs tag ts1 ts2 r1 r2 =
                      prin_tac (fn _=> "hypsubst:"^tag^":"^X) 
                      THEN (TRY(REPEAT(hypsubst_asm ts1 Subst cla 
                                                    r1  PPAT ST FV goal)))
                      THEN hypsubst_concl ts2 Subst cla r2 PPAT ST FV goal
    in  (case res of 
        (UNDEF _)=> hs "UNDEF"(TSa@TSb) TSa undef_fw_rev_rules undef_rules
       |(DEF  _) => hs "DEF"  TSa       TSa DEF_fw_rev_rules DEF_rules
       |(FALSE _)=> hs "FALSE"TSa       TSa FALSE_fw_rev_rules FALSE_rules
       |(TRUE _) => hs "TRUE"  TSa      TSa TRUE_fw_rev_rules TRUE_rules
       |(EQLS _) => hs "EQLS" (TSa@TSb) TSa strongEq_fw2_rules strongEq_rules
       |(EQRS _) => hs "EQRS" (TSa@TSb) TSa strongEq_fw_rules strongEq_rev_rules
       |(EQLW _) => hs "EQLW" (TSa@TSb) TSa strictEq_fw2_rules strictEq_rules
       |(EQRW _) => hs "EQRW" (TSa@TSb) TSa strictEq_fw_rules strictEq_rev_rules
       | _       => all_tac) (state)
    end
    handle _ => no_tac state;


(* known BUG: \<tau> \<MathOclValid> \<MathOclIsDefined> E - does not terminate always, in particular if E, but
              not \<MathOclIsDefined> E occurs in the conclusions \<dots> Check for occurence of
              E too simple \<dots> *)

   
fun OCL_hyp_subst_tac clasimp goal = 
        CHANGED(TRY(hyp_subst_tac goal) THEN
                REPEAT(hypsubst clasimp goal))

*}

ML{* 

fun get_context_of_matches thy (pat,obj) =
(* pre: must not contain loose bound variables . . .
        must be typeable . . .
   post: type is drawn from obj entirely, thus the
         intrusion of type-vars from pat is avoided *)
  let fun msub(bounds,obj) = 
              if Pattern.matches thy (pat,obj) 
              then (SOME(Bound (length bounds), type_of obj))
              else (case obj of
                    Abs(x,T,t) => (let val y = variant bounds x
                                       val f = Free(":" ^ y,T)
                                   in  msub(x::bounds,subst_bound(f,t)) end)
                    | s$t => (case (msub(bounds,s), msub(bounds,t)) of
                                (SOME (s',ty), SOME(t',_)) => SOME(s' $ t',ty)
                               |(SOME (s',ty), NONE)    => SOME(s' $ t,ty)
                               |(NONE, SOME (t',ty))    => SOME(s $ t',ty)
                               |(_ , _)            => NONE)
                    | _   => NONE)
  in case msub([],obj) of 
(*       SOME t => SOME(Abs(": Intern",(type_of pat) --> (type_of obj),t)) *)
       SOME(t,ty) => SOME(Abs(": Intern",ty,t))
     | _ => NONE
  end;

fun get_trivial_match (pat,obj) =
(* pre: must not contain loose bound variables . . .
        must be typeable . . . *)
(*   Abs(": Intern",(type_of pat) --> (type_of obj),obj) *)
   Abs(": Intern",type_of pat,obj)

*}
ML{*
fun find_patt thy (pat,PPAT) [] = NoAsmRes
   |find_patt thy (pat,PPAT) (obj::R) =
    (* PPAT  : equality under consideration
       pat: obj                      *)
    (if   Pattern.matches thy (PPAT,HOLogic.dest_Trueprop obj)
     then find_patt thy (pat,PPAT) R
     else case  Pattern.beta_eta_contract(HOLogic.dest_Trueprop obj) of 
         (Const ("OCL_Logic_core.OclLocalValid",ty2) $ ST $ E) 
            => (case get_context_of_matches thy (pat,E) of
                SOME s'   =>  Jdmt(s', type_of s') 
              | NONE      =>  find_patt thy (pat,PPAT) R)
        |(Const ("op =",_) $ (Const ("OCL_Logic_core.OclLocalValid",ty2)  $ ST  $ E)
                           $ (Const ("OCL_Logic_core.OclLocalValid",ty2') $ ST' $ E') )
            => (case (get_context_of_matches thy (pat,E), 
                      get_context_of_matches thy (pat,E')) of
                (SOME s', SOME t') => JdmtEquiv(s',t',type_of s')
               |(SOME s', NONE)    => JdmtEquiv(s',get_trivial_match (pat,E'),type_of s')
               |(NONE, SOME t')    => JdmtEquiv(get_trivial_match (pat,E),t',type_of t')
               |(_ , _)            => find_patt thy (pat,PPAT) R)         
        |(Const ("op =",ty2) $ (E$ST) $ (E'$ST'))
            => (case (get_context_of_matches thy (pat,E), 
                      get_context_of_matches thy (pat,E')) of
                (SOME s', SOME t') => GlobCong(s',t',type_of s')
               |(SOME s', NONE)    => GlobCong(s',get_trivial_match (pat,E'),type_of s')
               |(NONE, SOME t')    => GlobCong(get_trivial_match (pat,E),t',type_of t')
               |(_ , _)            => find_patt thy (pat,PPAT) R)
        |_ =>  find_patt thy (pat,PPAT) R)
   handle TERM _ => find_patt thy (pat,PPAT) R (* dest_Trueprop *)
     

*}

ML{*

fun subst_concl TS Subst cla thm_sum thmS ST goal state =
    (* subst_concl : (ctyp * ctyp) list
                          -> (cterm * cterm) list
                          -> claset -> thm list -> term -> int -> tactic
     *)
    let  val thy = theory_of_thm state
         val solve_asm = case thm_sum of
                           Inr _ => atac goal
                         | Inl t => rtac t goal
         fun body thm  =      res_terminst_tac TS Subst thm goal
                         THEN solve_asm 
                         THEN fast_tac cla (goal+1) 
                         THEN (if length(prems_of thm) = 4 
                               then fast_tac cla (goal+1)
                               else all_tac)
    in  (prin_tac (fn _ => "subst_concl") THEN 
         FIRST (map body thmS)) (state)
    end; 
*}

ML{* 
fun subst_asm TS S cla thmS thm_sum PPAT ST PAT goal state = 
            subst_asm_gen (find_patt(theory_of_thm state)) thm_sum
                TS S cla thmS PPAT ST PAT goal state

(* thm_sum is the rule/assumption that represents the rewrite-rule. *)
fun subst_asm_concl (no_asm_simp,no_concl_simp) css thm_sum goal state = 
    let val prems   = Logic.prems_of_goal (prop_of state) goal
                      (* replaces parameters (Bounds) by frees *)
        val thy     = theory_of_thm state 
        val cla     = fst css
        val res     = case thm_sum of
                        Inl thm1   => find_rw_rule [concl_of thm1]
                      | Inr(trm,n) => find_rw_rule [trm]
        val prems'  = if no_asm_simp then []
                      else case thm_sum of
                             Inl thm1   => prems
                           | Inr(trm,n) => List.take(prems,n-1) @ List.drop(prems,n)
        val (PPAT,ST,lhs $ rhs) = get_canon_free res
                      (*both may fail, caught to error see last line.*)
        val Xty     = type_of lhs
        val TSa     = [(Thm.ctyp_of thy ATVar, 
                        Thm.ctyp_of thy (domain_type Xty))]
        val TSb     = [(Thm.ctyp_of thy BTVar, 
                        Thm.ctyp_of thy (range_type Xty))]
        val Subst   = [(Thm.cterm_of thy (Var(("X",0),Xty)),
                        Thm.cterm_of thy lhs),
                       (Thm.cterm_of thy (Var(("\<tau>",0),domain_type Xty)),
                        Thm.cterm_of thy ST)]
        val X       =  Sign.string_of_term thy lhs
        fun sa ts1 r1  = subst_asm ts1 Subst cla r1 thm_sum PPAT ST lhs goal
        fun sc r1   = if no_concl_simp then no_tac
                      else subst_concl TSa Subst cla thm_sum r1 ST goal
        fun prin tag= prin_tac (fn _=>"subst_asm_concl:"^tag^X)
    in  (case res of 
        (UNDEF _) => prin "UNDEF:" THEN
                     (sa (TSa@TSb) undef_fw_rev_rules ORELSE sc undef_rules)
       |(DEF  _)  => prin "DEF:"   THEN
                     (sa TSa DEF_fw_rev_rules    ORELSE sc DEF_rules)
       |(FALSE _) => prin "FALSE:" THEN
                     (sa TSa FALSE_fw_rev_rules  ORELSE sc FALSE_rules)
       |(TRUE  _) => prin "TRUE:"  THEN
                     (sa TSa TRUE_fw_rev_rules   ORELSE sc TRUE_rules)
       |(EQLS  _) => prin "EQLS:"  THEN
                     (sa (TSa@TSb) strongEq_fw2_rules ORELSE sc strongEq_rules)
       |(EQLW  _) => prin "EQLW:"  THEN
                     (sa (TSa@TSb) strictEq_fw2_rules ORELSE sc strictEq_rules)) (state)
    end
    handle _ => error "Illegal Rewrite Pattern";

*}

ML{*
fun OCL_subst_tac(no_asm_use, no_asm_simp, no_concl_simp) (simps) css goal state =
    let val prems   = Logic.prems_of_goal (prop_of state) goal
                      (* replaces parameters (Bounds) by frees *)
        val asms  = if no_asm_use then [] else prems ~~ (1 upto (length prems))
        fun H (UNDEF  _) = true
           |H (DEF  _)   = true
           |H (FALSE  _) = true
           |H (TRUE  _)  = true
           |H (EQLS  _)  = true
           |H (EQLW  _)  = true
           |H _          = false
        val asms  = filter (fn (x,y) => H (find_rw_rule [x])) asms
        val fl    = (no_asm_simp,no_concl_simp)
    in  CHANGED(FIRST(map (fn X => subst_asm_concl fl css (Inl X) goal)(simps))
                ORELSE 
                FIRST(map (fn X => subst_asm_concl fl css (Inr X) goal)(asms)))
               (state)
    end

*}
(*>*)

section{* Testing \tactic{OCL\_hyp\_subst\_tac} *}

lemma "\<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> \<MathOclNot>B; \<tau> \<MathOclValid> A \<MathOclAnd> B ; \<tau> \<MathOclValid> \<MathOclIsUndefined> B\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> B \<MathOclAnd> C"
by(tactic"OCL_hyp_subst_tac (get_local_clasimpset ctxt) 1", simp)

lemma "\<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> A \<MathOclAnd> B ; \<tau> \<MathOclValid> \<MathOclIsUndefined> B; (\<tau> \<MathOclValid> A) = (\<tau> \<MathOclValid> C)\<rbrakk> \<Longrightarrow> A \<tau> = B \<tau>"
by(tactic"OCL_hyp_subst_tac (get_local_clasimpset ctxt) 1", simp)
 
lemma "!!\<tau> A . \<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> A \<MathOclAnd> B ; \<tau> \<MathOclValid> \<MathOclIsUndefined> B; (\<tau> \<MathOclValid> A) = (\<tau> \<MathOclValid> C)\<rbrakk> \<Longrightarrow> A \<tau> = B \<tau>"
by(tactic"OCL_hyp_subst_tac (get_local_clasimpset ctxt) 1", simp)

lemma "\<lbrakk>  \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> D \<MathOclStrongEq> E ;  \<tau> \<MathOclValid> D \<MathOclStrongEq> 1\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> E \<MathOclStrongEq> 1"
by(tactic"OCL_hyp_subst_tac (get_local_clasimpset ctxt) 1", simp)

lemma "\<lbrakk>  \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> D \<MathOclStrongEq> E ;  \<tau> \<MathOclValid> 1 \<MathOclStrongEq> D\<rbrakk>  \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> E \<MathOclStrongEq> 1"
by(tactic"OCL_hyp_subst_tac (get_local_clasimpset ctxt) 1", simp)

lemma "\<lbrakk>  \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> E \<MathOclStrongEq> D ; \<tau> \<MathOclValid> 1 \<MathOclStrongEq> E ;  \<tau> \<MathOclValid> 1 \<MathOclStrongEq> D\<rbrakk>  \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> E \<MathOclStrongEq> 1"
by(tactic"OCL_hyp_subst_tac (get_local_clasimpset ctxt) 1", simp)

(*ML{* OCL_hypsubst_trace:=false *} *)
lemma "\<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> \<MathOclNot>B; \<tau> \<MathOclValid> 1 \<MathOclStrongEq> E;  A \<tau> = B \<tau>; P \<tau> = \<MathOclTrue> \<tau>;
         \<tau> \<MathOclValid> D \<MathOclStrictEq> 1; \<tau> \<MathOclValid> D\<MathOclStrictEq>E\<rbrakk> 
       \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> B"
apply(tactic"OCL_hyp_subst_tac (get_local_clasimpset ctxt) 1")
apply(simp)
done

lemma "\<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> \<MathOclNot>B; \<tau> \<MathOclValid> E\<MathOclStrongEq>1; \<tau> \<MathOclValid> 1\<MathOclStrongEq>E; 
         \<tau> \<MathOclValid> E\<MathOclStrictEq>1;\<tau> \<MathOclValid> 1\<MathOclStrictEq>E;\<tau> \<MathOclValid> \<MathOclIsUndefined> B\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> B"
by(tactic"OCL_hyp_subst_tac (get_local_clasimpset ctxt) 1",simp)


section{* Testing \tactic{OCL\_subst\_tac} *}
(*<*)
ML{* val no_asm_use   = false
     val no_asm_simp  = false
     val no_concl_simp= false *}
(* commented out due to sorry \<dots> 

lemma hack1 : " \<tau> \<MathOclValid> A \<MathOclAnd> E \<MathOclStrongEq> 1" sorry
lemma hack2 : " \<tau> \<MathOclValid> \<MathOclIsDefined>(A \<MathOclAnd> C) " sorry

lemma "\<lbrakk>  \<tau> \<MathOclValid> D \<MathOclStrongEq> E ;  \<tau> \<MathOclValid> 1 \<MathOclStrongEq> D\<rbrakk>  \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> E \<MathOclStrongEq> 1  \<MathOclAnd> D \<MathOclStrongEq> E "
apply(tactic"OCL_subst_tac(no_asm_use, no_asm_simp, false) 
                       [thm\"hack1\"] (get_local_clasimpset ctxt) 1")
apply(simp)
done



lemma "\<lbrakk>  \<tau> \<MathOclValid> \<MathOclNot> (A \<MathOclAnd> E \<MathOclStrongEq> 1); \<tau> \<MathOclValid> D \<MathOclStrongEq> E ;  \<tau> \<MathOclValid> 1 \<MathOclStrongEq> D\<rbrakk>  \<Longrightarrow> \<tau> \<MathOclValid> F \<MathOclStrongEq> E "
apply(tactic"OCL_subst_tac(false, false, false) 
            [thm\"hack1\"] (get_local_clasimpset ctxt) 1")
apply simp
done
*)

(*>*)

lemma isDefined_cong5 : "(\<MathOclIsDefined> X \<MathOclOr> \<MathOclIsUndefined> X) = \<MathOclTrue>"
apply (rule ext, case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)")

apply (tactic"OCL_subst_tac(false, false, false) 
                       [] (get_local_clasimpset ctxt) 1", simp_all)
apply (tactic"OCL_subst_tac(false, false, false) 
                       [] (get_local_clasimpset ctxt) 1", simp_all)
done


section{* Some more Absurdities *}

(* TODO: Check if really here ! *)

lemma not_isDefined_if_isUndefined: 
  "\<tau> \<MathOclValid> \<MathOclIsDefined> A \<Longrightarrow> (~(\<tau> \<MathOclValid> \<MathOclIsUndefined> A))"
  by auto


lemma non_tertium_datur_if_isDefined: 
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>A; \<tau> \<MathOclValid> A \<Longrightarrow> P; \<tau> \<MathOclValid>  \<MathOclNot> A \<Longrightarrow> P \<rbrakk> \<Longrightarrow>  P"
  by (cut_tac \<tau> = "\<tau>" and A = "A" in non_quatrium_datur, auto)

lemma isDefined_if_valid':  
  "\<tau> \<MathOclValid> A \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> A"
  by (rule classical, auto)
  

lemma isDefined_if_invalid: 
  "\<tau> \<MathOclValid> \<MathOclNot> A \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> A"
  by (rule classical, auto)

(* Are these rules necessary absurdities ? *)





(**************************************************************************)
section {*  Trichotomy Specialized for $\isasymMathOclIf$ *}
(**************************************************************************)

lemma if_trichotomy: 
  "\<lbrakk> \<tau> \<MathOclValid> P \<MathOclUndefined>; \<tau> \<MathOclValid> P Y; \<tau> \<MathOclValid> P Z; cp P \<rbrakk> \<Longrightarrow> 
   \<tau> \<MathOclValid> P (\<MathOclIf> X \<MathOclThen> Y \<MathOclElse> Z \<MathOclEndif>)"
  apply (simp add: OclLocalValid_def)
  apply (rule_tac P = "\<lambda>x. P (\<MathOclIf> x \<MathOclThen> Y \<MathOclElse> Z \<MathOclEndif>)" in cp_distinct_core)
  apply auto
  apply (rule_tac P = "P" in cp_compose2)
  apply auto
  done

lemma if_trichotomyE: 
  "\<lbrakk> \<tau> \<MathOclValid> P (\<MathOclIf> X \<MathOclThen> Y \<MathOclElse> Z \<MathOclEndif>);
     \<tau> \<MathOclValid> P \<MathOclUndefined>  \<Longrightarrow> R;                   
     \<tau> \<MathOclValid> P Y  \<Longrightarrow> R;                   
     \<tau> \<MathOclValid> P Z  \<Longrightarrow> R;                   
     cp P \<rbrakk>                          
  \<Longrightarrow> R"
  apply (cut_tac \<tau> = "\<tau>" and A = "X" in non_quatrium_datur)
  apply safe

  apply (drule_tac [3] subst_LJ_undef_fw) 
  prefer 3
  apply (assumption)
  apply (rule_tac [3] P = "P" in cp_compose2, simp_all) 

  apply (drule_tac [2] subst_LJ_FALSE_fw) 
  prefer 2
  apply (assumption)
  apply (drule subst_LJ_TRUE_fw) 
  back
  apply  (assumption)
  apply auto
  apply (rule_tac [2] P = "P" in cp_compose2)
  apply (rule_tac P = "P" in cp_compose2)
  apply auto
  done


lemmas OclIf_LJE[simp] = 
       subst_TRUE_LJE [of _ _ "\<lambda>X. \<MathOclIf> X 
                                   \<MathOclThen> Y 
                                   \<MathOclElse> Z \<MathOclEndif>", 
                       simplified]
       subst_FALSE_LJE[of _ _ "\<lambda>X. \<MathOclIf> X 
                                   \<MathOclThen> Y 
                                   \<MathOclElse> Z \<MathOclEndif>", 
                       simplified]
       subst_undef_LJE[of _ _ "\<lambda>X. \<MathOclIf> X 
                                   \<MathOclThen> Y 
                                   \<MathOclElse> Z \<MathOclEndif>",
                       simplified]


(**************************************************************************)
section{* A Weak (Propositional) \acs{lje} Calculus for Propositional \OCL *}
(**************************************************************************)

text{* 
  The weak propositional \acs{lje} gives rise to a fairly promising
  decision procedure for a fragment of \OCL:
   \begin{itemize}
   \item blow away quantifiers
   \item infer from inclusions of variables the invariants
   \item infer from the invariants the definedness of variables
   \item all remaining variables occuring in formula: 
     case-splitting over definedness,
     congruence closure over undefinednesses,
   \item \tactic{simp\_all}: universal congruences
   \item \tactic{simp\_all}: weak propositional \acs{lje}
%   [case splitting over C's in all if ??? ]
   \item \tactic{blast\_tac}.
   \end{itemize}
    \emph{Weak} \acs{lje} Calculus means here that everything is restricted
       to defined expressions. 
*}

lemma localValidNot2not : 
   "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> A \<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> \<MathOclNot> A) = (\<not> (\<tau> \<MathOclValid> A))"
   by(auto elim!: non_tertium_datur_if_isDefined)


lemma localValidAnd2conj : 
   "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> A; \<tau> \<MathOclValid>  \<MathOclIsDefined> B \<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> A \<MathOclAnd> B) = ((\<tau> \<MathOclValid> A) \<and> (\<tau> \<MathOclValid> B))"
   apply(auto elim!: non_tertium_datur_if_isDefined)
   by(tactic"ALLGOALS(OCL_hyp_subst_tac (get_local_clasimpset ctxt))",simp_all)


lemma localValidOr2disj : 
   "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> A; \<tau> \<MathOclValid>  \<MathOclIsDefined> B \<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> A \<MathOclOr> B) = ((\<tau> \<MathOclValid> A) \<or> (\<tau> \<MathOclValid> B))"
   apply(auto elim!: non_tertium_datur_if_isDefined)
   by(tactic"ALLGOALS(OCL_hyp_subst_tac (get_local_clasimpset ctxt))",simp_all)

lemma localValidImplies2impl : 
   "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> A; \<tau> \<MathOclValid>  \<MathOclIsDefined> B \<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> A \<MathOclImplies> B) = ((\<tau> \<MathOclValid> A) \<longrightarrow> (\<tau> \<MathOclValid> B))"
   by(simp add: OclImplies_def localValidNot2not 
                localValidOr2disj localValidAnd2conj) 


lemma isDefined_notI: "\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsDefined> A\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> (\<MathOclNot> A)" by simp

lemma isDefined_andI: "\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsDefined> A; \<tau> \<MathOclValid> \<MathOclIsDefined> B\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> (A \<MathOclAnd> B)"
  apply(auto elim!: non_tertium_datur_if_isDefined)
  apply(tactic"ALLGOALS(OCL_hyp_subst_tac (get_local_clasimpset ctxt))")
  apply(simp_all)
  done

lemma isDefined_orI:"\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsDefined> A; \<tau> \<MathOclValid> \<MathOclIsDefined> B\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> (A \<MathOclOr> B)"
  apply(auto elim!: non_tertium_datur_if_isDefined)
  apply(tactic"ALLGOALS(OCL_hyp_subst_tac (get_local_clasimpset ctxt))")
  apply(simp_all)
  done

lemma isDefined_xorI:"\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsDefined> A; \<tau> \<MathOclValid> \<MathOclIsDefined> B\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> (A \<MathOclXor> B)"
  apply(auto elim!: non_tertium_datur_if_isDefined)
  apply(tactic"ALLGOALS(OCL_hyp_subst_tac (get_local_clasimpset ctxt))")
  apply(simp_all)
  done

lemma isDefined_impliesI:"\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsDefined> A; \<tau> \<MathOclValid> \<MathOclIsDefined> B\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> (A \<MathOclImplies> B)"
  apply(auto elim!: non_tertium_datur_if_isDefined)
  apply(tactic"ALLGOALS(OCL_hyp_subst_tac (get_local_clasimpset ctxt))")
  apply(simp_all)
  done

lemma isDefined_implies2I:"\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsDefined> A; \<tau> \<MathOclValid> \<MathOclIsDefined> B\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> (A \<MathOclVVImplies> B)"
  apply(auto elim!: non_tertium_datur_if_isDefined)
  apply(tactic"ALLGOALS(OCL_hyp_subst_tac (get_local_clasimpset ctxt))")
  apply(simp_all)
  done

lemma localValidImplies22impl : 
   "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> A; \<tau> \<MathOclValid>  \<MathOclIsDefined> B \<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> A  \<MathOclVVImplies> B) = ((\<tau> \<MathOclValid> A) \<longrightarrow> (\<tau> \<MathOclValid> B))"
  by(simp add: OclImplies2_def localValidNot2not isDefined_andI 
                  localValidOr2disj localValidAnd2conj) 

(*<*)
ML{* OCL_strict_syntax_checking:=false; *}
(*>*)

section {* \acs{isar} Interface Setup *}

text{* 
  This part of the so-called thymorpher generates 
  a number of rules and a specific setup for all operators
  given to him. 

  This includes (all examples are for the case of
  the weak (strict) equality here.
  \begin{enumerate}
  \item context passingness rules (cp) for all unary and 
        binary operators, e.g.:
        \[ \lbrakk cp P; cp P'\rbrakk \Longrightarrow cp (\lambda X. P X \isasymMathOclStrictEq P' X)\]
        These rules were used both as simplifier rules
        as well as safe introduction rules.
 
  \item all undefinedness reduction rules
        (unary and binary case), e.g.:
        \[ \isasymMathOclUndefined \isasymMathOclStrictEq X = \isasymMathOclUndefined" and "X \isasymMathOclStrictEq \isasymMathOclUndefined = \isasymMathOclUndefined \]
        These rules were used as global simplifier rules.
  \item universal congruences for definedness, e.g.:
        \[\isasymMathOclIsDefined X \isasymMathOclStrictEq Y = \isasymMathOclIsDefined X  \isasymMathOclAnd \isasymMathOclIsDefined Y \]
  \item the local judgement equivalence for definedness, e.g.:
        \[\tau \isasymMathOclValid \isasymMathOclIsDefined(X \isasymMathOclStrictEq Y) = (\tau \isasymMathOclValid \isasymMathOclIsDefined X \land \tau \isasymMathOclValid \isasymMathOclIsDefined Y)\]
        This equality is used for optimized reasoning
        over definedness inside some OCL-tactics.
  \end{enumerate} 
   *}

(*<*)
ML{*
(*
val OCL_initialized_ops = ref ([""]);
val OCL_is_def_optimized_ss = ref HOL_ss ;

fun mRS S (name,thm) = Library.mapfilter (fn a => SOME(name,thm RS a) 
                                          handle THM _ => NONE)
                                         (S);

val cp_lift1_fw                  = thm"cp_lift1_fw";
val cp_lift2_fw                  = thm"cp_lift2_fw";

val lift2_undef1a_fw             = thm"lift2_undef1a_fw";
val lift2_undef2a_fw             = thm"lift2_undef2a_fw";
val lift1b_undef_fw              = thm"lift1b_undef_fw";
val lift1_strict_is_isdef_fw     = thm"lift1_strict_is_isdef_fw";
val lift1_strictify_is_isdef_fw  = thm"lift1_strictify_is_isdef_fw";
val lift2_strict_is_isdef_fw     = thm"lift2_strict_is_isdef_fw";
val lift2_strictify_is_isdef_fw  = thm"lift2_strictify_is_isdef_fw";
val lift1_strictify_implies_LocalValid_defined =
                         thm"lift1_strictify_implies_LocalValid_defined";
val lift2_strictify_implies_LocalValid_defined =
                         thm"lift2_strictify_implies_LocalValid_defined";
*)
*}

ML{*
   (* OUTDATED !!! *)
   fun name_thm p  (name,thm) = 
       let val clean_dot = String.map (fn x => if x = #"." then #"_" else x)
       in (clean_dot((p())^name),thm) end

*}


ML {*

fun undef_rules1 rname = RULE
    {rule      = (rname, thm rname),
     simp_tac  = fn _ => fn x => x,
     is_atomic = fn _ => true, 
     register  = default_register_simponly, 
     name_gen  = default_namegen "OCL_undef_1_"
    }

fun undef_rules2 rname = RULE
    {rule      = (rname, thm rname),
     simp_tac  = fn _ => fn x => x,
     is_atomic = fn _ => true, 
     register  = default_register_simponly, 
     name_gen  = default_namegen "OCL_undef_2_"
    }

fun is_defs_optimized_rules rname = RULE
    {rule      = (rname, thm rname),
     simp_tac  = fn (cs,ss) => simplify ss,
     is_atomic = fn _ => true, 
     register  = default_register_simponly, 
     name_gen  = default_namegen "OCL_is_defopt_"
    }

fun is_defs_rules rname = RULE
    {rule      = (rname, thm rname),
     simp_tac  = fn (cs,ss) => simplify ss,
     is_atomic = fn _ => true, 
     register  = default_register_simponly, 
     name_gen  = default_namegen "OCL_is_def_"
    };

add_setup_rules (
   map undef_rules1 ["lift2_undef1a_fw","lift1b_undef_fw"] @
   map undef_rules2 ["lift2_undef2a_fw"] @
   map is_defs_optimized_rules 
       ["lift1_strictify_implies_LocalValid_defined",
        "lift2_strictify_implies_LocalValid_defined"] @
   map is_defs_rules 
       [(* "lift1_strict_is_isdef_fw", outdated ? *)
        "lift1_strictify_is_isdef_fw",
        (* "lift2_strict_is_isdef_fw", outdated ? *)
        "lift2_strictify_is_isdef_fw"]
   );



*}

(*<*)
subsection{* Definition of \tactic{ocl\_hypsubst} *}
text {*
  In the theory files, this section contains the \acs{isar}  setup
  code for  \tactic{ocl\_hypsubst}. 
*}

(*>*)
method_setup ocl_hypsubst = {* 
let
  fun m_ocl_hypsubst args proof =
  let
 
      val thy             = ProofContext.theory_of proof;

     fun clasimp_meth tac prems ctxt = Method.METHOD (fn facts =>
                                         ALLGOALS (Method.insert_tac (prems @ facts))
                                                THEN tac (get_local_clasimpset ctxt));
     fun clasimp_meth' tac prems ctxt = Method.METHOD (fn facts =>
                                         HEADGOAL (Method.insert_tac (prems @ facts)
                                                THEN' tac (get_local_clasimpset ctxt)));
     val clasimp_method' = Method.bang_sectioned_args clasimp_modifiers o clasimp_meth';
    fun   ocl_hypsubstP m  = Method.bang_sectioned_args' clasimp_modifiers 
                             (Scan.lift (Scan.option (Args.$$$ "[" |-- Args.nat --| Args.$$$ "]")))  m;

    (* WHY thm? *)
    fun ocl_hypsubst_meth NONE = clasimp_meth (CHANGED_PROP o ( fn css => OCL_hyp_subst_tac css 1))
      | ocl_hypsubst_meth (SOME n) = clasimp_meth (CHANGED_PROP o 
                 (fn css => OCL_hyp_subst_tac css n ))
  in
      ocl_hypsubstP ocl_hypsubst_meth  args proof
  end

in
   m_ocl_hypsubst
end

*} "OCL hypsubst"

(*>*)

subsection{* Definition of \tactic{ocl\_subst} *}
text {*
  In the theory files, this section contains the \acs{isar}  setup
  code for  \tactic{ocl\_subst}. 
*}

(*<*)
method_setup ocl_subst = {* 
let
  fun m_ocl_subst args proof =
  let
     val thy             = ProofContext.theory_of proof
     fun clasimp_meth tac prems ctxt = Method.METHOD (fn facts =>
                                         ALLGOALS (Method.insert_tac (prems @ facts))
                                                THEN tac (get_local_clasimpset ctxt))

    fun   ocl_substP m  =  Method.bang_sectioned_args' clasimp_modifiers  (Scan.lift (
	                     Scan.option (Args.$$$ "[" |-- Args.nat --| Args.$$$ "]")
                          -- Scan.option (Args.parens (Args.$$$ "no_asm_use"))
                          -- Scan.option (Args.parens (Args.$$$ "no_asm_simp"))
	                  -- Scan.option (Args.parens (Args.$$$ "no_concl_simp"))
                          -- Scan.option (Args.$$$ "[" -- Args.$$$ "add:" 
                                  |-- Args.list1 Args.name --| Args.$$$ "]"))
                          ) m

  fun ocl_subst_meth ((((goalNr, no_asm_use), no_asm_simp),no_concl_simp), thmList) 
                 =   let
                        val theGoal        = (case goalNr  of
                                                SOME g => g
                                              | NONE   => 1)
                        val theNoConclSimp = isSome no_concl_simp
                        val theNoAsmUse    = isSome no_asm_use 
                        val theNoAsmSimp   = isSome no_asm_simp 
                        val theThmList     = (case thmList  of
                                                SOME l => map (fn t => 
                                                                get_thm thy (Name t)) l
                                              | NONE   => [])
                     in 
                        clasimp_meth (CHANGED_PROP o 
                              (fn css => OCL_subst_tac 
                                             (theNoAsmUse, theNoAsmSimp, theNoConclSimp) 
                                              theThmList css theGoal )) 
                      end
  in
      ocl_substP ocl_subst_meth  args proof
  end

in
   m_ocl_subst
end

*} "OCL subst"

(*>*)
section{* Definition of \tactic{ocl\_auto} *}
text {*
  In the theory files, this section contains the \acs{isar}  setup
  code for  \tactic{ocl\_auto}. 
*}

(*<*)

method_setup ocl_auto = {* 
let
  fun m_ocl_auto args proof =
  let
     val ocl_auto_tac  = auto_tac ; 

     val thy             = ProofContext.theory_of proof;

     fun clasimp_meth tac prems ctxt = Method.METHOD (fn facts =>
                                         ALLGOALS (Method.insert_tac (prems @ facts))
                                                THEN tac (get_local_clasimpset ctxt));
     fun clasimp_meth' tac prems ctxt = Method.METHOD (fn facts =>
                                         HEADGOAL (Method.insert_tac (prems @ facts)
                                                THEN' tac (get_local_clasimpset ctxt)));
     val clasimp_method' = Method.bang_sectioned_args clasimp_modifiers o clasimp_meth';
    fun   ocl_autoP m  = Method.bang_sectioned_args' clasimp_modifiers 
                             (Scan.lift (Scan.option (Args.nat -- Args.nat)))  m;

    fun ocl_auto_meth NONE = clasimp_meth (CHANGED_PROP o ocl_auto_tac)
      | ocl_auto_meth (SOME (m, n)) = clasimp_meth (CHANGED_PROP o 
                 (*(fn css => mk_ocl_auto_tac css m n) *)
                 ocl_auto_tac);
  in
      ocl_autoP ocl_auto_meth  args proof
  end

in
   m_ocl_auto
end

*} "OCL auto"

(*>*)
end
