(* ****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Set.thy -
 * Copyright (C) 2001,2002,2003,2004 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNES 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ***************************************************************************** *)

header {* \OCL Unsmashed Set *}
theory OCL_Set 
imports 
  "$HOLOCL_HOME/src/library/collection/OCL_Collection" 
  "$HOLOCL_HOME/src/library/collection/nonsmashed/OCL_Bag"
  "$HOLOCL_HOME/src/library/collection/nonsmashed/OCL_OrderedSet"


begin   
 text_raw{* \label{thy:OCL_Set} *}


section{* Is \OCL Set Faithful to OCL 2.0? *}

text{* Unlike the basic library theories, for this first and
  foundational ``Collection''-type theory the question of faithfulness
  is not merely a syntactic matter (like fusing various semantic
  interpretation functions into one). In fact, it is not even
  trivial to answer. This is mainly for two reasons:
  \begin{enumerate}
  \item We generalize the concept of Set as required by OCL 2.0 to
        \textbf{infinite sets}. 
  \item We interprete the standard with respect to the question,
        if \verb+OclUndefind+ can be included in a set or not:
        we deny this question in this version of a set theory. 
  \end{enumerate}
The first decision has the advantage, that general Sets
can be used to represent the syntactic category of "types"
in the standard as ``characteristic sets'' in HOL and 
enables the possibility to reason over them. From the pragmatics point 
of view, this also allows for quantifications of sets of ``values'' in 
the sense of OCL 2.0. For example, it is possible in \holocl{} to
express the commutativity law on Integers \emph{inside} OCL
as follows:
\begin{verbatim}
  
  Integer ->forAll(x,y | x + y `=` y + x)

\end{verbatim}
The second decision is due to the fact that the standard avoids
a clear statement to the nature of undefinedness in OCL. This
led to two forms of interpretations by users and tool-developers:
is \verb+OclUndefined+ just ``null'', so the null-reference,
or is it a semantic construct denoting non-termination or
exceptional behaviour? In the former sense, it is possible
to include \verb+OclUndefined+ in a set, in the latter not.
Admitting undefinedness in sets also has consequenses in the
inclusion operation and the elementhood test (\verb+->including+ and
\verb+->includes+). The operational behaviour of JAVA-implementations
reflects the latter view, which we chose therefore as default.
Semantically, this leads to a quotient construction (called ``smashing''
in the literature) which identifies $\{x, \bottom\}$ with $\bottom$.

With respect to the question of faithfulness of our OCL semantics
we have the following answers:
\begin{enumerate}
\item the required \emph{properties} of OCL 2.0 should hold in 
      our setting for finite sets, and
\item the subtle issue of $\bottom$s in sets is not mentioned
      at all in them.
\end{enumerate}
Therefore, we show compliance in this sense by providing the individual
proof for these ``required properties''. However, we will only use
them in rare cases in definitions and provide a number of other
properties for them that are more suited for automated reasoning.

*}
        

section{* Foundational Properties of the smash-construction *}


text{* Global OCL type constructor for OCL Set types: *}
types    ('st,'b) Set  = "('st, 'b Set_0) VAL"

section{* Properties of the Congruence inside the Datatype Adaption *}

text{* The two Abs\_Sequence\_inject\_absurd rules proved very useful
       therefore I'll introduce them as well for sets. *}

lemma Abs_Set_inject_absurd11 [simp]:
  "\<lbrakk> \<bottom> \<notin> x \<rbrakk> \<Longrightarrow> ((Abs_Set_0 down) = (Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(subst Abs_Set_0_inject,
     simp_all add: OCL_Set_type.Set_0_def smash_def)

lemma Abs_Set_inject_absurd12 [simp]:
  "\<lbrakk> \<bottom> \<notin> x \<rbrakk> \<Longrightarrow> ((Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Set_0 down)) = False"
  by(subst Abs_Set_0_inject,
     simp_all add: OCL_Set_type.Set_0_def smash_def)

lemma Abs_Set_inject_absurd21 [simp]:
  "\<lbrakk> \<bottom> \<notin> x \<rbrakk> \<Longrightarrow> (\<bottom> = (Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(simp add: UU_Set_def)

lemma Abs_Set_inject_absurd22 [simp]:
  "\<lbrakk> \<bottom> \<notin> x \<rbrakk> \<Longrightarrow> ((Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<bottom>) = False"
  by(simp add: UU_Set_def)

lemma Abs_Set_inject_charn:
  "\<lbrakk> (\<bottom>::'a::bot) \<notin> x; (\<bottom>::'a::bot) \<notin> y \<rbrakk> \<Longrightarrow> 
   (((Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Set_0 \<HolOclLiftLeft>y\<HolOclLiftRight>)) = (x = y))"
  by(subst Abs_Set_0_inject, 
     simp_all add: OCL_Set_type.Set_0_def smash_def)

lemma Abs_Set_inverse_charn1 [simp]:
  "(\<bottom> \<notin> x) \<Longrightarrow> Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<HolOclLiftLeft>x\<HolOclLiftRight>"
  by (subst Abs_Set_0_inverse, 
      simp_all add: OCL_Set_type.Set_0_def smash_def)

lemma Abs_Set_inverse_charn2 [simp]:
  "(\<bottom> \<notin> x) \<Longrightarrow> \<HolOclDropLeft>Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)\<HolOclDropRight> = x"
  by (subst Abs_Set_0_inverse, 
      simp_all add: OCL_Set_type.Set_0_def smash_def)


text{* the rest *}

lemma inj_on_Abs_Set: "inj_on Abs_Set_0 OCL_Set_type.Set_0";
   by(rule inj_on_inverseI,rule Set_0.Abs_Set_0_inverse,assumption)

lemma inj_Rep_Set : "inj Rep_Set_0"
   by(rule inj_on_inverseI, rule Set_0.Rep_Set_0_inverse)
(*
lemma smashed_set_charn: "(\<bottom> \<notin> X) = ((\<HolOclLiftLeft>X\<HolOclLiftRight>) \<in> OCL_Set_type.Set_0)"
   apply(unfold smash_def OCL_Set_type.Set_0_def UU_Set_def,auto) (*characterization 1 *)
*)
lemma UU_in_smashed_set[simp]: "\<bottom> : OCL_Set_type.Set_0"
   by(unfold smash_def OCL_Set_type.Set_0_def UU_Set_def,auto) (*characterization 2a *)

lemma down_in_smashed_set[simp]: "down : OCL_Set_type.Set_0"
   by(unfold smash_def OCL_Set_type.Set_0_def UU_Set_def,auto) (*characterization 2b *)

lemma mt_in_smashed_set: "\<HolOclLiftLeft>{}\<HolOclLiftRight> \<in> OCL_Set_type.Set_0"
   by (unfold smash_def OCL_Set_type.Set_0_def,auto)

(*
text{* Note that universal sets are never in \verb+OCL_Set_type.Set_0+,
   since they must contain bottom elements (HOL's type discipline 
   forces \verb+UNIV+ to be of type \verb+'a::bot set+) and therefore
   their smashed value is $\bottom$. *}
lemma UNIV_notin_smashed_set: "\<HolOclLiftLeft>UNIV\<HolOclLiftRight> \<notin> OCL_Set_type.Set_0"
   by (simp add: smash_def OCL_Set_type.Set_0_def)
*)
text{* In contrast, the set consisting of all 1-lifted elements 
   is in \verb+OCL_Set_type.Set_0+. This justifies the construction
   of characteristic sets for Boolean, Integer, Real, and String. *}
lemma liftUNIV_in_smashed_set: "\<HolOclLiftLeft>lift ` UNIV\<HolOclLiftRight> \<in> OCL_Set_type.Set_0"
   by (auto simp: smash_def OCL_Set_type.Set_0_def)



lemma DEF_Abs_Set: "\<bottom> \<notin> X \<Longrightarrow> DEF (Abs_Set_0 (\<HolOclLiftLeft>X\<HolOclLiftRight>))"
  by(simp add: DEF_def)

lemma DEF_Rep_Set: "DEF X \<Longrightarrow> DEF (Rep_Set_0 X)"
apply(unfold DEF_def UU_Set_def,auto)
apply(drule_tac f = "Abs_Set_0" in arg_cong)
apply(simp add: Rep_Set_0_inverse)
done

lemma not_DEF_Rep_Set: "\<not> DEF X \<Longrightarrow> \<not> DEF (Rep_Set_0 X)"
apply(unfold DEF_def UU_Set_def,auto)
apply(simp add: Abs_Set_0_inverse)
done


lemma exists_lift_Set: "DEF X \<Longrightarrow> \<exists>c. Rep_Set_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
by (drule DEF_Rep_Set,simp add: DEF_X_up)
(*
lemma exists_lift_Set2: "DEF X \<Longrightarrow> \<exists>c. \<bottom> \<notin> c \<and> Rep_Set_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
apply(frule exists_lift_Set)
apply(auto simp: smashed_set_charn)
apply(drule sym)
apply(simp add: Rep_Set_0)
done

lemma Rep_Set_cases: 
"Rep_Set_0 X = \<bottom> \<or> (\<exists>c. \<bottom> \<notin> c \<and> Rep_Set_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
apply(case_tac "DEF X")
apply (drule exists_lift_Set2)
apply (drule_tac [2] not_DEF_Rep_Set)
apply(simp_all add: DEF_def)
done

lemma DEF_X_Set_0 : "DEF X = (\<exists>c. \<bottom> \<notin> c \<and> Rep_Set_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
apply (insert  Rep_Set_cases [of "X"], auto simp: smashed_set_charn)
apply (drule exists_lift_Set,auto)
apply (rule swap)
prefer 2
apply (rule not_DEF_Rep_Set, auto)
done


text{* These four lemmas are used by ocl\_setup\_op to reason about definedness: *}

lemma lift2_strict_is_isdef_fw_Set_Val:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Set_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Set_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> a; (\<bottom>::'a::bot) \<noteq> b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> (g a b))"
  shows          " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def lift0_def lift1_def lift2_def
                  strictify_def DEF_def)
  apply(rule ext)
  apply(case_tac "DEF (X St)")
  apply(frule DEF_X_Set_0[THEN iffD1])
  apply(auto simp: strictify_def DEF_def)
  apply(drule_tac b="(Y St)" in  inv_g)
  prefer 2
  apply(drule DEF_Abs_Set, simp add: DEF_def, auto)
  done

lemma lift2_strict_is_isdef_fw_Set_Set:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Set_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Set_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_Set_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> a; \<bottom> \<notin> b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> (g a b))"
  shows          " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def lift0_def lift1_def lift2_def
                  strictify_def DEF_def)
  apply(rule ext)
  apply(case_tac "DEF (X St)")
  apply(case_tac "DEF (Y St)")
  apply(drule DEF_X_Set_0[THEN iffD1])+
  apply(auto simp: strictify_def DEF_def)
  apply(drule_tac b="ca" in inv_g, assumption)
  apply(rotate_tac -1)
  apply(drule DEF_Abs_Set, simp add: DEF_def)
  done

lemma lift2_strictify_implies_LocalValid_defined_Set_Val:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Set_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Set_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> a; (\<bottom>::'a::bot) \<noteq> b\<rbrakk> \<Longrightarrow> \<bottom> \<notin> (g a b))"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_Set_Val)
  apply(rule inv_g, assumption+)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done

lemma lift2_strictify_implies_LocalValid_defined_Set_Set:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Set_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Set_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_Set_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> a; \<bottom> \<notin> b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> (g a b))"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_Set_Set)
  apply(rule inv_g, assumption+)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done


text{* The following lemma is crucial for the thy\_morpher: *} 
lemma DEF_X_Set: "DEF X = (\<exists>c. \<bottom> \<notin> c \<and> X = Abs_Set_0 (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
apply (auto simp: DEF_Abs_Set smashed_set_charn)
apply (simp add:DEF_X_Set_0)
apply (erule exE, rule exI, auto simp: smashed_set_charn)
apply (rule injD [OF inj_Rep_Set])
apply (auto simp: Abs_Set_0_inverse)
done 


lemma [simp]: "\<bottom> \<notin> ca \<Longrightarrow> Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>ca\<HolOclLiftRight>) = \<HolOclLiftLeft>ca\<HolOclLiftRight>"
apply (rule  Abs_Set_0_inverse)
apply (simp add: smash_def OCL_Set_type.Set_0_def)
done
 

lemma RepAbs_UNIV[simp]:
"Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>lift ` UNIV\<HolOclLiftRight>) = \<HolOclLiftLeft>lift ` UNIV\<HolOclLiftRight>"
apply (subst Abs_Set_0_inverse)
apply (simp_all add: Abs_Set_0_inverse liftUNIV_in_smashed_set)
done



lemma Abs_Set_rangelift_DEF[simp]:
"DEF(Abs_Set_0 \<HolOclLiftLeft>range lift\<HolOclLiftRight>)"
by(auto simp: DEF_X_Set_0)

*)

section{* The Definition of the Set Operators *}

subsection{* The Powerset Operator *}

constdefs
   Set     :: "('st, ('b::bot) Set_0) VAL \<Rightarrow> ('st, 'b Set_0 Set_0) VAL"
  "Set     \<equiv> lift1(strictify (\<lambda>X. 
                     Abs_Set_0\<HolOclLiftLeft>(Abs_Set_0 o lift) ` {x. x \<subseteq> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>}\<HolOclLiftRight>))"

ML{* OCL_setup_trace := true; *}

ocl_setup_op [Set]
(*
lemma Set_defined :
" \<tau> \<MathOclValid> \<MathOclIsDefined> X \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsDefined>(Set X)"
apply(simp add: Set_def localValidDefined2sem)
apply (auto simp: strictify_def 
                 o_def lift0_def lift1_def lift2_def image_def
                 OclTrue_def OclFalse_def  OclUndefined_def) 
apply(rule DEF_Abs_Set, simp)
apply(rule allI, rule impI)
apply(rule neq_commute[THEN iffD1])
apply(simp only:DEF_def[symmetric])
apply(rule DEF_Abs_Set, simp)
thm exists_lift_Set2
apply(drule exists_lift_Set2, auto)
done

lemma Set_definedR :
"(\<tau> \<MathOclValid> \<MathOclIsDefined>(Set X)) =  (\<tau> \<MathOclValid> \<MathOclIsDefined> X) "
apply (auto simp: Set_defined )
oops
*)
subsection{* The Empty Set *}

constdefs  
    OclMtSet  :: "('a, 'b::bot Set_0) VAL" 
   "OclMtSet  \<equiv> lift0(Abs_Set_0(\<HolOclLiftLeft>{}\<HolOclLiftRight>))"


    syntax  (* standard syntax *)
    "_OclMtSet_std"   :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("\<OclMtSet>")
    syntax  (* ascii syntax *)
    "_OclMtSet_ascii" :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("mtSet")
    syntax  (xsymbols)
    "_OclMtSet_math"  :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("\<MathOclMtSet>")
    parse_translation {* flat(map (emb trans_const) ["OclMtSet"]) *}
    print_translation {* map emb_print ["OclMtSet"] *}



subsection{* Foundational Properties of MtSet *}

(*
lemma Abs_Set_Mt_DEF[simp]:
"DEF(Abs_Set_0 \<HolOclLiftLeft>{}\<HolOclLiftRight>)"
by(simp add: DEF_X_Set_0)
 

lemma mtSet_Defined [simp]:
"\<MathOclIsDefined> \<MathOclMtSet> = \<MathOclTrue>"
by(simp add: localValidDefined2sem Abs_Set_Mt_DEF OclMtSet_def 
             OclIsDefined_def lift1_def lift0_def OclTrue_def)

*)


section{* Foundational Properties of includes and excludes *}

defs 
   (* includes     :: "[('a, 'b Set) VAL,('a, 'b::bot) VAL] => 'a Boolean" *)
   OclIncludes_def :  "OclIncludes \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>x. 
                                           \<HolOclLiftLeft>x \<in> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>)))"


   (* excludes     :: "[('a, 'b Set) VAL, ('a, 'b::bot) VAL] => 'a Boolean" *) 
   OclExcludes_def : "OclExcludes \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>x.
                                           \<HolOclLiftLeft>x \<notin> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>)))" 

ocl_setup_op [OclIncludes, OclExcludes]


lemma includes_charn1 : "(\<tau> \<MathOclValid> x \<MathOclIncludes> S) \<Longrightarrow> (x \<tau> \<in> \<HolOclDropLeft>Rep_Set_0 (S \<tau>)\<HolOclDropRight>)" 
apply (frule isDefined_if_valid,
       drule OCL_is_defopt_OclIncludes[THEN iffD1])
apply (simp add: OclLocalValid_def OclIsDefined_def strictify_def
                 OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclFalse_def  OclUndefined_def DEF_def) 
done

(*
lemma includes_charn2 : "\<lbrakk> DEF(S \<tau>); x \<tau> \<in> \<HolOclDropLeft>Rep_Set_0 (S \<tau>)\<HolOclDropRight> \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> x \<MathOclIncludes> S"
apply (frule DEF_X_Set [THEN iffD1], erule exE)
apply (simp add: OclLocalValid_def OclIsDefined_def strictify_def
                 OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclFalse_def  OclUndefined_def DEF_def)
apply auto
done
*)

section{* Higher Properties of includes and excludes *}

(* We prefer a normal form where excludes is eliminated. *)
lemma exludes_not_includes[simp]: 
"x \<MathOclExcludes> X = \<MathOclNot> (x::('a, 'b)VAL) \<MathOclIncludes> (X::('a, 'b::bot) Set)"
apply (rule ext)
apply (simp add:  OclExcludes_def OclIncludes_def)
apply (simp only: OclNot_def)
apply (simp only: OclIncludes_def OclExcludes_def lift1_def lift2_def 
                  strictify_def o_def not_def)
apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def 
                             split add: split_if up.split)
done

(*
lemma mtSet_Includes_charn :
" \<tau> \<MathOclValid> \<MathOclIsDefined> x \<Longrightarrow> 
  \<tau> \<MathOclValid> (x::('a, 'b)VAL) \<MathOclExcludes> (\<MathOclMtSet>::('a, 'b::bot) Set)"
apply(simp add: localValidDefined2sem)
apply(simp only: exludes_not_includes OclMtSet_def OclLocalValid_def
                OclIncludes_def  lift0_def lift1_def lift2_def 
                strictify_def o_def OclNot_def OclTrue_def DEF_def) 
apply(simp add: Abs_Set_Mt_DEF [simplified DEF_def] strictify_def)
done


lemma X_in_Set_X:
" \<tau> \<MathOclValid> \<MathOclIsDefined> (S::('a, 'b::bot)Set) \<Longrightarrow> 
  \<tau> \<MathOclValid> S \<MathOclIncludes> (Set S)"
oops

lemma mtSet_in_Set:
" \<tau> \<MathOclValid> \<MathOclIsDefined> (S::('a, 'b::bot)Set) \<Longrightarrow> 
  \<tau> \<MathOclValid> (\<MathOclMtSet>::('a,'b::bot)Set) \<MathOclIncludes> (Set S)"
oops


subsection{* Consequences of Smashedness of Sets *}

text{* This is the characteristic property of smashed sets:
       if \verb+X ->includes(x)+ is valid, we know that x must be defined. *}

lemma smashed_sets_charn:
"\<tau> \<MathOclValid> ((x::('a,'b)VAL) \<MathOclIncludes> (X::('a,'b::bot)Set)) \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> x"
apply (simp add: OclIncludes_def lift2_def lift1_def lift0_def 
                 strictify_def OclTrue_def OclIsDefined_def o_def OclLocalValid_def)
apply(case_tac "X \<tau> = \<bottom>", auto)
apply(simp add: strictify_def DEF_def)
apply(case_tac "x \<tau> = \<bottom>", auto)
done


section{* Fundamental Theorem on (infinite) Sets *}

lemma Set_Ext0 : 
"(\<forall>  x::(('a,'b)VAL). \<tau> \<MathOclValid> \<MathOclIsDefined> x \<longrightarrow> (x \<MathOclIncludes> (S::('a,'b::bot)Set))\<tau> = (x \<MathOclIncludes> T)\<tau>) 
 \<Longrightarrow> S \<tau> = T \<tau>" 
apply (simp add: OclIncludes_def lift2_def lift1_def lift0_def 
                 strictify_def OclTrue_def OclIsDefined_def o_def OclLocalValid_def)
apply(case_tac "S \<tau> = \<bottom>", auto)
apply(case_tac "T \<tau> = \<bottom>", auto)
apply(insert exists_DEF, auto)
apply(case_tac "T \<tau> = \<bottom>", auto)
apply(simp add: DEF_def[symmetric])
apply (frule DEF_Rep_Set)
apply (frule DEF_Rep_Set) back 
apply (simp add: DEF_X_Set)
apply auto
apply (rule_tac f = "\<lambda> c. Abs_Set_0 \<HolOclLiftLeft>c\<HolOclLiftRight>" in arg_cong)
apply (rule set_ext)
apply (erule_tac x="\<lambda> r. xa" in allE) 
apply(case_tac "x=\<bottom>", simp_all add: DEF_def[symmetric])
apply (auto simp: DEF_def)
done


lemma Set_Ext : 
assumes H :"\<And>x::(('a,'b)VAL). \<tau> \<MathOclValid> \<MathOclIsDefined> x \<Longrightarrow> (x \<MathOclIncludes> (S::('a,'b::bot)Set))\<tau> = (x \<MathOclIncludes> T)\<tau>"
shows      "S \<tau> = T \<tau>" 
apply(rule Set_Ext0)back
by(auto intro: H)

*)
section{* Semantic Definitions of the Standard Set Operations *}

defs 
   (* size         :: "('a,'b::bot Set_0)VAL=>'a Integer" *)
   OclSize_def     :"(OclSize::('a, 'b::bot Set_0) VAL => 'a Integer) \<equiv> 
                         (lift1 (strictify (\<lambda>X. 
                               if infinite \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> then  \<bottom> 
                               else (\<HolOclLiftLeft>(int(card(\<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>)))\<HolOclLiftRight>))))" 

   (* count        :: "[('a, 'b Set_0) VAL, ('a, 'b::bot) VAL] => 'a Integer" *)
   OclCount_def    : "OclCount \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>x. 
                                           \<HolOclLiftLeft>if x \<in> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> 
                                            then int 1 
                                            else int 0\<HolOclLiftRight>)))"
  
   (* includesAll  :: "[('a, 'b::bot Set_0) VAL, ('a, 'b Set_0) VAL] => 'a Boolean" *)
   OclIncludesAll_def: "OclIncludesAll \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. 
                                           \<HolOclLiftLeft>\<HolOclDropLeft>Rep_Set_0 Y\<HolOclDropRight> \<subseteq> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>)))" 

   (* excludesAll  :: "[('a, 'b::bot Set_0) VAL, ('a, 'b Set_0) VAL] => 'a Boolean" *)
   OclExcludesAll_def: "OclExcludesAll \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. 
                                           \<HolOclLiftLeft>\<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> \<inter> \<HolOclDropLeft>Rep_Set_0 Y\<HolOclDropRight> = {}\<HolOclLiftRight>)))"

 
   (* OclIsEmpty      ::"('a, 'b::bot Set_0) VAL => 'a Boolean" *)
   OclIsEmpty_def  : "OclIsEmpty \<equiv> lift1 (strictify (\<lambda>X. 
                                           \<HolOclLiftLeft>{} =  \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>))"
   OclNotEmpty_def : "OclNotEmpty \<equiv> lift1 (strictify (\<lambda>X. 
                                           \<HolOclLiftLeft>{}  \<noteq>  \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>))"


defs
(* PROBLEM: wrongly defined in the standard: Idempotence!.
   TODO:    A definition that captures the intention of the standard. *)
   OclSum_def    : "OclSum (self::('a, 'b::bot Set_0) VAL)  \<equiv> \<MathOclUndefined>"
text{* 
  \OCLglitch[\inlineocl{sum}]{\inlineocl{sum} wrongly defined in the standard: 
  idempotence} 
*}

defs
   OclAsSet_def  : "OclAsSet(self::('a, 'b::bot Set_0) VAL) \<equiv> self"  


ocl_setup_op [OclSize, OclCount, OclIncludesAll, OclExcludesAll,
              OclIsEmpty, OclNotEmpty]

defs
   OclFlatten_def: "(OclFlatten ::('a,'b::bot Set_0 Set_0)VAL \<Rightarrow> ('a,'b Set_0)VAL) \<equiv> 
                   lift1 (strictify (\<lambda>X. Abs_Set_0 
                                           (\<HolOclLiftLeft>\<Union> (Lifting.drop \<circ> Rep_Set_0) ` \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>)))"

ocl_setup_op [OclFlatten]


defs 
(* Set operations *)
(* OclIncluding  :: "[('a,('b::bot) Set_0) VAL,('a,'b) VAL] => ('a,'b Set_0) VAL" *)
   OclIncluding_def:"OclIncluding     \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Set_0 
                                           (\<HolOclLiftLeft>insert Y \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>))))"


(* OclExcluding  :: "[('a,('b::bot) Set_0) VAL,('a,'b) VAL] => ('a,'b Set_0) VAL" *)
   OclExcluding_def:"OclExcluding     \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Set_0 
                                           (\<HolOclLiftLeft> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> - {Y}\<HolOclLiftRight>))))"


(* OclComplement :: "('a, 'b::bot Set_0) VAL => ('a, 'b Set_0) VAL"  *)
   OclComplement_def:"OclComplement   \<equiv> lift1 (strictify(\<lambda>X. Abs_Set_0 
                                           (\<HolOclLiftLeft>- (\<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> - {\<bottom>})\<HolOclLiftRight>)))"


(* OclUnion      :: "[('a,('b::bot)Set_0)VAL, ('a,'b Set_0)VAL] => ('a,'b Set_0)VAL" *)
   OclUnion_def  :  "OclUnion      \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Set_0 
                                           (\<HolOclLiftLeft> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> \<union>  \<HolOclDropLeft>Rep_Set_0 Y\<HolOclDropRight>\<HolOclLiftRight>))))"


(* OclIntersection:: "[('a,('b::bot) Set_0)VAL,('a,'b Set_0)VAL]\<Rightarrow>('a,'b Set_0)VAL" *)
   OclIntersection_def: "OclIntersection \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Set_0 
                                           (\<HolOclLiftLeft> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> \<inter> \<HolOclDropLeft>Rep_Set_0 Y\<HolOclDropRight>\<HolOclLiftRight>))))"


ocl_setup_op [OclIncluding, OclExcluding,
             OclComplement, OclUnion]

ocl_setup_op [OclIntersection]

(*
thm OCL_cp_OclComplement

thm OCL_undef_1_OclIntersection 
    OCL_undef_2_OclIntersection
*)

declare OCL_cp_OclIntersection[simp,intro!]


text{* Note that also sequencing of declarations are possible. *}


syntax
   "@OclFinSet"      :: "args \<Rightarrow> ('a,'b Set_0)VAL"           ("mkSet{(_)}")

translations
  "mkSet{x, xs}"     == "OclIncluding (mkSet{xs}) x"
  "mkSet{x}"         == "OclIncluding OclMtSet x"

lemma "mkSet{1,1,1,0,1} = ?X" oops (*test *)


section{* Properties of the Operations *}

subsection{* Useful Lemmas (necessary in this theory) *}


lemma OclIntersection_idem[simp]: "((S::('a,'b::bot)Set) \<MathOclIntersection> S)  = S "
apply (rule ext)
apply (case_tac "x \<MathOclValid>  \<MathOclIsDefined>(S)")
apply (simp_all, ocl_hypsubst [2], simp_all)
apply (simp only: OclIntersection_def
                  OclLocalValid_def OclIsDefined_def strictify_def
                  OclNot_def 
                  o_def lift0_def lift1_def lift2_def
                  OclTrue_def DEF_def)
apply (simp_all add: strictify_def lift_drop_idem DEF_def[symmetric] 
                     DEF_Rep_Set Rep_Set_0_inverse)
done


subsection{* Lemmas required by the Standard *}
(*
("_ ->excludes'( _ ')" [66,65]65)
    lemma REQ11_7_1_3:     
          "self ->excludes(obj) \<equiv> (self ->count(obj) `==` 0)" 
    oops
    lemma REQ11_7_1_3_alt: 
          "self ->excludes(obj) \<equiv> not(self ->includes(obj))" 
    oops

consts   
    OclIncludesAll  :: "[('a,'b::collection)VAL,('a,'b::collection)VAL] 
                       => 'a Boolean" 
    ("_ ->includesAll'( _ ')" [66,65]65)
    lemma REQ11_7_1_5: 
    "(self ->includesAll(obj)) \<equiv> (obj ->forAll(elem | (self ->includes(elem))))"
    oops


consts
    OclExcludesAll  :: 
    "[('a,'b::collection)VAL,('a,'b::collection)VAL] => 'a Boolean" 
     ("_ ->excludesAll'( _ ')" [66,65]65) 

    lemma REQ11_7_1_5:
    "(self ->excludesAll(obj)) \<equiv> (obj ->forAll(elem|(self ->excludes(elem))))" 
    oops

    ("_ ->isEmpty'( ')" [66])



*)


lemma REQ11_7_1_8: "OclNotEmpty(X::('a, 'b::bot) Set) =  \<MathOclNot> (OclIsEmpty(X))" 
apply (rule ext)
apply (simp add:  OclNotEmpty_def OclIsEmpty_def OclNot_def)
apply (simp only: lift1_def lift2_def 
                  strictify_def o_def not_def)
apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def 
                             split add: split_if up.split)
done

(*
lemma OclIsEmpty_charn : "OclIsEmpty(X) = (X \<MathOclStrictEq> \<MathOclMtSet>)"
(* abstract part *)
apply (rule ext)
apply (case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)")
apply (simp_all)
apply (ocl_hypsubst [2], simp_all)
(* TODO : redo with OCL_subst *)
apply (rule sym, rule trans)
apply (rule strictEq_is_strongEq_LE)
apply (simp_all)
(* semantic part *)
apply (simp only: OclIsDefined_def OclTrue_def OclLocalValid_def OclMtSet_def
                  OclIsEmpty_def OclStrongEq_def
                  lift0_def lift1_def lift2_def 
                  strictify_def o_def not_def)
apply (auto simp: DEF_def)
apply (rule_tac t = "{}" in ssubst, assumption)
apply (subst lift_drop_idem)
apply (rule DEF_Rep_Set)
apply (auto simp: DEF_def Rep_Set_0_inverse)
done


lemma OclIsNotEmpty_local_charn_LJE :
    "\<tau>  \<MathOclValid>  \<MathOclIsDefined>(X::('a,'b::bot)Set) \<Longrightarrow> 
     (\<tau> \<MathOclValid> (\<MathOclNotEmpty>(X))) = 
     (\<exists> a::('a,'b::bot)VAL. \<tau> \<MathOclValid> \<MathOclIsDefined>a \<and> \<tau> \<MathOclValid> a \<MathOclIncludes> X)"
apply (simp only: OclIsDefined_def OclTrue_def OclLocalValid_def 
                  OclNotEmpty_def OclIncludes_def
                  OclIsEmpty_def OclStrongEq_def
                  lift0_def lift1_def lift2_def 
                  strictify_def o_def not_def DEF_def)
apply (simp)
apply (subst neq_commute, subst ex_in_conv[symmetric])
apply auto
apply (rule_tac x="\<lambda> c. x" in exI)
apply (simp add: strictify_def DEF_def[symmetric] DEF_X_Set)
apply auto
apply (subgoal_tac "Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>c\<HolOclLiftRight>) = \<HolOclLiftLeft>c\<HolOclLiftRight>")
apply (auto simp:DEF_def)
done



text{* Sets with defined size must be finite. *}
lemma finite_Sets: "\<tau>  \<MathOclValid>  \<MathOclIsDefined>((self::('a,'b::bot)Set)->size()) \<Longrightarrow> 
                    finite \<HolOclDropLeft>Rep_Set_0 (self \<tau>)\<HolOclDropRight>"
apply (simp add:  OclIsDefined_def OclTrue_def lift0_def lift1_def lift2_def 
                  strictify_def o_def not_def)
apply (simp only: OclSize_def)
apply (simp add:  lift0_def lift1_def lift2_def strictify_def DEF_def UU_up_def)
apply (cases "self \<tau>  = \<bottom>")
apply (auto  simp: OclLocalValid_def OclTrue_def lift0_def)
apply (cases "finite (Lifting.drop (Rep_Set_0 (self \<tau>)))")
apply (auto  simp: OclLocalValid_def)
done

text{* Sets with defined size must be defined. *}
lemma def_if_size_def : "\<tau> \<MathOclValid> \<MathOclIsDefined> (self::('a,'b::bot)Set) ->size() \<Longrightarrow> 
                         \<tau> \<MathOclValid> \<MathOclIsDefined> self"
apply (simp add:  OclIsDefined_def OclTrue_def lift0_def lift1_def lift2_def 
                  strictify_def o_def not_def)
apply (simp only: OclSize_def)
apply (simp add:  lift0_def lift1_def lift2_def OclLocalValid_def
                  strictify_def DEF_def UU_up_def)
apply (cases "self \<tau> = \<bottom>")
apply (auto  simp: OclTrue_def lift0_def)
done

lemma def_if_size_def_global: 
      "\<MathOclValid> \<MathOclIsDefined>(self::('a, 'b::bot)Set) ->size() \<Longrightarrow>  \<MathOclValid> \<MathOclIsDefined> self"
apply (rule valid_intro)
apply (drule_tac \<tau>=\<tau> in valid_elim)
apply (simp add: def_if_size_def)
done
 

text{* Due to our extension to infinite sets, the usual connection between
       isEmpty and size holds only under a precondition *}
lemma REQ11_7_1_7: 
"\<MathOclValid> \<MathOclIsDefined>(self ->size()) 
 \<Longrightarrow> self ->isEmpty() = (self::('a,'b::bot)Set) ->size() \<MathOclStrictEq> 0"
apply (rule ext)
apply (drule_tac \<tau> = x in valid_elim)
apply (subgoal_tac "finite (Lifting.drop (Rep_Set_0 (self x)))")
prefer 2
apply (erule finite_Sets)
apply (simp add:  OclIsEmpty_def OclSize_def OclStrictEq_def 
                  OCL_Integer.Zero_ocl_int_def)
apply (simp only: lift0_def lift1_def lift2_def 
                  strictify_def o_def not_def)
apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def 
                             split add: split_if up.split)
apply (simp only: lift0_def lift1_def lift2_def 
                  strictify_def o_def not_def OclIsDefined_def DEF_def OclTrue_def)
apply auto
done
*)


text{* The standard requires the following  "definition" which
       is erroneous (ignoring required idempotence and typing problems): 
     *}
(*
lemma REQ_11_7_1_4: 
   "self ->count(obj)  = 
    (self ->iterate(elem;acc = 0 | if elem \<MathOclStrictEq> obj 
                                   then acc + 1 
                                   else acc endif))" 
oops
*)
(*
text{* Instead, we suggest the following alternative: *}
lemma REQ_11_7_1_4_alt1: 
   "(self::('a, 'b::bot)Set) ->count((obj::('a,'b)VAL)) = 
    ((self ->select(X | (X \<MathOclStrictEq>  obj))) ->size())"
*)
(*
apply(rule ext)
apply(case_tac "x  \<MathOclValid>  \<MathOclIsDefined>(obj)")
apply(simp_all, ocl_hypsubst [2], simp_all)
   (* not enough stuff about select in order to male hypsubst work here. Do later" *)
*)
(*
apply (rule ext)
apply (simp add:  count_def size_def select_def OclStrictEq_def)
apply (simp only: lift0_def lift1_def lift2_def 
                  strictify_def o_def not_def oclIsDefined_def TRUE_def)
apply (case_tac "self x = \<bottom>")
apply simp_all
apply (simp add:DEF_def)
apply (simp only: lift0_def lift1_def lift2_def 
                  strictify_def o_def not_def oclIsDefined_def TRUE_def)
apply (case_tac "obj x = \<bottom>")
apply simp_all
 *)  
(* do this more abstract : case dist over definedness helps a lot \<dots> *)

lemma REQ_11_7_1_4_alt2: 
   "(self::('a,'b::bot) Set) ->count((obj::('a, 'b)VAL)) = 
    (if obj  \<MathOclIncludes> self then 1 else 0 endif)"
apply (rule ext)
apply (simp add:  OclCount_def OclIncludes_def OclIf_def OclStrictEq_def 
                  OCL_Integer.Zero_ocl_int_def OCL_Integer.One_ocl_int_def)
apply (simp only: lift0_def lift1_def lift2_def lift3_def
                  strictify_def o_def not_def OclIsDefined_def OclTrue_def)
apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def split 
                             add: split_if up.split)
done



section{* Semantic Definitions of the Iterators *}

text{* With "iterator", the standard attepts to subsume several vaguely
  related higher-order concepts such as fold, map, filter and 
  comprehensions. The quantifier $ForAll$ is in fact be defined via 
  fold. Since this approach only works for finite sets, we do not 
  follow the standard here and give a more general definition, but 
  show that this coincides with the standards requirements for finite 
  sets. *}


subsection{* Universal Quantifiers *}


defs 
(* OclForAll       ::"[('a,('b::bot)Set)VAL,('a,'b)VAL => 'a Boolean] => 'a Boolean" *)
   OclForAll_def:  "OclForAll S P \<equiv> \<lambda> \<tau>.  
                    if DEF (S \<tau>)
                    then if  \<forall>x\<in>\<HolOclDropLeft>Rep_Set_0(S \<tau>)\<HolOclDropRight>.  P (\<lambda> \<tau>. x) \<tau> = (\<HolOclLiftLeft>True\<HolOclLiftRight>) 
                         then \<HolOclLiftLeft>True\<HolOclLiftRight>
                         else if \<exists>x\<in>\<HolOclDropLeft>Rep_Set_0(S \<tau>)\<HolOclDropRight>. P (\<lambda> \<tau>. x) \<tau> = (\<HolOclLiftLeft>False\<HolOclLiftRight>) 
                              then \<HolOclLiftLeft>False\<HolOclLiftRight>
                              else \<bottom>
                    else \<bottom>"


lemma cp_forall1 [simp,intro!]: 
"\<lbrakk> cp (S :: (('a,'c)VAL \<Rightarrow> ('a,('b::bot))Set)) \<rbrakk> 
 \<Longrightarrow> cp(\<lambda>X. \<MathOclForAll> Y \<in> S X \<bullet> P (Y\<Colon>'a \<Rightarrow> 'b))"
apply(simp only: cp_def OclForAll_def) 
apply(erule exE, rule exI, rule allI, rule allI)
apply(erule_tac x=X in allE)
apply(erule_tac x=\<tau>  in allE)
apply(simp only:)
done


lemma cp_forall2 [simp,intro!]: 
"\<lbrakk> cp (\<lambda> X St.(\<lambda>x. P (\<lambda>\<tau>. x) X St)); 
   cp (S :: (('a,'c)VAL \<Rightarrow> ('a,('b::bot))Set)) \<rbrakk> 
 \<Longrightarrow> cp(\<lambda>X. \<MathOclForAll> Y \<in> S X \<bullet> P (Y\<Colon>'a \<Rightarrow> 'b) X) "
apply(simp only: cp_def OclForAll_def) 
apply(erule exE)+
apply(rule exI, rule allI, rule allI)
apply (simp only:)
apply(rule_tac t = "(\<lambda>x. P (\<lambda>\<tau>. x) X \<tau> )" and 
               s = "f (X \<tau> ) \<tau> " in subst)
prefer 2 
ML{* Unify.search_bound:=1000; *}
apply(rule refl)
ML{* Unify.search_bound:=20; *}
(* Miracle ! This works. Definitively a unification problem !!! *)
apply simp
done  (* temporary solution. *)
      (* TODO: improve !!! *)



lemma OclForAll_UU1 [simp]:
      "(\<MathOclForAll> x \<in> (\<MathOclUndefined>::('a, 'b::bot)Set) \<bullet> P (x::('a, 'b)VAL)) = \<MathOclUndefined>"
  by(simp add: OclForAll_def OclUndefined_def lift0_def)


lemma OclExists_UU1 [simp]:
      "(\<MathOclExists> x \<in> (\<MathOclUndefined>::('a, 'b::bot)Set) \<bullet> P (x::('a, 'b)VAL)) = \<MathOclUndefined>"
  by(simp add: OclExists_def)

(*

lemma OclForAllMt [simp]:
      "(\<MathOclForAll> x \<in> (\<MathOclMtSet>::('a, 'b::bot)Set)  \<bullet> P (x::('a, 'b::bot)VAL)) = \<MathOclTrue>"
  by(simp add: OclForAll_def OclTrue_def lift0_def OclMtSet_def)

*)

(*
lemma OclExistsMt [simp]:
      "(\<MathOclExists> x \<in> (\<MathOclMtSet>::('a, 'b::bot)Set)  \<bullet> P (x::('a, 'b::bot)VAL)) = \<MathOclFalse>"
  by(simp add: OclExists_def)

*)

(*

lemma ForAllI:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Set)"
and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid> P x" 
shows          "\<tau> \<MathOclValid> (\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b::bot))"
apply (insert defS)
apply (simp add: OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def)
apply (frule DEF_X_Set[THEN iffD1])
apply (simp add: strictify_def DEF_def, erule exE)
apply simp
apply (subgoal_tac "Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>c\<HolOclLiftRight>) = \<HolOclLiftLeft>c\<HolOclLiftRight>")
apply (subgoal_tac "(\<forall>x\<Colon>'b\<in>c. P (\<lambda>\<tau>\<Colon>'a. x) \<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>)") 
apply simp
apply (rule ballI)
apply (rule_tac \<tau>1 = \<tau> in localValid2sem[THEN iffD1])
apply (rule allP)
apply (rule includes_charn2)
apply (simp_all add: DEF_def)
done


lemma undefForAllI1:
"\<lbrakk>   \<tau> \<MathOclValid> \<MathOclIsUndefined> (S::('a,('b::bot))Set)\<rbrakk> 
 \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsUndefined> (\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b))"
by(erule_tac P = "\<lambda> X. (\<MathOclIsUndefined> (\<MathOclForAll> x \<in> X \<bullet> P x))" in subst_LJ_undef, simp_all)


lemma undefForAllI2:
assumes exUndef1: "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('a,('b::bot))Set)"
and     exUndef2: "\<tau> \<MathOclValid> \<MathOclIsUndefined>(P x)"
and     allP    : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid> (P x) \<MathOclOr> \<MathOclIsUndefined>(P x)" 
and     cpP     : "cp P"
shows             "\<tau> \<MathOclValid> \<MathOclIsUndefined>(\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b::bot))"
apply (insert exUndef1 exUndef2)
apply (frule isDefined_if_valid)
apply (drule OCL_is_defopt_OclIncludes[THEN iffD1])
apply (simp add: OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def)
apply (erule conjE)
apply (frule DEF_X_Set[THEN iffD1])
apply (simp add: strictify_def DEF_def, erule exE)
apply simp
apply (subgoal_tac "Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>c\<HolOclLiftRight>) = \<HolOclLiftLeft>c\<HolOclLiftRight>")
apply simp
apply safe
apply (rule_tac x = xa in bexI, simp_all)
prefer 2
apply (rule_tac x = "x \<tau>" in bexI, simp_all)
apply (subgoal_tac "P (\<lambda>\<tau>a. x \<tau>) \<tau> = P x \<tau>", simp)
apply (rule_tac St = \<tau> and P = P in cp_charn, simp add: cpP, rule cpP)
apply (erule_tac Q = "P (\<lambda>\<tau>. xa) \<tau> = \<HolOclLiftLeft>False\<HolOclLiftRight>" in contrapos_pp)
apply (rule_tac \<tau>1 = \<tau> in localValidToNotFalse2sem [THEN iffD1]) back
apply (rule allP)
apply (simp_all add: includes_charn2 DEF_def)
done


lemma isdefForAllI1[simp, intro!]:
"\<tau> \<MathOclValid> \<MathOclIsDefined> (\<MathOclForAll> x \<in> (\<MathOclMtSet>::('a,'b::bot)Set) \<bullet> P(x\<Colon>'a\<Rightarrow>'b))"
by simp
(* Not analytic, but works for rewriting \<dots> *)

lemma isdefForAllI2:
assumes exNot1: "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('a,('b::bot))Set)"
and     exNot2: "\<tau> \<MathOclValid> \<MathOclNot> (P x)"
and     cpP   : "cp P"
shows           "\<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b::bot))"
apply (insert exNot1 exNot2)
apply (frule isDefined_if_valid,
       drule OCL_is_defopt_OclIncludes[THEN iffD1])
apply (simp add: OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def)
apply (erule conjE)
apply (frule DEF_X_Set[THEN iffD1])
apply (simp add: strictify_def DEF_def, erule exE)
apply simp
apply (subgoal_tac "Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>c\<HolOclLiftRight>) = \<HolOclLiftLeft>c\<HolOclLiftRight>")
apply simp
apply safe
apply (case_tac "P x \<tau> = down", simp_all)
apply (auto simp: not_down_exists_lift)
apply (erule_tac Q = "Ball ?X ?Y" in contrapos_pp,simp)
apply (rule_tac x = "x \<tau>" in bexI)
apply (rule trans) prefer 2 apply(assumption)
apply (rule_tac P=P and St=\<tau> in cp_charn, simp_all add:cpP)
done


lemmas isdefForAllI3 = ForAllI[THEN isDefined_if_valid]

lemma   notForAllI:
assumes isin : "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('a,('b::bot))Set)"
and     not  : "\<tau> \<MathOclValid> \<MathOclNot> (P x)"
and     cpP  : "cp P"
shows          "\<tau> \<MathOclValid> \<MathOclNot> (\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b::bot))"
apply (insert isin not)
apply (frule isDefined_if_valid,
       drule OCL_is_defopt_OclIncludes[THEN iffD1])
apply (simp add: OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def)
apply (erule conjE)
apply (frule DEF_X_Set[THEN iffD1])
apply (simp add: strictify_def DEF_def, erule exE)
apply simp
apply (subgoal_tac "Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>c\<HolOclLiftRight>) = \<HolOclLiftLeft>c\<HolOclLiftRight>")
apply (case_tac "P x \<tau> = down", simp_all)
apply (auto simp: not_down_exists_lift)
apply (rule_tac x = "xa" in bexI, simp_all)
apply (rule_tac x = "x \<tau>" in bexI, simp_all)
apply (subgoal_tac "(P x \<tau>) = (P (\<lambda>\<tau>a. x \<tau>)\<tau>)")
apply (simp only:)
apply (rule_tac P=P and St=\<tau> in cp_charn, simp_all add:cpP)
apply (erule_tac Q = "\<forall>x\<in>c. P (\<lambda>\<tau>. x) \<tau> \<noteq> \<HolOclLiftLeft>False\<HolOclLiftRight>" in contrapos_pp,simp)
apply (rule_tac x = "x \<tau>" in bexI, simp_all)
apply (subgoal_tac "(P x \<tau>) = (P (\<lambda>\<tau>a. x \<tau>)\<tau>)")
apply (simp only:)
apply (rule_tac P=P and St=\<tau> in cp_charn, simp_all add:cpP)
done


lemma   ExistsI:
assumes isin : "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('a,('b::bot))Set)"
and     not  : "\<tau> \<MathOclValid> (P x)"
and     cpP  : "cp P"
shows          "\<tau> \<MathOclValid> (\<MathOclExists> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b::bot))"
by (auto simp:OclExists_def intro!: notForAllI isin not cpP)


lemma   NotExistsI:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Set)"
and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid>  \<MathOclNot> P x" 
shows          "\<tau> \<MathOclValid>  \<MathOclNot>(\<MathOclExists> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b::bot))"
by (auto simp:OclExists_def intro!: ForAllI defS allP)


subsection{* Lemmas required by the standard (NOTE: added by Simon) *}


(* TODO: write a correct alternative formulation *)

text{* 
  \OCLglitch{Wrong definition of the correspondence of includesAll and iterate}

  If \verb|(self = \<MathOclUndefined>) and (obj = \<MathOclMtSet>)| then the left-hand side of the equation evaluates to
  \verb|text \<MathOclUndefined>| whereas the right-hand side evaluates to \verb|text \<MathOclTrue>|.
*}

(* TODO: how can one prove such a statement by a counterexample? *)

lemma REQ11_7_1_5: 
  "(((self::('a,'b::bot)Set) ->includesAll(obj)) = (obj ->forAll(elem | (self ->includes((elem::('a,'b::bot)VAL)))))) = False"
  apply(simp add: OclIncludes_def OclIncludesAll_def OclForAll_def
                  lift0_def lift1_def lift2_def strictify_def DEF_def)
  oops

text{* 
  \OCLglitch{Wrong definition of the correspondence of excludesAll and iterate}

  If \verb|(self = \<MathOclUndefined>) and (obj = \<MathOclMtSet>)| then the 
  left-hand side of the equation evaluates to
  \verb|\<MathOclUndefined>| whereas the right-hand side evaluates to 
  \verb|\<MathOclTrue>|.
*}
lemma REQ11_7_1_6:
  "((self::('a,'b::bot)Set) ->excludesAll(obj)) = (obj ->forAll(elem | (self ->excludes((elem::('a,'b::bot)VAL)))))"
  apply(rule ext)
  apply(simp add: OclExcludes_def OclExcludesAll_def OclForAll_def
                  lift0_def lift1_def lift2_def strictify_def DEF_def)
  apply(auto)
  oops
*)

subsection{* Select and Collect *}

defs
   OclSelect_def: "OclSelect S P \<equiv>  \<lambda> \<tau>. 
                   if DEF (S \<tau>)
                   then if    \<forall>x\<in>\<HolOclDropLeft>Rep_Set_0 (S \<tau>)\<HolOclDropRight>. DEF (P (\<lambda>\<tau>. x) \<tau>)
                        then  Abs_Set_0(\<HolOclLiftLeft>{x\<in>\<HolOclDropLeft>Rep_Set_0(S \<tau>)\<HolOclDropRight>. P(\<lambda>\<tau>. x)\<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>}\<HolOclLiftRight>)
                        else \<bottom>
                   else \<bottom>"
(*

lemma cp_select [simp,intro!]: 
"\<lbrakk> cp (\<lambda> X St.(\<lambda>x. P (\<lambda>\<tau>. x) X St)); 
   cp (S :: (('a,'c)VAL \<Rightarrow> ('a,('b::bot))Set)) \<rbrakk> 
 \<Longrightarrow> cp(\<lambda>X. (\<MathOclSelectLeft>x : S X | (P (x\<Colon>'a \<Rightarrow> 'b::bot)) X\<MathOclSelectRight>))"
apply(simp only: cp_def OclSelect_def) 
apply(erule exE)+
apply(rule exI, rule allI, rule allI)
apply (simp only:)
apply(rule_tac t = "(\<lambda>x. P (\<lambda>\<tau>. x) X St)" and 
               s = "f (X St) St" in subst)
prefer 2 
ML{* Unify.search_bound:=1000; *}
apply(rule refl)
ML{* Unify.search_bound:=20; *}
(* Miracle ! This works. Definitively a unification problem !!! *)
apply simp
done  (* temporary solution. *)
      (* TODO: improve !!! *)


lemma [simp]: 
"(\<MathOclSelectLeft>x : (\<MathOclUndefined>::('a,('b::bot))Set) | (P (x\<Colon>'a \<Rightarrow> 'b::bot))\<MathOclSelectRight>) = \<MathOclUndefined>"
  by(simp add: OclSelect_def OclUndefined_def lift0_def)


lemma SelectPartialUndef : 
assumes ainS : "\<tau> \<MathOclValid> a \<MathOclIncludes> (S::('a,('b::bot))Set)" 
and     allP : "\<tau> \<MathOclValid> \<MathOclIsUndefined> (P a)" 
and     cpP  : "cp P"
shows          "(\<MathOclSelectLeft>x : S | P (x::'a\<Rightarrow>'b::bot)\<MathOclSelectRight>) \<tau> = \<MathOclUndefined> \<tau>"
apply (insert ainS allP)
apply (frule isDefined_if_valid,
       drule OCL_is_defopt_OclIncludes[THEN iffD1])
apply (simp add: OclSelect_def OclIncluding_def OclUndefined_def
                 OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def DEF_def)
apply auto
apply (erule_tac x="a \<tau>" in ballE, simp_all)
apply (subgoal_tac "P (\<lambda>\<tau>'. a \<tau>) \<tau> = P a \<tau>", simp)
apply (rule_tac St = \<tau> and P = P in cp_charn, simp add: cpP,rule cpP)
done


lemma SelectPartialUndef2[simp] : 
assumes ainS : "\<tau> \<MathOclValid> (a::'a\<Rightarrow>'b::bot) \<MathOclIncludes> (S::('a,('b::bot))Set)" 
shows          "(\<MathOclSelectLeft>x : S | (\<lambda>x. \<MathOclUndefined>)((x::'a\<Rightarrow>'b::bot))\<MathOclSelectRight>) \<tau> = \<MathOclUndefined> \<tau>"
by(rule SelectPartialUndef, auto intro!: ainS)


lemma mem_OclSelect_eq:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Set)"
and     defa : "\<tau> \<MathOclValid> \<MathOclIsDefined>(a::('a\<Rightarrow>'b::bot))"
and     ainS : "\<tau> \<MathOclValid> a \<MathOclIncludes> S" 
and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid>  \<MathOclIsDefined>(P x)" 
and     cpP  : "cp P"
shows          "(\<tau> \<MathOclValid> a \<MathOclIncludes> (\<MathOclSelectLeft>x : S | P (x::'a\<Rightarrow>'b::bot)\<MathOclSelectRight>)) 
               = (\<tau> \<MathOclValid> P a)"
apply (insert defS defa)
apply (simp add: OclSelect_def OclIncluding_def
                 OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def)
apply (frule DEF_X_Set[THEN iffD1])
apply (erule exE)
apply (subgoal_tac "Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>c\<HolOclLiftRight>) = \<HolOclLiftLeft>c\<HolOclLiftRight>") prefer 2
apply (rule Abs_Set_0_inverse,simp_all add: smashed_set_charn)
apply (erule conjE)
apply (drule smashed_set_charn[THEN iffD2])
apply (simp add: DEF_def)
apply (subgoal_tac "(\<forall>x\<in>c. P (\<lambda>\<tau>. x) \<tau> \<noteq> down)", simp) 
apply (subgoal_tac "Abs_Set_0 \<HolOclLiftLeft>{x \<in> c. P (\<lambda>\<tau>. x) \<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>}\<HolOclLiftRight> \<noteq> \<bottom>", simp)
apply (subgoal_tac "a \<tau> \<in> c", simp) 
apply (rule_tac t="P a \<tau>" in subst, simp_all)
apply (rule_tac St = \<tau> and P = P in cp_charn, simp add: cpP,rule cpP)
apply (insert  ainS, simp add: OclSelect_def OclIncluding_def
                 OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def)
apply (rule ballI)
apply (rule allP[simplified 
                 OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def DEF_def, simplified])
apply (simp add: defS[THEN localValidDefined2sem[THEN iffD1], simplified DEF_def]
                 strictify_def)
apply auto
done



lemma nonmem_OclSelect:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Set)"
and     defa : "\<tau> \<MathOclValid> \<MathOclIsDefined>(a::('a\<Rightarrow>'b::bot))"
and     aninS: "\<not> (\<tau> \<MathOclValid> a \<MathOclIncludes> S)" 
and     cpP  : "cp P"
shows          "\<not> (\<tau> \<MathOclValid> a \<MathOclIncludes> (\<MathOclSelectLeft>x : S | P (x::'a\<Rightarrow>'b::bot)\<MathOclSelectRight>))"
apply (insert defS defa aninS)
apply (simp add: OclSelect_def OclIncluding_def
                 OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def)
apply (frule DEF_X_Set [THEN iffD1], erule exE, simp add: DEF_def)
done


lemma mem_Select_charn:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Set)"
and     defa : "\<tau> \<MathOclValid> \<MathOclIsDefined>(a::('a\<Rightarrow>'b::bot))"
and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid>  \<MathOclIsDefined>(P x)" 
and     cpP  : "cp P"
shows          "(\<tau> \<MathOclValid> a \<MathOclIncludes> (\<MathOclSelectLeft>x:S | P(x::'a\<Rightarrow>'b::bot)\<MathOclSelectRight>))= 
                (\<tau> \<MathOclValid> P a \<and> \<tau> \<MathOclValid> a \<MathOclIncludes> S)"
apply(case_tac "\<tau> \<MathOclValid> a \<MathOclIncludes> S", simp_all)
by   (intro mem_OclSelect_eq nonmem_OclSelect,
      simp_all add: allP defS defa cpP)+


lemma select_mem_charn:
assumes defT : "\<tau> \<MathOclValid> \<MathOclIsDefined>(T::('a,('b::bot))Set)"
shows          "(\<MathOclSelectLeft>x : S | ((x::('a\<Rightarrow>'b::bot)) \<MathOclIncludes> T)\<MathOclSelectRight>) \<tau> = 
                ((S::('a,'b)Set)  \<MathOclIntersection> T) \<tau>"
(* abstract part *)
apply(case_tac "\<tau>  \<MathOclValid>  \<MathOclIsDefined>(S)")
apply(simp_all, ocl_hypsubst [2], simp_all)
(* semantic part *)
apply (insert defT)
apply (simp only: OclSelect_def OclIncluding_def OclIntersection_def
                 OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def)
apply (simp add: DEF_def)
apply (simp add: DEF_def[symmetric])
apply (frule DEF_Rep_Set)
apply (frule DEF_Rep_Set) back 
apply (simp add:strictify_def )
apply (subgoal_tac "T \<tau> \<noteq> \<bottom>")
apply (simp add: DEF_X_Set)
apply auto
apply (rule_tac t = "c \<inter> ca" in subst)
prefer 2 apply simp
apply blast
done


lemma select_mem_charn_global[simp]: (* really [simp] ? *)
assumes defT : "\<MathOclValid> \<MathOclIsDefined>(T::('a,('b::bot))Set)"
shows          "(\<MathOclSelectLeft>x : S | ((x::('a\<Rightarrow>'b::bot)) \<MathOclIncludes> T)\<MathOclSelectRight>) = 
                ((S::('a,'b)Set)  \<MathOclIntersection> T)"
apply (rule ext, insert defT)
apply (rule select_mem_charn)
apply (auto simp: global_vs_local_validity[symmetric])
done



lemma select_mem_charn_universal[simp]:
shows  "(\<MathOclSelectLeft>x : S | ((x::('a\<Rightarrow>'b::bot)) \<MathOclIncludes> S)\<MathOclSelectRight>)  = 
        ((S::('a,'b)Set))"
apply (rule ext)
apply (case_tac "x  \<MathOclValid>  \<MathOclIsDefined>(S)")
apply (simp_all, ocl_hypsubst [2],simp_all)
apply (rule trans)
apply (rule select_mem_charn)
apply (simp_all)
done

*)
defs
   OclCollect_def: "OclCollect S P \<equiv>  \<lambda> \<tau>. 
                       if DEF (S \<tau>)
                       then if \<forall>x\<in>\<HolOclDropLeft>Rep_Set_0 (S \<tau>)\<HolOclDropRight>. DEF (P (\<lambda>\<tau>. x) \<tau>)
                            then Abs_Set_0 \<HolOclLiftLeft>(\<lambda>x. P (\<lambda>\<tau>. x) \<tau>) ` \<HolOclDropLeft>Rep_Set_0 (S \<tau>)\<HolOclDropRight>\<HolOclLiftRight> else \<bottom>
                       else \<bottom>"


subsection{* Any and One *}


defs
   OclAny_def      : "OclAny S P \<equiv>  \<lambda> \<tau>. 
                   if DEF (S \<tau>)
                   then if   \<forall>x\<in>\<HolOclDropLeft>Rep_Set_0(S \<tau>)\<HolOclDropRight>. DEF(P(\<lambda> \<tau>. x) \<tau>) 
                        then SOME x.(x \<in> \<HolOclDropLeft>Rep_Set_0(S \<tau>)\<HolOclDropRight> \<and> 
                                    P(\<lambda>\<tau>. x) \<tau> = (\<HolOclLiftLeft>True\<HolOclLiftRight>))
                        else \<bottom>
                   else \<bottom>" 

   OclOne_def      : "OclOne (S::('a,('b::bot) Set_0)VAL) P  \<equiv>
                     \<MathOclSizeLeft>(\<MathOclSelectLeft>x : S | P (x\<Colon>'a\<Colon>type \<Rightarrow> 'b\<Colon>bot)\<MathOclSelectRight>)\<MathOclSizeRight> \<MathOclStrongEq> 1"

thm fold_def
text{* The iterator for Sets is not adequately defined in the standard
       (idempotence). This can be overcome by Paulsons theory of
       the finite fold operator of sets: it is always defined,
       but only for bodies P that are ACI, there is an unfold theorem.

       This definition essentially captures the intention of the standard. *}


defs

   OclIterate_def  : "OclIterate (S::('a,('b::bot) Set_0)VAL) P A \<equiv> 
                      \<lambda> \<tau>. if DEF (S \<tau>)
                           then if   infinite \<HolOclDropLeft>Rep_Set_0 (S \<tau>)\<HolOclDropRight> 
                                then \<bottom> 
                                else (fold P (\<lambda>c \<tau>. c) A (\<HolOclDropLeft>Rep_Set_0 (S \<tau>)\<HolOclDropRight>))\<tau>
                           else \<bottom>"

text{* \textbf{Known Limitation}:
We have a slight problem here; the type :
\[
"('a \Rightarrow 'b:: bot Set)
   \Rightarrow (('a \Rightarrow 'b) \Rightarrow ('a \Rightarrow 'b) 
   \Rightarrow 'a \Rightarrow 'b)
      \Rightarrow ('a \Rightarrow 'b) \Rightarrow 'a \Rightarrow 'b"
\]
resulting from this definition is slightly less general than
required from the standard. The reason for this is that the
format of $fold$ from the HOL library requires a separation of
$P$ into $P = \lambda x y. f(g x) y$. Since such a reconstruction
is not possible automatically, we had to provide $\lambda c \tau. c$
for $g$ which leads to the requirement that the type of the
set and the type of the accumulator must be identical.

This is unfortunate since, for the case of the $OclSize$ definition,
the direct formulation as required in the standard via an
iterator falls in the class of not representable 
uses of the Iterator. *}

text{* The theory of this operator develops as follows. *}

subsection{* Other Stuff *}

section{* Properties of the Operations *}

text{* Standard theorems were generated by the thy\_morpher. More to come. *}


end
