(* ****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Set.thy -
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNES 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ***************************************************************************** *)

header {* \OCL Set *}
theory OCL_Set 
imports 
  "$HOLOCL_HOME/src/library/collection/smashed/OCL_Bag" 
  "$HOLOCL_HOME/src/library/collection/smashed/OCL_OrderedSet" 
begin   
 text_raw{* \label{thy:OCL_Set} *}


section{* Is \OCL Set Faithful to the \OCL Standard? *}

text{* Unlike the basic library theories, for this first and
  foundational ``Collection''-type theory the question of faithfulness
  is not merely a syntactic matter (like fusing various semantic
  interpretation functions into one). In fact, it is not even
  trivial to answer. This is mainly for two reasons:
  \begin{enumerate}
  \item We generalize the concept of Set as required by OCL 2.0 to
        \emph{infinite sets}. 
  \item We interprete the standard with respect to the question,
        if \inlineocl{OclUndefind} can be included in a set or not:
        we deny this question in this version of a set theory. 
  \end{enumerate}
The first decision has the advantage, that general Sets can be used to
represent the syntactic category of "types" in the standard as
``characteristic sets'' in \HOL and enables the possibility to reason
over them. From the pragmatics point of view, this also allows for
quantifications of sets of ``values'' in the sense of \OCL 2.0. For
example, it is possible in \holocl{} to express the commutativity law
on Integers \emph{inside} \OCL.

The second decision is due to the fact that the standard avoids a
clear statement to the nature of undefinedness in \OCL. This led to two
forms of interpretations by users and tool-developers: is
\inlineocl+OclUndefined+ just ``null'', so the null-reference, or is it a
semantic construct denoting non-termination or exceptional behaviour?
In the former sense, it is possible to include \inlineocl+OclUndefined+ in
a set, in the latter not.  Admitting undefinedness in sets also has
consequenses in the inclusion operation and the elementhood test
(\inlineocl+->including+ and \inlineocl+->includes+). 
The operational behaviour of Java-implementations reflects the 
latter view, which we chose therefore as default.  Semantically, 
this leads to a quotient construction (called ``smashing'' in the literature) 
which identifies $\{x, \bottom\}$ with $\bottom$.

With respect to the question of faithfulness of our \OCL semantics
we have the following answers:
\begin{enumerate}
\item the required \emph{properties} of \OCL 2.0 should hold in 
      our setting for finite sets, and
\item the subtle issue of $\bottom$s in sets is not mentioned
      at all in them.
\end{enumerate}
Therefore, we show compliance in this sense by providing the individual
proof for these ``required properties''. However, we will only use
them in rare cases in definitions and provide a number of other
properties for them that are more suited for automated reasoning.
*}
        
section{* Foundational Properties of the smash-construction *}
text{* Global \OCL type constructor for \OCL Set types: *}
types    ('\<tau>,'\<alpha>) Set  = "('\<tau>, '\<alpha> Set_0) VAL"

section{* Properties of the Congruence inside the Datatype Adaption *}
lemma Abs_Set_inject_absurd11 [simp]:
  "\<lbrakk> \<bottom> \<notin> x \<rbrakk> \<Longrightarrow> ((Abs_Set_0 down) = (Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(subst Abs_Set_0_inject,
     simp_all add: OCL_Set_type.Set_0_def smash_def)

lemma Abs_Set_0_inject_absurd12 [simp]:
  "\<lbrakk> \<bottom> \<notin> x \<rbrakk> \<Longrightarrow> ((Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Set_0 down)) = False"
  by(subst Abs_Set_0_inject,
     simp_all add: OCL_Set_type.Set_0_def smash_def)

lemma Abs_Set_0_inject_absurd21 [simp]:
  "\<lbrakk> \<bottom> \<notin> x \<rbrakk> \<Longrightarrow> (\<bottom> = (Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(simp add: UU_Set_def)

lemma Abs_Set_0_inject_absurd22 [simp]:
  "\<lbrakk> \<bottom> \<notin> x \<rbrakk> \<Longrightarrow> ((Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<bottom>) = False"
  by(simp add: UU_Set_def)

lemma Abs_Set_0_inject_charn:
  "\<lbrakk> (\<bottom>::'\<alpha>::bot) \<notin> x; (\<bottom>::'\<alpha>::bot) \<notin> y \<rbrakk> \<Longrightarrow> 
   (((Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Set_0 \<HolOclLiftLeft>y\<HolOclLiftRight>)) = (x = y))"
  by(subst Abs_Set_0_inject, 
     simp_all add: OCL_Set_type.Set_0_def smash_def)

lemma Abs_Set_0_inverse_charn1 [simp]:
  "(\<bottom> \<notin> x) \<Longrightarrow> Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<HolOclLiftLeft>x\<HolOclLiftRight>"
  by (subst Abs_Set_0_inverse, 
      simp_all add: OCL_Set_type.Set_0_def smash_def)

lemma Abs_Set_0_inverse_charn2 [simp]:
  "(\<bottom> \<notin> x) \<Longrightarrow> \<HolOclDropLeft>Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)\<HolOclDropRight> = x"
  by (subst Abs_Set_0_inverse, 
      simp_all add: OCL_Set_type.Set_0_def smash_def)

lemma Abs_Set_0_cases_charn:
  assumes bottomCase : "\<lbrakk> x = \<bottom> \<rbrakk> \<Longrightarrow> P"
  assumes listCase   : "\<And>y. (\<lbrakk> x = Abs_Set_0 \<HolOclLiftLeft>y\<HolOclLiftRight>; \<bottom> \<notin> y \<rbrakk> \<Longrightarrow> P)"
  shows                "P"
  apply(rule_tac x="x" in Abs_Set_0_cases)
  apply(case_tac "y = \<bottom>")
  apply(rule bottomCase, simp add: UU_Set_def)
  apply(frule not_down_exists_lift2[THEN iffD1])
  apply(rule_tac y="\<HolOclDropLeft>y\<HolOclDropRight>" in listCase)
  apply(auto simp: OCL_Set_type.Set_0_def smash_def)
  done

lemma Abs_Set_0_induct_charn:
  assumes bottomCase : "P \<bottom>"
  assumes stepCase   : "\<And>y. (\<lbrakk> \<bottom> \<notin> y \<rbrakk> \<Longrightarrow> P (Abs_Set_0 \<HolOclLiftLeft>y\<HolOclLiftRight>))"
  shows                "P x"
  apply(rule_tac x="x" in Abs_Set_0_induct)
  apply(rule_tac  x="Abs_Set_0 y" in Abs_Set_0_cases_charn)
  apply(auto intro!: bottomCase elim: stepCase)
  done


text{* the rest *}

lemma inj_on_Abs_Set: "inj_on Abs_Set_0 OCL_Set_type.Set_0";
   by(rule inj_on_inverseI,rule Set_0.Abs_Set_0_inverse,assumption)

lemma inj_Rep_Set : "inj Rep_Set_0"
   by(rule inj_on_inverseI, rule Set_0.Rep_Set_0_inverse)

lemma smashed_set_charn: "(\<bottom> \<notin> X) = ((\<HolOclLiftLeft>X\<HolOclLiftRight>) \<in> OCL_Set_type.Set_0)"
   by(unfold smash_def OCL_Set_type.Set_0_def UU_Set_def,auto) (*characterization 1 *)

lemma UU_in_smashed_set[simp]: "\<bottom> : OCL_Set_type.Set_0"
   by(unfold smash_def OCL_Set_type.Set_0_def UU_Set_def,auto) (*characterization 2a *)

lemma down_in_smashed_set[simp]: "down : OCL_Set_type.Set_0"
   by(unfold smash_def OCL_Set_type.Set_0_def UU_Set_def,auto) (*characterization 2b *)

lemma mt_in_smashed_set: "\<HolOclLiftLeft>{}\<HolOclLiftRight> \<in> OCL_Set_type.Set_0"
   by (unfold smash_def OCL_Set_type.Set_0_def,auto)


text{* Universal sets are never in @{text "OCL_Set_type.Set_0"},
   since they must contain bottom elements (\HOL's type discipline 
   forces @{text "UNIV"} to be of type @{text "'\<alpha>::bot set"}) and therefore
   their smashed value is $\bottom$. *}
lemma UNIV_notin_smashed_set: "\<HolOclLiftLeft>UNIV\<HolOclLiftRight> \<notin> OCL_Set_type.Set_0"
   by (simp add: smash_def OCL_Set_type.Set_0_def)

text{* In contrast, the set consisting of all 1-lifted elements 
   is in @{text "OCL_Set_type.Set_0"}. This justifies the construction
   of characteristic sets for Boolean, Integer, Real, and String. *}
lemma liftUNIV_in_smashed_set: "\<HolOclLiftLeft>lift ` UNIV\<HolOclLiftRight> \<in> OCL_Set_type.Set_0"
   by (auto simp: smash_def OCL_Set_type.Set_0_def)



lemma DEF_Abs_Set: "\<bottom> \<notin> X \<Longrightarrow> DEF (Abs_Set_0 (\<HolOclLiftLeft>X\<HolOclLiftRight>))"
  by(simp add: DEF_def)

lemma DEF_Rep_Set: "DEF X \<Longrightarrow> DEF (Rep_Set_0 X)"
apply(unfold DEF_def UU_Set_def,auto)
apply(drule_tac f = "Abs_Set_0" in arg_cong)
apply(simp add: Rep_Set_0_inverse)
done

lemma not_DEF_Rep_Set: "\<not> DEF X \<Longrightarrow> \<not> DEF (Rep_Set_0 X)"
apply(unfold DEF_def UU_Set_def,auto)
apply(simp add: Abs_Set_0_inverse)
done


lemma exists_lift_Set: "DEF X \<Longrightarrow> \<exists>c. Rep_Set_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
by (drule DEF_Rep_Set,simp add: DEF_X_up)

lemma exists_lift_Set2: "DEF X \<Longrightarrow> \<exists>c. \<bottom> \<notin> c \<and> Rep_Set_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
apply(frule exists_lift_Set)
apply(auto simp: smashed_set_charn)
apply(drule sym)
apply(simp add: Rep_Set_0)
done

lemma Rep_Set_cases: 
"Rep_Set_0 X = \<bottom> \<or> (\<exists>c. \<bottom> \<notin> c \<and> Rep_Set_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
apply(case_tac "DEF X")
apply (drule exists_lift_Set2)
apply (drule_tac [2] not_DEF_Rep_Set)
apply(simp_all add: DEF_def)
done

lemma DEF_X_Set_0 : "DEF X = (\<exists>c. \<bottom> \<notin> c \<and> Rep_Set_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
apply (insert  Rep_Set_cases [of "X"], auto simp: smashed_set_charn)
apply (drule exists_lift_Set,auto)
apply (rule swap)
prefer 2
apply (rule not_DEF_Rep_Set, auto)
done

text{* The following lemma is crucial for the theory morpher: *} 
lemma DEF_X_Set: "DEF X = (\<exists>c. \<bottom> \<notin> c \<and> X = Abs_Set_0 (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
apply (auto simp: DEF_Abs_Set smashed_set_charn)
apply (simp add:DEF_X_Set_0)
apply (erule exE, rule exI, auto simp: smashed_set_charn)
apply (rule injD [OF inj_Rep_Set])
apply (auto simp: Abs_Set_0_inverse)
done 


lemma [simp]: "\<bottom> \<notin> ca \<Longrightarrow> Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>ca\<HolOclLiftRight>) = \<HolOclLiftLeft>ca\<HolOclLiftRight>"
apply (rule  Abs_Set_0_inverse)
apply (simp add: smash_def OCL_Set_type.Set_0_def)
done
 

lemma RepAbs_UNIV[simp]:
"Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>lift ` UNIV\<HolOclLiftRight>) = \<HolOclLiftLeft>lift ` UNIV\<HolOclLiftRight>"
apply (subst Abs_Set_0_inverse)
apply (simp_all add: Abs_Set_0_inverse liftUNIV_in_smashed_set)
done



lemma Abs_Set_rangelift_DEF[simp]:
"DEF(Abs_Set_0 \<HolOclLiftLeft>range lift\<HolOclLiftRight>)"
by(auto simp: DEF_X_Set_0)


lemma DEF_X_Set':
  "\<lbrakk> X \<noteq> \<bottom> \<rbrakk>  \<Longrightarrow> 
   (\<exists>c. (\<bottom> \<notin> c) \<and> (Rep_Set_0 X = \<HolOclLiftLeft>c\<HolOclLiftRight>)) \<and> 
   (\<exists>c. (\<bottom> \<notin> c) \<and> (X = (Abs_Set_0 \<HolOclLiftLeft>c\<HolOclLiftRight>)))"
  apply(fold DEF_def)
  apply(frule DEF_X_Set[THEN iffD1])
  apply(drule DEF_X_Set_0[THEN iffD1])
  apply(simp)
  done
 
(* TODO: does this hurt to have both versions in the assumptions? *)
lemma Set_sem_cases_0:
  assumes defC: "\<And> c d. \<lbrakk> X \<noteq> \<bottom>; \<bottom> \<noteq> X; 
                          \<bottom> \<notin> c; \<bottom> \<notin> d; 
                          Rep_Set_0 X = \<HolOclLiftLeft>c\<HolOclLiftRight>;
                          X = Abs_Set_0 \<HolOclLiftLeft>d\<HolOclLiftRight> \<rbrakk> \<Longrightarrow> P X"
  and   undefC: "\<lbrakk> X = \<bottom> \<rbrakk> \<Longrightarrow> P X"
  shows "P X"
  apply(rule Abs_Set_0_cases_charn, erule undefC)
  apply(rule defC, auto)
  done

lemma Set_sem_cases:
  assumes defC: "\<And> c d. \<lbrakk> (X \<tau>) \<noteq> \<bottom>; \<bottom> \<noteq> (X \<tau>); 
                          \<bottom> \<notin> c; \<bottom> \<notin> d; 
                          Rep_Set_0 (X \<tau>) = \<HolOclLiftLeft>c\<HolOclLiftRight>;
                          (X \<tau>) = Abs_Set_0 \<HolOclLiftLeft>d\<HolOclLiftRight> \<rbrakk> \<Longrightarrow> P X \<tau>"
  and   undefC: "\<lbrakk> (X \<tau>) = \<bottom> \<rbrakk> \<Longrightarrow> P X \<tau>"
  and      cpP: "cp P"
  shows "P X \<tau>"
  apply(rule_tac P2="P" in subst[OF sym[OF cp_subst]], rule cpP)
  apply(rule Set_sem_cases_0)
  apply(rule_tac P1="P" in subst[OF cp_subst], rule cpP)
  apply(rule defC) prefer 7
  apply(rule_tac P1="P" in subst[OF cp_subst], rule cpP)
  apply(rule undefC, simp_all)
  done

lemma Set_sem_cases_ext:
  assumes defC: "\<And> c d \<tau>. \<lbrakk> (X \<tau>) \<noteq> \<bottom>; \<bottom> \<noteq> (X \<tau>); 
                          (\<bottom> \<notin> c); (\<bottom> \<notin> d); 
                          Rep_Set_0 (X \<tau>) = \<HolOclLiftLeft>c\<HolOclLiftRight>;
                          (X \<tau>) = Abs_Set_0 \<HolOclLiftLeft>d\<HolOclLiftRight> \<rbrakk> \<Longrightarrow> P X \<tau> = Q X \<tau>"
  and   undefC: "\<And> \<tau>. \<lbrakk> (X \<tau>) = \<bottom> \<rbrakk> \<Longrightarrow> P X \<tau> = Q X \<tau>"
  and      cpP: "cp P"
  and      cpQ: "cp Q"
  shows "P X = Q X"
  apply(rule ext)
  apply(rule_tac X="X" in Set_sem_cases)
  apply(rule defC) prefer 7
  apply(rule undefC)
  apply(simp_all add: cpP cpQ)
  done


text{* These four lemmas are used by \inlineholocl{ocl_setup_op} to reason about definedness: *}

lemma lift2_strict_is_isdef_fw_Set_Val:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Set_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Set_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> a; (\<bottom>::'\<alpha>::bot) \<noteq> b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> (g a b))"
  shows          " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def lift0_def lift1_def lift2_def
                  strictify_def DEF_def)
  apply(rule ext)
  apply(case_tac "DEF (X \<tau>)")
  apply(frule DEF_X_Set_0[THEN iffD1])
  apply(auto simp: strictify_def DEF_def)
  apply(drule_tac b="(Y \<tau>)" in  inv_g)
  prefer 2
  apply(drule DEF_Abs_Set, simp add: DEF_def, auto)
  done

lemma lift2_strict_is_isdef_fw_Set_Set:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Set_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Set_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_Set_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> a; \<bottom> \<notin> b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> (g a b))"
  shows          " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def lift0_def lift1_def lift2_def
                  strictify_def DEF_def)
  apply(rule ext)
  apply(case_tac "DEF (X \<tau>)")
  apply(case_tac "DEF (Y \<tau>)")
  apply(drule DEF_X_Set_0[THEN iffD1])+
  apply(auto simp: strictify_def DEF_def)
  apply(drule_tac b="ca" in inv_g, assumption)
  apply(rotate_tac -1)
  apply(drule DEF_Abs_Set, simp add: DEF_def)
  done

lemma lift2_strictify_implies_LocalValid_defined_Set_Val:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Set_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Set_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> a; (\<bottom>::'\<alpha>::bot) \<noteq> b\<rbrakk> \<Longrightarrow> \<bottom> \<notin> (g a b))"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_Set_Val)
  apply(rule inv_g, assumption+)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done

lemma lift2_strictify_implies_LocalValid_defined_Set_Set:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Set_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Set_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_Set_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> a; \<bottom> \<notin> b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> (g a b))"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_Set_Set)
  apply(rule inv_g, assumption+)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done


section{* The conversion operators *}
subsection{* From sets to sequences *}
text{* 
  Because there is no ordering on the elements of the sets the @{text
  "AsSequence"} operator choses some sequence which has exactly the
  same elements as the set.

  Because this is just some sequence the proof about statements
  containing the @{text "AsSequence"} operator, have to show the same
  properties as one has to show for the general Hilbert Choice
  operator.*}
defs
   (* asSequence     :: "('\<tau>,'\<alpha>::bot) Set \<Rightarrow> ('\<tau>,'\<alpha>) Sequence" *)
   OclAsSequence_def  : 
   "OclAsSequence     \<equiv> lift1 (strictify (\<lambda>X. 
                          Abs_Sequence_0 \<HolOclLiftLeft>list_of_set \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>))"
ocl_setup_op [OclAsSequence]

(* TODO: prove if the whole construction with respect to OrderedSet becomes clear 

text{* This rule is crucial for the simplifier to reason about membership in
  this inverse construction. *}
lemma sem_set2seqI:
  assumes finite: "finite X"
  and   condCase: "\<And> xs. \<lbrakk> X = set xs; distinct xs \<rbrakk> \<Longrightarrow> P xs"
  shows           "P (SOME xs. X = set xs \<and> distinct xs)"
  apply(rule someI2_ex)
  apply(rule_tac[2] condCase)
  apply(rule finite_induct)
  apply(auto simp: finite)
  apply(rule_tac x="x#a" in exI, simp)
  done
  

lemma in_set_of_some_sequence [simp]:
  "finite X \<Longrightarrow> a \<in> set (SOME xs. X = set xs \<and> distinct xs) = (a \<in> X)"
  by(rule_tac X="X" in sem_set2seqI, simp_all)

*)
(* THEOREMS from OCL_Sequence.thy

   sadly they hold only in the finite case, because Sequence is only finite.

lemma OCL_is_defopt_OclAsSequence: 
  "(\<tau> \<MathOclValid> \<MathOclIsDefined> (((self::('\<tau>,'\<alpha>::bot)Set) ->asSequence())::(('\<tau>, '\<alpha>::bot)Sequence))) = (\<tau> \<MathOclValid> \<MathOclIsDefined> self)"
  apply(simp add: localValidDefined2sem OclAsSequence_def ss_lifting')
  apply(rule impI, frule DEF_X_Set', auto)
  done

lemma OCL_is_def_OclAsSequence:
  "(\<MathOclIsDefined> (((self::('\<tau>,'\<alpha>::bot)Set) ->asSequence())::(('\<tau>, '\<alpha>::bot)Sequence))) = (\<MathOclIsDefined> self)"
  apply(rule Set_sem_cases_ext, simp_all)
  apply(simp_all add: OclIsDefined_def OclAsSequence_def ss_lifting')
  done


subsubsection{* Characterisation theorems of the conversion from sets to sequences *}


lemma OclAsSequence_charn_0:
  assumes undefCase: "P \<MathOclUndefined> \<bottom> \<tau>"
  and     defCase  : "\<And> xs. \<lbrakk> \<HolOclDropLeft>Rep_Set_0 B\<HolOclDropRight> = multiset_of xs; DEF B; \<bottom> \<notin> set xs \<rbrakk> \<Longrightarrow> 
                            P (\<lambda> s. (Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>)) (\<lambda> s. B) \<tau>"
  shows              "P ((OclAsSequence (\<lambda> s. (B::'\<alpha>::bot Set_0)))::('\<tau>,'\<alpha>::bot)Sequence) (\<lambda> s. B) \<tau>"
  apply(insert undefCase)
  apply(simp add: localValidDefined2sem OclIsDefined_def OclAsSequence_def
                  OclUndefined_def ss_lifting')
  apply(rule impI)
  apply(frule DEF_X_Set_0[THEN iffD1, simplified DEF_def], clarify)
  apply(rule someI2_ex[OF surj_multiset_of[THEN surjD]])
  apply(rule defCase)
  apply(simp_all add: localValidDefined2sem DEF_def mem_set_multiset_eq)
  done

lemma OclAsSequence_charn:
  assumes undefCase: "P \<MathOclUndefined> \<MathOclUndefined> \<tau>"
  and     defCase  : "\<And> xs. \<lbrakk> \<HolOclDropLeft>Rep_Set_0 (B \<tau>)\<HolOclDropRight> = multiset_of xs; \<bottom> \<notin> set xs; \<tau> \<MathOclValid> \<MathOclIsDefined>B \<rbrakk> \<Longrightarrow> 
                            P (\<lambda> s. (Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>)) B \<tau>"
  and     cp1_P    : "\<And> y. cp (\<lambda> x. P x y)"
  and     cp2_P    : "\<And> x. cp (P x)"
  shows              "P ((OclAsSequence (B::('\<tau>,'\<alpha>::bot)Set))::('\<tau>,'\<alpha>::bot)Sequence) B \<tau>"
  apply(rule subst, rule_tac x1="B" in sym[OF cp_subst]) 
  apply(rule_tac F="OclAsSequence" and F'="\<lambda> x. x" in cp_compose_binary) 
  apply(simp_all add: cp1_P cp2_P lift0_def)
  apply(rule_tac P=P in OclAsSequence_charn_0)
  apply(insert undefCase, simp add: OclUndefined_def lift0_def)
  apply(drule defCase, simp_all add: localValidDefined2sem)
  apply(rule subst)
  apply(rule_tac P="\<lambda> x St. (P (\<lambda> s. (Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>)) x) St" and A="B" in cp_charn) 
  apply(simp_all add: cp2_P)
  done

subsection{* Showing the distributivity over the equivalences *}

(* \textbf{Note:} Actually it should be possible to develop the theory about the inverse
  of \verb{multiset_of} in a more general way by combinig the results of Hilbert_Choice.thy
  and Multiset.thy. But I'm currently not using this approach because only few specifically
  tailored theorems are needed *)
lemma count_filter_inv_multiset_of:
  "count M x = length (List.filter (\<lambda> y. y = x) (SOME xs. M = multiset_of xs))"
  apply(subst count_filter[symmetric])
  apply(rule someI2_ex[OF surj_multiset_of[THEN surjD]])
  apply(simp)
  done

lemma set2seq_holEq:
  "((X::('\<tau>,'\<alpha>::bot)Set) = Y) = (((\<MathOclAsSequence> X)::('\<tau>,'\<alpha>)Sequence) = (\<MathOclAsSequence> Y))"
  apply(auto simp: OclAsSequence_def ss_lifting' expand_fun_eq)
  apply(erule_tac x="x" in allE)
  apply(case_tac "X x \<noteq> \<bottom>", frule DEF_X_Set')
  apply(case_tac "Y x \<noteq> \<bottom>", rotate_tac -1, frule DEF_X_Set')
  apply(auto)
  apply(subst (asm) Abs_Sequence_0_inject_charn, auto)
  apply(subst multiset_eq_conv_count_eq)
  apply(subst count_filter_inv_multiset_of)+
  apply(simp)
  apply(case_tac "Y x \<noteq> \<bottom>", rotate_tac -1, frule DEF_X_Set')
  apply(auto)
  done


lemma set2seq_strictEq:
  "((X::('\<tau>,'\<alpha>::bot)Set) \<MathOclStrictEq> Y) = (((\<MathOclAsSequence> X)::('\<tau>,'\<alpha>)Sequence) \<MathOclStrictEq> (\<MathOclAsSequence> Y))"
  apply(rule ext)
  apply(simp add: OclStrictEq_def OclAsSequence_def ss_lifting')
  apply(rule impI)+
  apply(frule DEF_X_Set', rotate_tac 1)
  apply(frule DEF_X_Set', clarify, simp)
  apply(subst Abs_Sequence_0_inject_charn, auto)
  apply(subst multiset_eq_conv_count_eq)
  apply(subst count_filter_inv_multiset_of)+
  apply(simp)
  done

lemma set2seq_strongEq:
  "((X::('\<tau>,'\<alpha>::bot)Set) \<MathOclStrongEq> Y) = (((\<MathOclAsSequence> X)::('\<tau>,'\<alpha>)Sequence) \<MathOclStrongEq> (\<MathOclAsSequence> Y))"
  apply(rule ext)
  apply(auto simp: OclStrongEq_def OclAsSequence_def ss_lifting')
  apply(frule DEF_X_Set', clarify, simp)
  apply(frule DEF_X_Set', clarify, simp)
  apply(frule DEF_X_Set', rotate_tac 1)
  apply(frule DEF_X_Set', clarify, simp)
  apply(subst (asm) Abs_Sequence_0_inject_charn, auto)
  apply(subst multiset_eq_conv_count_eq)
  apply(subst count_filter_inv_multiset_of)+
  apply(simp)
  done

lemmas ss_set2seq = set2seq_strictEq set2seq_strongEq


subsection{* Simplifying multiple conversions *}

text{* asSet is the inverse of asSequence *}
lemma seq2set_of_set2seq_UC:
  "(OclAsSet ((OclAsSequence (B::('\<tau>,'\<alpha>::bot)Set))::('\<tau>,'\<alpha>::bot)Sequence)) = B"
  apply(rule Set_sem_cases_ext, simp_all)
  apply(simp_all add: OclAsSet_def OclAsSequence_def ss_lifting')
  apply(rule someI2_ex[OF surj_multiset_of[THEN surjD]], simp)
  done

text{* This blowup theorem can be quite useful for conversion proofs about
  functions that map from a collection type to the same collection type. *}
lemmas blowup_set2seq = seq2set_of_set2seq_UC[symmetric]

*)
subsection{* From sets to bags *}
defs
   (* asBag_def  :: "('\<tau>,'\<alpha>::bot) Set \<Rightarrow> ('\<tau>,'\<alpha>::bot) Bag" *)
   OclAsBag_def  : 
   "OclAsBag     \<equiv> lift1 (strictify (\<lambda>X. 
                      Abs_Bag_0 \<HolOclLiftLeft>multiset_of_set \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>))"
ocl_setup_op [OclAsBag]

subsection{* From sets to ordered sets *}
defs
   (* asOrderedSet     :: "('\<tau>,'\<alpha>::bot) Set \<Rightarrow> ('\<tau>,'\<alpha>::bot) Set" *)
   OclAsOrderedSet_def : 
   "OclAsOrderedSet    \<equiv> lift1 (strictify (\<lambda>X. 
                           Abs_OrderedSet_0 \<HolOclLiftLeft>list_of_set \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>))"
ocl_setup_op [OclAsOrderedSet]
(*
Verifying the construction: it should hold that \verb|Bag\<rightarrow>Seq\<rightarrow>OSet = Bag\<rightarrow>Set\<rightarrow>OSet| 
*)
lemma bag2seq2oset_eq_bag2set2oset:
  "\<MathOclAsOrderedSet> ((\<MathOclAsSequence> (B::('\<tau>,'\<alpha>::bot)Bag))::('\<tau>,'\<alpha>)Sequence) =
   ((\<MathOclAsOrderedSet> ((\<MathOclAsSet> B)::('\<tau>,'\<alpha>)Set))::('\<tau>,'\<alpha>)OrderedSet)"
  apply(rule ext)
  apply(case_tac "x \<MathOclValid> \<MathOclIsUndefined> B", ocl_hypsubst, simp)
  apply(simp add: localValidDefined2sem DEF_def_both, clarify)
  apply(frule DEF_X_Bag', clarify)
  apply(simp add: OCL_Bag.OclAsSequence_def OCL_Bag.OclAsSet_def 
                  OCL_Sequence.OclAsOrderedSet_def
                  OCL_Set.OclAsOrderedSet_def ss_lifting')
  apply(subst Abs_OrderedSet_0_inject_charn)
  apply(simp_all add: rev_remdups rev_remstutter[symmetric])
  done

lemma oset2seq2bag_eq_oset2set2bag:
  "\<MathOclAsBag> ((\<MathOclAsSequence> (X::('\<tau>,'\<alpha>::bot)OrderedSet))::('\<tau>,'\<alpha>)Sequence) =
   ((\<MathOclAsBag> ((\<MathOclAsSet> X)::('\<tau>,'\<alpha>)Set))::('\<tau>,'\<alpha>)Bag)"
  apply(rule OrderedSet_sem_cases_ext, simp_all)
  apply(simp_all add: OCL_OrderedSet.OclAsSequence_def OCL_OrderedSet.OclAsSet_def
                      OCL_Sequence.OclAsBag_def
                      OCL_Set.OclAsBag_def ss_lifting')
  done



subsection{* From sets to sets *}

defs
   (* asSet     :: "('\<tau>,'\<alpha>::bot) Set \<Rightarrow> ('\<tau>,'\<alpha>::bot) Set" *)
   OclAsSet_def : 
   "OclAsSet    \<equiv> id"


lemma set2set_id [simp]:
  "((OclAsSet (B::('\<tau>,'\<alpha>::bot)Set))::('\<tau>,'\<alpha>)Set) = B"
  by(simp add: OclAsSet_def)


section{* The Definition of the Set Operators *}
subsection{* The Powerset Operator *}

constdefs
   Set     :: "('st, ('\<alpha>::bot) Set_0) VAL \<Rightarrow> ('st, '\<alpha> Set_0 Set_0) VAL"
  "Set     \<equiv> lift1(strictify (\<lambda>X. 
                     Abs_Set_0\<HolOclLiftLeft>(Abs_Set_0 o lift) ` {x. x \<subseteq> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>}\<HolOclLiftRight>))"
(*<*)
ML{* OCL_setup_trace := true; *}
(*>*)
ocl_setup_op [Set]

lemma Set_defined :
" \<tau> \<MathOclValid> \<MathOclIsDefined> X \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsDefined>(Set X)"
apply(simp add: Set_def localValidDefined2sem)
apply (auto simp: strictify_def 
                 o_def lift0_def lift1_def lift2_def image_def
                 OclTrue_def OclFalse_def  OclUndefined_def) 
apply(rule DEF_Abs_Set, simp)
apply(rule allI, rule impI)
apply(rule neq_commute[THEN iffD1])
apply(simp only:DEF_def[symmetric])
apply(rule DEF_Abs_Set, simp)
thm exists_lift_Set2
apply(drule exists_lift_Set2, auto)
done

lemma Set_definedR :
"(\<tau> \<MathOclValid> \<MathOclIsDefined>(Set X)) =  (\<tau> \<MathOclValid> \<MathOclIsDefined> X) "
apply (auto simp: Set_defined )
oops

subsection{* The Empty Set *}

constdefs  
    OclMtSet  :: "('\<tau>, '\<alpha>::bot Set_0) VAL" 
   "OclMtSet  \<equiv> lift0(Abs_Set_0(\<HolOclLiftLeft>{}\<HolOclLiftRight>))"


    syntax  (* standard syntax *)
    "_OclMtSet_std"   :: "('\<tau>,'\<alpha>\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"     ("\<OclMtSet>")
    syntax  (* ascii syntax *)
    "_OclMtSet_ascii" :: "('\<tau>,'\<alpha>\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"     ("mtSet")
    syntax  (xsymbols)
    "_OclMtSet_math"  :: "('\<tau>,'\<alpha>\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"     ("\<MathOclMtSet>")
(*<*)
    parse_translation {* flat(map (emb trans_const) ["OclMtSet"]) *}
    print_translation {* map emb_print ["OclMtSet"] *}
(*>*)


subsection{* Foundational Properties of MtSet *}


lemma Abs_Set_Mt_DEF[simp]:
"DEF(Abs_Set_0 \<HolOclLiftLeft>{}\<HolOclLiftRight>)"
by(simp add: DEF_X_Set_0)
 

lemma mtSet_Defined [simp]:
"\<MathOclIsDefined> \<MathOclMtSet> = \<MathOclTrue>"
by(simp add: localValidDefined2sem Abs_Set_Mt_DEF OclMtSet_def 
             OclIsDefined_def lift1_def lift0_def OclTrue_def)




section{* Foundational Properties of includes and excludes *}

defs 
   (* includes     :: "[('\<tau>, '\<alpha> Set) VAL,('\<tau>, '\<alpha>::bot) VAL] => '\<tau> Boolean" *)
   OclIncludes_def :  "OclIncludes \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>x. 
                                           \<HolOclLiftLeft>x \<in> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>)))"


   (* excludes     :: "[('\<tau>, '\<alpha> Set) VAL, ('\<tau>, '\<alpha>::bot) VAL] => '\<tau> Boolean" *) 
   OclExcludes_def : "OclExcludes \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>x.
                                           \<HolOclLiftLeft>x \<notin> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>)))" 

ocl_setup_op [OclIncludes, OclExcludes]


lemma includes_charn1 : "(\<tau> \<MathOclValid> x \<MathOclIncludes> S) \<Longrightarrow> (x \<tau> \<in> \<HolOclDropLeft>Rep_Set_0 (S \<tau>)\<HolOclDropRight>)" 
apply (frule isDefined_if_valid,
       drule OCL_is_defopt_OclIncludes[THEN iffD1])
apply (simp add: OclLocalValid_def OclIsDefined_def strictify_def
                 OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclFalse_def  OclUndefined_def DEF_def) 
done


lemma includes_charn2 : "\<lbrakk> DEF(S \<tau>); x \<tau> \<in> \<HolOclDropLeft>Rep_Set_0 (S \<tau>)\<HolOclDropRight> \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> x \<MathOclIncludes> S"
apply (frule DEF_X_Set [THEN iffD1], erule exE)
apply (simp add: OclLocalValid_def OclIsDefined_def strictify_def
                 OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclFalse_def  OclUndefined_def DEF_def)
apply auto
done



section{* Higher Properties of includes and excludes *}

(* We prefer a normal form where excludes is eliminated. *)
lemma exludes_not_includes[simp]: 
"x \<MathOclExcludes> X = \<MathOclNot> (x::('\<tau>, '\<alpha>)VAL) \<MathOclIncludes> (X::('\<tau>, '\<alpha>::bot) Set)"
apply (rule ext)
apply (simp add:  OclExcludes_def OclIncludes_def)
apply (simp only: OclNot_def)
apply (simp only: OclIncludes_def OclExcludes_def lift1_def lift2_def 
                  strictify_def o_def not_def)
apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def 
                             split add: split_if up.split)
done


lemma mtSet_Includes_charn :
" \<tau> \<MathOclValid> \<MathOclIsDefined> x \<Longrightarrow> 
  \<tau> \<MathOclValid> (x::('\<tau>, '\<alpha>)VAL) \<MathOclExcludes> (\<MathOclMtSet>::('\<tau>, '\<alpha>::bot) Set)"
apply(simp add: localValidDefined2sem)
apply(simp only: exludes_not_includes OclMtSet_def OclLocalValid_def
                OclIncludes_def  lift0_def lift1_def lift2_def 
                strictify_def o_def OclNot_def OclTrue_def DEF_def) 
apply(simp add: Abs_Set_Mt_DEF [simplified DEF_def] strictify_def)
done


lemma X_in_Set_X:
" \<tau> \<MathOclValid> \<MathOclIsDefined> (S::('\<tau>, '\<alpha>::bot)Set) \<Longrightarrow> 
  \<tau> \<MathOclValid> S \<MathOclIncludes> (Set S)"
oops

lemma mtSet_in_Set:
" \<tau> \<MathOclValid> \<MathOclIsDefined> (S::('\<tau>, '\<alpha>::bot)Set) \<Longrightarrow> 
  \<tau> \<MathOclValid> (\<MathOclMtSet>::('\<tau>,'\<alpha>::bot)Set) \<MathOclIncludes> (Set S)"
oops


subsection{* Consequences of Smashedness of Sets *}

text{* This is the characteristic property of smashed sets:
       if \verb+X ->includes(x)+ is valid, we know that x must be defined. *}

lemma smashed_sets_charn:
"\<tau> \<MathOclValid> ((x::('\<tau>,'\<alpha>)VAL) \<MathOclIncludes> (X::('\<tau>,'\<alpha>::bot)Set)) \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> x"
apply (simp add: OclIncludes_def lift2_def lift1_def lift0_def 
                 strictify_def OclTrue_def OclIsDefined_def o_def OclLocalValid_def)
apply(case_tac "X \<tau> = \<bottom>", auto)
apply(simp add: strictify_def DEF_def)
apply(case_tac "x \<tau> = \<bottom>", auto)
done


section{* Fundamental Theorem on (infinite) Sets *}

lemma Set_Ext0 : 
"(\<forall>  x::(('\<tau>,'\<alpha>)VAL). \<tau> \<MathOclValid> \<MathOclIsDefined> x \<longrightarrow> (x \<MathOclIncludes> (S::('\<tau>,'\<alpha>::bot)Set))\<tau> = (x \<MathOclIncludes> T)\<tau>) 
 \<Longrightarrow> S \<tau> = T \<tau>" 
apply (simp add: OclIncludes_def lift2_def lift1_def lift0_def 
                 strictify_def OclTrue_def OclIsDefined_def o_def OclLocalValid_def)
apply(case_tac "S \<tau> = \<bottom>", auto)
apply(case_tac "T \<tau> = \<bottom>", auto)
apply(insert exists_DEF, auto)
apply(case_tac "T \<tau> = \<bottom>", auto)
apply(simp add: DEF_def[symmetric])
apply (frule DEF_Rep_Set)
apply (frule DEF_Rep_Set) back 
apply (simp add: DEF_X_Set)
apply auto
apply (rule_tac f = "\<lambda> c. Abs_Set_0 \<HolOclLiftLeft>c\<HolOclLiftRight>" in arg_cong)
apply (rule set_ext)
apply (erule_tac x="\<lambda> r. xa" in allE) 
apply(case_tac "x=\<bottom>", simp_all add: DEF_def[symmetric])
apply (auto simp: DEF_def)
done


lemma Set_Ext : 
assumes H :"\<And>x::(('\<tau>,'\<alpha>)VAL). \<tau> \<MathOclValid> \<MathOclIsDefined> x \<Longrightarrow> (x \<MathOclIncludes> (S::('\<tau>,'\<alpha>::bot)Set))\<tau> = (x \<MathOclIncludes> T)\<tau>"
shows      "S \<tau> = T \<tau>" 
apply(rule Set_Ext0)back
by(auto intro: H)


section{* Semantic Definitions of the Standard Set Operations *}

defs 
   (* size         :: "('\<tau>,'\<alpha>::bot Set_0)VAL=>'\<tau> Integer" *)
   OclSize_def     :"(OclSize::('\<tau>, '\<alpha>::bot Set_0) VAL => '\<tau> Integer) \<equiv> 
                         (lift1 (strictify (\<lambda>X. 
                               if infinite \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> then  \<bottom> 
                               else (\<HolOclLiftLeft>(int(card(\<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>)))\<HolOclLiftRight>))))" 

   (* count        :: "[('\<tau>, '\<alpha> Set_0) VAL, ('\<tau>, '\<alpha>::bot) VAL] => '\<tau> Integer" *)
   OclCount_def    : "OclCount \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>x. 
                                           \<HolOclLiftLeft>if x \<in> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> 
                                            then int 1 
                                            else int 0\<HolOclLiftRight>)))"
  
   (* includesAll  :: "[('\<tau>, '\<alpha>::bot Set_0) VAL, ('\<tau>, '\<alpha> Set_0) VAL] => '\<tau> Boolean" *)
   OclIncludesAll_def: "OclIncludesAll \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. 
                                           \<HolOclLiftLeft>\<HolOclDropLeft>Rep_Set_0 Y\<HolOclDropRight> \<subseteq> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>)))" 

   (* excludesAll  :: "[('\<tau>, '\<alpha>::bot Set_0) VAL, ('\<tau>, '\<alpha> Set_0) VAL] => '\<tau> Boolean" *)
   OclExcludesAll_def: "OclExcludesAll \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. 
                                           \<HolOclLiftLeft>\<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> \<inter> \<HolOclDropLeft>Rep_Set_0 Y\<HolOclDropRight> = {}\<HolOclLiftRight>)))"

 
   (* OclIsEmpty      ::"('\<tau>, '\<alpha>::bot Set_0) VAL => '\<tau> Boolean" *)
   OclIsEmpty_def  : "OclIsEmpty \<equiv> lift1 (strictify (\<lambda>X. 
                                           \<HolOclLiftLeft>{} =  \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>))"
   OclNotEmpty_def : "OclNotEmpty \<equiv> lift1 (strictify (\<lambda>X. 
                                           \<HolOclLiftLeft>{}  \<noteq>  \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>))"


defs
(* PROBLEM: wrongly defined in the standard: Idempotence!.
   TODO:    A definition that captures the intention of the standard. *)
   OclSum_def    : "OclSum (self::('\<tau>, '\<alpha>::bot Set_0) VAL)  \<equiv> \<MathOclUndefined>"
text{* 
  \OCLglitch[\inlineocl{sum}]{\inlineocl{sum} wrongly defined in the standard: 
  idempotence} 
*}


ocl_setup_op [OclSize, OclCount, OclIncludesAll, OclExcludesAll,
              OclIsEmpty, OclNotEmpty]

defs
   OclFlatten_def: "(OclFlatten ::('\<tau>,'\<alpha>::bot Set_0 Set_0)VAL \<Rightarrow> ('\<tau>,'\<alpha> Set_0)VAL) \<equiv> 
                   lift1 (strictify (\<lambda>X. Abs_Set_0 
                                           (\<HolOclLiftLeft>\<Union> (Lifting.drop \<circ> Rep_Set_0) ` \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>)))"

ocl_setup_op [OclFlatten]


defs 
(* Set operations *)
(* OclIncluding  :: "[('\<tau>,('\<alpha>::bot) Set_0) VAL,('\<tau>,'\<alpha>) VAL] => ('\<tau>,'\<alpha> Set_0) VAL" *)
   OclIncluding_def:"OclIncluding     \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Set_0 
                                           (\<HolOclLiftLeft>insert Y \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight>\<HolOclLiftRight>))))"


(* OclExcluding  :: "[('\<tau>,('\<alpha>::bot) Set_0) VAL,('\<tau>,'\<alpha>) VAL] => ('\<tau>,'\<alpha> Set_0) VAL" *)
   OclExcluding_def:"OclExcluding     \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Set_0 
                                           (\<HolOclLiftLeft> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> - {Y}\<HolOclLiftRight>))))"


(* OclComplement :: "('\<tau>, '\<alpha>::bot Set_0) VAL => ('\<tau>, '\<alpha> Set_0) VAL"  *)
   OclComplement_def:"OclComplement   \<equiv> lift1 (strictify(\<lambda>X. Abs_Set_0 
                                           (\<HolOclLiftLeft>- (\<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> - {\<bottom>})\<HolOclLiftRight>)))"


(* OclUnion      :: "[('\<tau>,('\<alpha>::bot)Set_0)VAL, ('\<tau>,'\<alpha> Set_0)VAL] => ('\<tau>,'\<alpha> Set_0)VAL" *)
   OclUnion_def  :  "OclUnion      \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Set_0 
                                           (\<HolOclLiftLeft> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> \<union>  \<HolOclDropLeft>Rep_Set_0 Y\<HolOclDropRight>\<HolOclLiftRight>))))"


(* OclIntersection:: "[('\<tau>,('\<alpha>::bot) Set_0)VAL,('\<tau>,'\<alpha> Set_0)VAL]\<Rightarrow>('\<tau>,'\<alpha> Set_0)VAL" *)
   OclIntersection_def: "OclIntersection \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Set_0 
                                           (\<HolOclLiftLeft> \<HolOclDropLeft>Rep_Set_0 X\<HolOclDropRight> \<inter> \<HolOclDropLeft>Rep_Set_0 Y\<HolOclDropRight>\<HolOclLiftRight>))))"


ocl_setup_op [OclIncluding, OclExcluding,
             OclComplement, OclUnion]

ocl_setup_op [OclIntersection]

(*
thm OCL_cp_OclComplement

thm OCL_undef_1_OclIntersection 
    OCL_undef_2_OclIntersection
*)

declare OCL_cp_OclIntersection[simp,intro!]


text{* Note that also sequencing of declarations are possible. *}


syntax
   "@OclFinSet"      :: "args \<Rightarrow> ('\<tau>,'\<alpha> Set_0)VAL"           ("mkSet{(_)}")

(*<*)
translations
  "mkSet{x, xs}"     == "OclIncluding (mkSet{xs}) x"
  "mkSet{x}"         == "OclIncluding OclMtSet x"
(*>*)

lemma "mkSet{1,1,1,0,1} = ?X" oops (*test *)


section{* Properties of the Operations *}

subsection{* Useful Lemmas (necessary in this theory) *}


lemma OclIntersection_idem[simp]: "((S::('\<tau>,'\<alpha>::bot)Set) \<MathOclIntersection> S)  = S "
apply (rule ext)
apply (case_tac "x \<MathOclValid>  \<MathOclIsDefined>(S)")
apply (simp_all, ocl_hypsubst [2], simp_all)
apply (simp only: OclIntersection_def
                  OclLocalValid_def OclIsDefined_def strictify_def
                  OclNot_def 
                  o_def lift0_def lift1_def lift2_def
                  OclTrue_def DEF_def)
apply (simp_all add: strictify_def lift_drop_idem DEF_def[symmetric] 
                     DEF_Rep_Set Rep_Set_0_inverse)
done

lemma not_empty: "OclNotEmpty(X::('\<tau>, '\<alpha>::bot) Set) =  \<MathOclNot> (OclIsEmpty(X))" 
apply (rule ext)
apply (simp add:  OclNotEmpty_def OclIsEmpty_def OclNot_def)
apply (simp only: lift1_def lift2_def 
                  strictify_def o_def not_def)
apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def 
                             split add: split_if up.split)
done


lemma OclIsEmpty_charn : "OclIsEmpty(X) = (X \<MathOclStrictEq> \<MathOclMtSet>)"
(* abstract part *)
apply (rule ext)
apply (case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)")
apply (simp_all)
apply (ocl_hypsubst [2], simp_all)
(* TODO : redo with OCL_subst *)
apply (rule sym, rule trans)
apply (rule strictEq_is_strongEq_LE)
apply (simp_all)
(* semantic part *)
apply (simp only: OclIsDefined_def OclTrue_def OclLocalValid_def OclMtSet_def
                  OclIsEmpty_def OclStrongEq_def
                  lift0_def lift1_def lift2_def 
                  strictify_def o_def not_def)
apply (auto simp: DEF_def)
apply (rule_tac t = "{}" in ssubst, assumption)
apply (subst lift_drop_idem)
apply (rule DEF_Rep_Set)
apply (auto simp: DEF_def Rep_Set_0_inverse)
done



lemma OclIsNotEmpty_local_charn_LJE :
    "\<tau>  \<MathOclValid>  \<MathOclIsDefined>(X::('\<tau>,'\<alpha>::bot)Set) \<Longrightarrow> 
     (\<tau> \<MathOclValid> (\<MathOclNotEmpty>(X))) = 
     (\<exists> a::('\<tau>,'\<alpha>::bot)VAL. \<tau> \<MathOclValid> \<MathOclIsDefined>a \<and> \<tau> \<MathOclValid> a \<MathOclIncludes> X)"
apply (simp only: OclIsDefined_def OclTrue_def OclLocalValid_def 
                  OclNotEmpty_def OclIncludes_def
                  OclIsEmpty_def OclStrongEq_def
                  lift0_def lift1_def lift2_def 
                  strictify_def o_def not_def DEF_def)
apply (simp)
apply (subst neq_commute, subst ex_in_conv[symmetric])
apply auto
apply (rule_tac x="\<lambda> c. x" in exI)
apply (simp add: strictify_def DEF_def[symmetric] DEF_X_Set)
apply auto
apply (subgoal_tac "Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>c\<HolOclLiftRight>) = \<HolOclLiftLeft>c\<HolOclLiftRight>")
apply (auto simp:DEF_def)
done



text{* Sets with defined size must be finite. *}
lemma finite_Sets: "\<tau>  \<MathOclValid>  \<MathOclIsDefined>((self::('\<tau>,'\<alpha>::bot)Set)->size()) \<Longrightarrow> 
                    finite \<HolOclDropLeft>Rep_Set_0 (self \<tau>)\<HolOclDropRight>"
apply (simp add:  OclIsDefined_def OclTrue_def lift0_def lift1_def lift2_def 
                  strictify_def o_def not_def)
apply (simp only: OclSize_def)
apply (simp add:  lift0_def lift1_def lift2_def strictify_def DEF_def UU_up_def)
apply (cases "self \<tau>  = \<bottom>")
apply (auto  simp: OclLocalValid_def OclTrue_def lift0_def)
apply (cases "finite (Lifting.drop (Rep_Set_0 (self \<tau>)))")
apply (auto  simp: OclLocalValid_def)
done

text{* Sets with defined size must be defined. *}
lemma def_if_size_def : "\<tau> \<MathOclValid> \<MathOclIsDefined> (self::('\<tau>,'\<alpha>::bot)Set) ->size() \<Longrightarrow> 
                         \<tau> \<MathOclValid> \<MathOclIsDefined> self"
apply (simp add:  OclIsDefined_def OclTrue_def lift0_def lift1_def lift2_def 
                  strictify_def o_def not_def)
apply (simp only: OclSize_def)
apply (simp add:  lift0_def lift1_def lift2_def OclLocalValid_def
                  strictify_def DEF_def UU_up_def)
apply (cases "self \<tau> = \<bottom>")
apply (auto  simp: OclTrue_def lift0_def)
done

lemma def_if_size_def_global: 
      "\<MathOclValid> \<MathOclIsDefined>(self::('\<tau>, '\<alpha>::bot)Set) ->size() \<Longrightarrow>  \<MathOclValid> \<MathOclIsDefined> self"
apply (rule valid_intro)
apply (drule_tac \<tau>=\<tau> in valid_elim)
apply (simp add: def_if_size_def)
done
 

lemma REQ_11_7_1_4_alt2: 
   "(self::('\<tau>,'\<alpha>::bot) Set) ->count((obj::('\<tau>, '\<alpha>)VAL)) = 
    (if obj  \<MathOclIncludes> self then 1 else 0 endif)"
apply (rule ext)
apply (simp add:  OclCount_def OclIncludes_def OclIf_def OclStrictEq_def 
                  OCL_Integer.Zero_ocl_int_def OCL_Integer.One_ocl_int_def)
apply (simp only: lift0_def lift1_def lift2_def lift3_def
                  strictify_def o_def not_def OclIsDefined_def OclTrue_def)
apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def split 
                             add: split_if up.split)
done



section{* Semantic Definitions of the Iterators *}

text{* With "iterator", the standard attepts to subsume several vaguely
  related higher-order concepts such as fold, map, filter and 
  comprehensions. The quantifier $ForAll$ is in fact be defined via 
  fold. Since this approach only works for finite sets, we do not 
  follow the standard here and give a more general definition, but 
  show that this coincides with the standards requirements for finite 
  sets. *}


subsection{* Universal Quantifiers *}


defs 
(* OclForAll       ::"[('\<tau>,('\<alpha>::bot)Set)VAL,('\<tau>,'\<alpha>)VAL => '\<tau> Boolean] => '\<tau> Boolean" *)
   OclForAll_def: "OclForAll \<equiv> (lift2' lift_arg0 lift_arg1) (strictify( \<lambda> S. (\<lambda> P.  
                      if  \<forall>x \<in> \<HolOclDropLeft>Rep_Set_0 S\<HolOclDropRight>.  P x = \<HolOclLiftLeft>True\<HolOclLiftRight> 
                      then \<HolOclLiftLeft>True\<HolOclLiftRight>
                      else if \<exists>x \<in> \<HolOclDropLeft>Rep_Set_0 S\<HolOclDropRight>. P x = \<HolOclLiftLeft>False\<HolOclLiftRight>
                           then \<HolOclLiftLeft>False\<HolOclLiftRight>
                           else \<bottom>)))"

ocl_setup_op[OclForAll]

(*
lemma cp_forall1 [simp,intro!]: 
"\<lbrakk> cp (S :: (('\<tau>,'\<beta>)VAL \<Rightarrow> ('\<tau>,('\<alpha>::bot))Set)) \<rbrakk> 
 \<Longrightarrow> cp(\<lambda>X. \<MathOclForAll> Y \<in> S X \<bullet> P (Y\<Colon>'\<tau> \<Rightarrow> '\<alpha>))"
apply(simp only: cp_def OclForAll_def) 
apply(erule exE, rule exI, rule allI, rule allI)
apply(erule_tac x=X in allE)
apply(erule_tac x=St in allE)
apply(simp only:)
done


lemma cp_forall2 [simp,intro!]: 
"\<lbrakk> cp (\<lambda> X St.(\<lambda>x. P (\<lambda>\<tau>. x) X St)); 
   cp (S :: (('\<tau>,'\<beta>)VAL \<Rightarrow> ('\<tau>,('\<alpha>::bot))Set)) \<rbrakk> 
 \<Longrightarrow> cp(\<lambda>X. \<MathOclForAll> Y \<in> S X \<bullet> P (Y\<Colon>'\<tau> \<Rightarrow> '\<alpha>) X) "
apply(simp only: cp_def OclForAll_def) 
apply(erule exE)+
apply(rule exI, rule allI, rule allI)
apply (simp only:)
apply(rule_tac t = "(\<lambda>x. P (\<lambda>\<tau>. x) X St)" and 
               s = "f (X St) St" in subst)
prefer 2 
ML{* Unify.search_bound:=1000; *}
apply(rule refl)
ML{* Unify.search_bound:=20; *}
(* Miracle ! This works. Definitively a unification problem !!! *)
apply simp
done  (* temporary solution. *)
      (* TODO: improve !!! *)


lemma OclForAll_UU1 [simp]:
      "(\<MathOclForAll> x \<in> (\<MathOclUndefined>::('\<tau>, '\<alpha>::bot)Set) \<bullet> P (x::('\<tau>, '\<alpha>)VAL)) = \<MathOclUndefined>"
  by(simp add: OclForAll_def OclUndefined_def lift0_def)
*)

(* TODO: provide alternative definitions for OclExist to enable support
         through ocl_setup_op
*)

lemma OclExists_UU1 [simp]:
      "(\<MathOclExists> x \<in> (\<MathOclUndefined>::('\<tau>, '\<alpha>::bot)Set) \<bullet> P (x::('\<tau>, '\<alpha>)VAL)) = \<MathOclUndefined>"
  by(simp add: OclExists_def)


lemma OclForAllMt [simp]:
      "(\<MathOclForAll> x \<in> (\<MathOclMtSet>::('\<tau>, '\<alpha>::bot)Set)  \<bullet> P (x::('\<tau>, '\<alpha>::bot)VAL)) = \<MathOclTrue>"
  by(simp add: OclForAll_def OclTrue_def OclMtSet_def ss_lifting')

lemma OclExistsMt [simp]:
      "(\<MathOclExists> x \<in> (\<MathOclMtSet>::('\<tau>, '\<alpha>::bot)Set)  \<bullet> P (x::('\<tau>, '\<alpha>::bot)VAL)) = \<MathOclFalse>"
  by(simp add: OclExists_def)



lemma ForAllI:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Set)"
and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid> P x" 
shows          "\<tau> \<MathOclValid> (\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
  apply(insert defS)
  apply(rule_tac X="S" in Set_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(subgoal_tac "(\<forall>x \<in> c. P (\<lambda>\<tau>. x) \<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>)") 
  apply(auto simp: OclForAll_def localValid2sem ss_lifting')
  apply(rule_tac \<tau>1 ="\<tau>" in localValid2sem[THEN iffD1], rule allP)
  apply(auto simp: localValid2sem OclIncludes_def ss_lifting')
  done

lemma   notForAllI:
  assumes isin : "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('\<tau>,('\<alpha>::bot))Set)"
  and     not  : "\<tau> \<MathOclValid> \<MathOclNot> (P x)"
  and     cpP  : "cp P"
  shows          "\<tau> \<MathOclValid> \<MathOclNot> (\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
  apply(insert isin not cpP)
  apply(rule_tac X="S" in Set_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(subgoal_tac "(P x \<tau>) = (P (\<lambda>s. (x \<tau>)) \<tau>)")
  apply(case_tac "(P x) \<tau> \<noteq> \<bottom>", frule neq_commute[THEN iffD1])
  apply(case_tac "x \<tau> \<noteq> \<bottom>", rotate_tac -1, frule neq_commute[THEN iffD1])
  apply(subgoal_tac "(\<exists>x\<in> c. ((P (\<lambda>s. x) \<tau>) = \<HolOclLiftLeft>False\<HolOclLiftRight>))")
  apply(simp_all add: OclForAll_def OclIncludes_def OclNot_def
                      ss_lifting' localValid2sem o_def)
  apply(rule_tac x="x \<tau>" in bexI)
  apply(rule_tac x="P (\<lambda>s. (x \<tau>)) \<tau>" in boolean_cases_sem)
  apply(simp_all)
  apply(rule_tac x="x \<tau>" in bexI)
  apply(rule_tac x="P (\<lambda>s. (x \<tau>)) \<tau>" in boolean_cases_sem)
  apply(simp_all)
  apply(rule trans, rule_tac x="x" in cp_subst)
  apply(simp_all add: lift0_def)
  done

lemma ForAllD:
  assumes allP : "\<tau> \<MathOclValid> (\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
  and     cpP     : "cp P"
  shows   "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> (S::('\<tau>,('\<alpha>::bot))Set) \<Longrightarrow> \<tau> \<MathOclValid> P x" 
  apply (insert allP cpP)
  apply (simp add: localValid2sem OclForAll_def OclNot_def OclIncludes_def
                   ss_lifting')
  apply (simp add:   strictify_def
              split: split_if_asm)
  apply (erule_tac x="x \<tau>" in ballE)
  apply (simp_all add: cp_by_cpify lift0_def)
  done

lemma localValidForall2forall :
  assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Set)" 
  and     cpP  : "cp P"
  shows   "(\<tau> \<MathOclValid> \<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot)) =
           (\<forall> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<longrightarrow> \<tau> \<MathOclValid> P x)"
  apply (safe)
  apply (drule  ForAllD, simp_all add: cpP)
  apply (rule   ForAllI, simp_all add: defS)
  done


lemma undefForAllI1:
  "\<lbrakk>   \<tau> \<MathOclValid> \<MathOclIsUndefined> (S::('\<tau>,('\<alpha>::bot))Set)\<rbrakk> 
   \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsUndefined> (\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>))"
  by(erule_tac P = "\<lambda> X. (\<MathOclIsUndefined> (\<MathOclForAll> x \<in> X \<bullet> P x))" in subst_LJ_undef, simp_all)

lemma undefForAllI2:
  assumes exUndef1: "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('\<tau>,('\<alpha>::bot))Set)"
  and     exUndef2: "\<tau> \<MathOclValid> \<MathOclIsUndefined>(P x)"
  and     allP    : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid> (P x) \<MathOclOr> \<MathOclIsUndefined>(P x)" 
  and     cpP     : "cp P"
  shows             "\<tau> \<MathOclValid> \<MathOclIsUndefined>(\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
  apply(insert exUndef1 exUndef2)
  apply(frule isDefined_if_valid)
  apply(drule OCL_is_defopt_OclIncludes[THEN iffD1])
  apply(simp add: OclLocalValid_def OclIsDefined_def OclTrue_def
                  OclForAll_def OclNot_def OclIncludes_def
                  ss_lifting')
  apply(erule conjE)
  apply(frule DEF_X_Set', clarify)
  apply(simp, safe)
  apply(rule_tac x = xa in bexI, simp_all) prefer 2
  apply(rule_tac x = "x \<tau>" in bexI, simp_all)
  apply(subgoal_tac "P (\<lambda>\<tau>a. x \<tau>) \<tau> = P x \<tau>", simp)
  apply(rule_tac \<tau> = \<tau> and P = P in cp_charn, simp add: cpP, rule cpP)
  apply(erule_tac Q = "P (\<lambda>\<tau>. xa) \<tau> = \<HolOclLiftLeft>False\<HolOclLiftRight>" in contrapos_pp)
  apply(rule_tac \<tau>1 = \<tau> in localValidToNotFalse2sem [THEN iffD1]) back
  apply(rule allP)
  apply(simp add: includes_charn2 DEF_def)
  done

(* TODO: not needed anymore, as this expression is reduced
         to \<MathOclFalse> automatically 
lemma isdefForAllI1[simp, intro!]:
"\<tau> \<MathOclValid> \<MathOclIsDefined> (\<MathOclForAll> x \<in> (\<MathOclMtSet>::('\<tau>,'\<alpha>::bot)Set) \<bullet> P(x\<Colon>'\<tau>\<Rightarrow>'\<alpha>))"
  by simp
(* Not analytic, but works for rewriting \<dots> *)

*)


lemma isdefForAllI2:
  assumes exNot1: "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('\<tau>,('\<alpha>::bot))Set)"
  and     exNot2: "\<tau> \<MathOclValid> \<MathOclNot> (P x)"
  and     cpP   : "cp P"
  shows           "\<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
  apply(insert exNot1 exNot2 cpP)
  apply(drule_tac P="P" in notForAllI)
  apply(simp_all add: isDefined_if_invalid)
  done

lemma isDefinedForall_UC:
assumes  cpP   : "cp P"
shows    "(\<MathOclIsDefined>(\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))) =
          (( \<MathOclIsDefined> (S::('\<tau>,'\<alpha>)Set)) \<MathOclAnd>  
           ((\<MathOclExists> x \<in> S \<bullet> (\<MathOclNot> (P x))) \<MathOclOr>
            (\<MathOclForAll> x \<in> S \<bullet> (\<MathOclIsDefined> (P x)))))"
oops

lemma isDefinedForall_LJE:
assumes  cpP   : "cp P"
shows    "\<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot)) =
          ((\<tau> \<MathOclValid> \<MathOclIsDefined> S) \<and> 
           ((\<exists> x. (\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('\<tau>,'\<alpha>)Set)) \<and> \<not> (\<tau> \<MathOclValid> P x)) \<or> 
           (\<forall> x.  (\<tau> \<MathOclValid> \<MathOclIsDefined> x \<and> \<tau> \<MathOclValid> x \<MathOclIncludes> S)  \<longrightarrow> (\<tau> \<MathOclValid> \<MathOclIsDefined> (P x)))))"
oops 


lemmas isdefForAllI3 = ForAllI[THEN isDefined_if_valid]


lemma   ExistsI:
assumes isin : "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('\<tau>,('\<alpha>::bot))Set)"
and     not  : "\<tau> \<MathOclValid> (P x)"
and     cpP  : "cp P"
shows          "\<tau> \<MathOclValid> (\<MathOclExists> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
by (auto simp:OclExists_def intro!: notForAllI isin not cpP)


lemma   NotExistsI:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Set)"
and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid>  \<MathOclNot> P x" 
shows          "\<tau> \<MathOclValid>  \<MathOclNot>(\<MathOclExists> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
by (auto simp:OclExists_def intro!: ForAllI defS allP)


lemma   ExistsD:
  assumes ExP : "\<tau> \<MathOclValid> (\<MathOclExists> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
  shows         "\<exists> x. \<tau> \<MathOclValid> x \<MathOclIncludes> (S::('\<tau>,('\<alpha>::bot))Set) \<and> \<tau> \<MathOclValid> (P x)"
  apply (insert ExP)
  apply (simp add: localValid2sem OclForAll_def OclNot_def OclIncludes_def
                   OclExists_def o_def ss_lifting)
  apply (simp add:   strictify_def
              split: split_if_asm)
  apply(frule DEF_X_Set', clarify)
  apply(rule_tac x="lift0 x" in exI)
  apply(auto simp: lift0_def not_down_exists_lift)
  done

lemma   localValidExists2exists:
  assumes cpP  : "cp P"
  assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,'\<alpha>::bot)Set)"
  shows          "(\<tau> \<MathOclValid> (\<MathOclExists> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))) = 
                  (\<exists> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<and> \<tau> \<MathOclValid> (P x))"
  apply (safe)
  apply (drule ExistsD, simp)
  apply (auto intro!:  ExistsI cpP)
  done

lemmas weak_pred_LJE = weak_prop_LJE 
                       localValidForall2forall
                       localValidExists2exists

thm weak_pred_LJE

subsection{* Select and Collect *}
defs
   (* select    :: "[('\<tau>,'\<beta>::bot Set) VAL, ('\<tau>,'\<alpha>::bot)VAL => '\<tau> Boolean] *)
   OclSelect_def:  "OclSelect \<equiv>  (lift2' lift_arg0 lift_arg1) (strictify (\<lambda> S P.
                       if    \<forall> x \<in> \<HolOclDropLeft>Rep_Set_0 S\<HolOclDropRight>. DEF (P x)
                       then  Abs_Set_0(\<HolOclLiftLeft>{x\<in>\<HolOclDropLeft>Rep_Set_0 S\<HolOclDropRight>. P x = \<HolOclLiftLeft>True\<HolOclLiftRight>}\<HolOclLiftRight>)
                       else  \<bottom>))"
ocl_setup_op [OclSelect]

lemma SelectPartialUndef : 
assumes ainS : "\<tau> \<MathOclValid> a \<MathOclIncludes> (S::('\<tau>,('\<alpha>::bot))Set)" 
and     allP : "\<tau> \<MathOclValid> \<MathOclIsUndefined> (P a)" 
and     cpP  : "cp P"
shows          "(\<MathOclSelectLeft>x : S | P (x::'\<tau>\<Rightarrow>'\<alpha>::bot)\<MathOclSelectRight>) \<tau> = \<MathOclUndefined> \<tau>"
apply (insert ainS allP)
apply (frule isDefined_if_valid,
       drule OCL_is_defopt_OclIncludes[THEN iffD1])
apply (simp add: OclSelect_def OclIncluding_def OclUndefined_def
                 OclLocalValid_def OclIsDefined_def OclTrue_def
                 OclForAll_def OclNot_def OclIncludes_def
                 ss_lifting')
apply auto
apply (erule_tac x="a \<tau>" in ballE, simp_all)
apply (subgoal_tac "P (\<lambda>\<tau>'. a \<tau>) \<tau> = P a \<tau>", simp)
apply (rule_tac \<tau> = \<tau> and P = P in cp_charn, simp add: cpP,rule cpP)
done

lemma SelectPartialUndef2[simp] : 
assumes ainS : "\<tau> \<MathOclValid> (a::'\<tau>\<Rightarrow>'\<alpha>::bot) \<MathOclIncludes> (S::('\<tau>,('\<alpha>::bot))Set)" 
shows          "(\<MathOclSelectLeft>x : S | (\<lambda>x. \<MathOclUndefined>)((x::'\<tau>\<Rightarrow>'\<alpha>::bot))\<MathOclSelectRight>) \<tau> = \<MathOclUndefined> \<tau>"
by(rule SelectPartialUndef, auto intro!: ainS)


lemma mem_OclSelect_eq:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Set)"
and     defa : "\<tau> \<MathOclValid> \<MathOclIsDefined>(a::('\<tau>\<Rightarrow>'\<alpha>::bot))"
and     ainS : "\<tau> \<MathOclValid> a \<MathOclIncludes> S" 
and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid>  \<MathOclIsDefined>(P x)" 
and     cpP  : "cp P"
shows          "(\<tau> \<MathOclValid> a \<MathOclIncludes> (\<MathOclSelectLeft>x : S | P (x::'\<tau>\<Rightarrow>'\<alpha>::bot)\<MathOclSelectRight>)) 
               = (\<tau> \<MathOclValid> P a)"
apply (insert defS defa ainS)
apply (simp add: OclSelect_def OclIncluding_def OclUndefined_def
                 OclLocalValid_def OclIsDefined_def OclTrue_def
                 OclForAll_def OclNot_def OclIncludes_def
                 ss_lifting')
apply (frule DEF_X_Set', clarify)
apply(simp)
apply(rule_tac t="P a \<tau>" in subst)
apply(rule sym, rule_tac P=P and x=a in cp_subst, rule cpP)
apply(rule_tac x="P (lift0 (a \<tau>)) \<tau>" in boolean_cases_sem)
apply(auto simp: lift0_def)
apply(rotate_tac -1, erule contrapos_pp)
apply(rule allP[simplified 
                 OclLocalValid_def OclIsDefined_def OclTrue_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def ss_lifting', simplified])
apply(auto simp: strictify_def)
done


lemma nonmem_OclSelect:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Set)"
and     defa : "\<tau> \<MathOclValid> \<MathOclIsDefined>(a::('\<tau>\<Rightarrow>'\<alpha>::bot))"
and     aninS: "\<not> (\<tau> \<MathOclValid> a \<MathOclIncludes> S)" 
and     cpP  : "cp P"
shows          "\<not> (\<tau> \<MathOclValid> a \<MathOclIncludes> (\<MathOclSelectLeft>x : S | P (x::'\<tau>\<Rightarrow>'\<alpha>::bot)\<MathOclSelectRight>))"
apply (insert defS defa aninS)
apply (simp add: OclSelect_def OclIncluding_def
                 OclLocalValid_def OclIsDefined_def OclTrue_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def ss_lifting')
apply (frule DEF_X_Set', clarify, simp)
done


lemma mem_Select_charn:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Set)"
and     defa : "\<tau> \<MathOclValid> \<MathOclIsDefined>(a::('\<tau>\<Rightarrow>'\<alpha>::bot))"
and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid>  \<MathOclIsDefined>(P x)" 
and     cpP  : "cp P"
shows          "(\<tau> \<MathOclValid> a \<MathOclIncludes> (\<MathOclSelectLeft>x:S | P(x::'\<tau>\<Rightarrow>'\<alpha>::bot)\<MathOclSelectRight>))= 
                (\<tau> \<MathOclValid> P a \<and> \<tau> \<MathOclValid> a \<MathOclIncludes> S)"
apply(case_tac "\<tau> \<MathOclValid> a \<MathOclIncludes> S", simp_all)
by   (intro mem_OclSelect_eq nonmem_OclSelect,
      simp_all add: allP defS defa cpP)+


lemma select_mem_charn:
assumes defT : "\<tau> \<MathOclValid> \<MathOclIsDefined>(T::('\<tau>,('\<alpha>::bot))Set)"
shows          "(\<MathOclSelectLeft>x : S | ((x::('\<tau>\<Rightarrow>'\<alpha>::bot)) \<MathOclIncludes> T)\<MathOclSelectRight>) \<tau> = 
                ((S::('\<tau>,'\<alpha>)Set)  \<MathOclIntersection> T) \<tau>"
apply(insert defT)
apply(simp add: localValidDefined2sem DEF_def)
apply(frule DEF_X_Set', clarify)
apply(rule_tac X=S in Set_sem_cases, simp_all)
apply(simp_all add: OclSelect_def OclIncludes_def 
                    OclIntersection_def ss_lifting')
apply(safe)
apply(subst Abs_Set_0_inject_charn)
apply(auto)
done


lemma select_mem_charn_global[simp]: (* really [simp] ? *)
assumes defT : "\<MathOclValid> \<MathOclIsDefined>(T::('\<tau>,('\<alpha>::bot))Set)"
shows          "(\<MathOclSelectLeft>x : S | ((x::('\<tau>\<Rightarrow>'\<alpha>::bot)) \<MathOclIncludes> T)\<MathOclSelectRight>) = 
                ((S::('\<tau>,'\<alpha>)Set)  \<MathOclIntersection> T)"
apply (rule ext, insert defT)
apply (rule select_mem_charn)
apply (auto simp: global_vs_local_validity[symmetric])
done



lemma select_mem_charn_universal[simp]:
shows  "(\<MathOclSelectLeft>x : S | ((x::('\<tau>\<Rightarrow>'\<alpha>::bot)) \<MathOclIncludes> S)\<MathOclSelectRight>)  = 
        ((S::('\<tau>,'\<alpha>)Set))"
apply (rule ext)
apply (case_tac "x  \<MathOclValid>  \<MathOclIsDefined>(S)")
apply (simp_all, ocl_hypsubst [2],simp_all)
apply (rule trans)
apply (rule select_mem_charn)
apply (simp_all)
done


(* TODO: introduce OclCollectNested and adapt OclCollect to its defintion*)
defs
   OclCollect_def: "OclCollect S P \<equiv>  \<lambda> \<tau>. 
                       if DEF (S \<tau>)
                       then if \<forall>x\<in>\<HolOclDropLeft>Rep_Set_0 (S \<tau>)\<HolOclDropRight>. DEF (P (\<lambda>\<tau>. x) \<tau>)
                            then Abs_Set_0 \<HolOclLiftLeft>(\<lambda>x. P (\<lambda>\<tau>. x) \<tau>) ` \<HolOclDropLeft>Rep_Set_0 (S \<tau>)\<HolOclDropRight>\<HolOclLiftRight> else \<bottom>
                       else \<bottom>"


subsection{* Any and One *}
defs
   OclAny_def : 
  "OclAny     \<equiv> (lift2' lift_arg0 lift_arg1) (strictify (\<lambda> S P. 
                 if   (\<forall> x \<in> \<HolOclDropLeft>Rep_Set_0 S\<HolOclDropRight>. DEF (P x)) \<and>
                      (\<exists> x \<in> \<HolOclDropLeft>Rep_Set_0 S\<HolOclDropRight>. P x = \<HolOclLiftLeft>True\<HolOclLiftRight>)
                 then SOME x.(x \<in> \<HolOclDropLeft>Rep_Set_0 S\<HolOclDropRight> \<and> P x = \<HolOclLiftLeft>True\<HolOclLiftRight>)
                 else \<bottom>))"

ocl_setup_op [OclAny]

defs
   (* one         ::"[('\<tau>,'\<beta>::bot Set) VAL, ('\<tau>,'\<alpha>::bot)VAL => '\<tau> Boolean] => '\<tau> Boolean" *)
   OclOne_def     :
   "OclOne \<equiv> (lift2' lift_arg0 lift_arg1) (strictify (\<lambda> S P.
                 if \<forall>x \<in> \<HolOclDropLeft>Rep_Set_0 S\<HolOclDropRight>. DEF (P x)
                 then \<HolOclLiftLeft>card {x \<in> \<HolOclDropLeft>Rep_Set_0 S\<HolOclDropRight>. P x = \<HolOclLiftLeft>True\<HolOclLiftRight>} = 1\<HolOclLiftRight> 
                 else \<bottom>))"

ocl_setup_op [OclOne]

text{* 
  The iterator for Sets is not adequately defined in the standard
  (idempotence). This can be overcome by Paulsons theory of the finite
  fold operator of sets: it is always defined, but only for bodies P
  that are \acs{aci}, there is an unfold theorem.

  This definition essentially captures the intention of the
  standard. *}
defs

   OclIterate_def : 
   "OclIterate \<equiv> (lift3' lift_arg0 lift_arg2 lift_arg0) (strictify (\<lambda> S P A.
                  if   (finite \<HolOclDropLeft>Rep_Set_0 S\<HolOclDropRight>) 
                  then (fold_set (\<lambda> x y. P y x) A (\<HolOclDropLeft>Rep_Set_0 S\<HolOclDropRight>))
                  else \<bottom>))"
ocl_setup_op[OclIterate]

subsection{* Properties of iterate on sets *}

lemma iterate_infinite:
  "\<tau> \<MathOclValid> \<MathOclIsUndefined> (\<MathOclSizeLeft> (X::('\<tau>,'\<alpha>::bot)Set) \<MathOclSizeRight>) \<Longrightarrow> 
   \<tau> \<MathOclValid> \<MathOclIsUndefined> (OclIterate X (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau>,'\<beta>::bot)VAL \<Rightarrow> ('\<tau>,'\<beta>)VAL) A)"
  apply(rule_tac X=X in Set_sem_cases, simp_all)
  apply(case_tac "finite c")
  apply(simp_all add: localValidUnDefined2sem OclSize_def OclIterate_def
                      ss_lifting')
  done


text{* 
  Using this theorem all requirements proven about iterate on bags
  will carry over to iterate on \emph{finite} sets quite
  easily. Because they are actually equivalent in this case.  *}
lemma iterate_set2bag:
  "\<tau> \<MathOclValid> \<MathOclIsDefined> (\<MathOclSizeLeft> (X::('\<tau>,'\<alpha>::bot)Set) \<MathOclSizeRight>) \<Longrightarrow> 
   \<tau> \<MathOclValid> (OclIterate X (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau>,'\<beta>::bot)VAL \<Rightarrow> ('\<tau>,'\<beta>)VAL) A) \<MathOclStrongEq>
      (OclIterate ((\<MathOclAsBag> X)::('\<tau>,'\<alpha>::bot)Bag) P A)"
  apply(rule_tac X=X in Set_sem_cases, simp_all)
  apply(simp_all add: localValidDefined2sem)
  apply(case_tac "finite c")
  apply(simp_all add: OclSize_def OclIterate_def OclLocalValid_def
                      OclStrongEq_def OCL_Bag.OclIterate_def 
                      fold_set_fold_multiset_conv
                      OclAsBag_def OclTrue_def ss_lifting')
  done
end
