 (*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * thy_morpher.ML - Generic Theorem-Morpher Lifting Meta-Level Theorems
 * Copyright (C) 2001,2002,2003,2004 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

(* The technique of the "thy_morpher" (constructing a theory morphism 
   between HOL (Main.thy) and OCL (OCL.thy) is described in the paper:
   "Using Theory Morphisms for Implementing Formal Methods Tools",
   A.D.Brucker and B. Wolff, Types 2002, LNCS 2646 *)


signature THY_MORPHER =
  sig
    val SEM_COM     : string list

    val self_test   : unit -> unit

    val state       : {thms: (string * Thm.thm) list, 
                       trans: (string * Thm.thm) list,
                       undefs: (string * Thm.thm) list, 
                       is_defs: (string * Thm.thm) list,
                       liftables: (string * Thm.thm) list,
                       convertables: (string * Thm.cterm) list, 
                       proven_lifts: (string * Thm.thm) list,
                       trace_morpher: bool} ref

    val thy_morpher : Theory.theory -> unit
  end



structure Thy_morpher (* : THY_MORPHER *) =
struct
 

(******************************************************************************)
(*                                                                            *)
(*  Basic Operations                                                          *)
(*                                                                            *)
(******************************************************************************)



fun make_global_valid X = Const("op =",dummyT) $ X $ Const ("OCL_Boolean.TRUE",dummyT)

fun make_local_valid X st = Const("op =",dummyT) $ (X $ st) $
                          (Const ("OCL_Boolean.TRUE",dummyT) $ st)


fun err_dups names = error ("Duplicate entries in thy_morpher: " ^ commas_quote names); 


   fun store_thm2 p (name,thm) = 
       let val clean_dot = BasisLibrary.String.map (fn x => if x = #"." then #"_" else x)
       in bind_thm(clean_dot((p())^name),thm) end;

   fun constants_of (Const(s,t)) = [(s,t)]
      |constants_of (Abs(_,_,t)) = constants_of t
      |constants_of (s $ t)      = (constants_of s)@(constants_of t)
      |constants_of _            = [];

   fun free_of (Free(s,t))       = [(s,t)]
      |free_of (Abs(_,_,t))      = free_of t
      |free_of (s $ t)           = (free_of s)@(free_of t)
      |free_of _                 = []; 


   val SEM_COM        = ["op o", "Let",
                         "Lifting.lift1","Lifting.lift2",
                         "Lifting.up.lift", "Lifting.drop", "Lifting.UU",
                         "OCL_Set.Abs_Set", "OCL_Set.Rep_Set",
                         "OCL_Sequence.Abs_Sequence", "OCL_Sequence.Rep_Sequence",
                         "OCL_OrderedSet.Abs_OrderedSet", "OCL_OrderedSet.Rep_OrderedSet",
                         "OCL_Bag.Abs_Bag", "OCL_Bag.Rep_Bag"] 


   val strict_symbols = ["Lifting.strictify"]

   fun subtract xs xt = filter (fn (x,_) => forall(fn y => not(x = y)) xt)(xs)
                        
   fun find_thym_pair (name,thm) = (* (cterm,cterm,thm) option *)
       case concl_of thm of
          (Const("==",_ ) $ lhs $
            (Const ("Lifting.lift2", _) $
               (Const(strict1_op,_) $ Abs(_,_, 
                Const(strict2_op,_) $ Abs(_,_,rhs))))) 
           =>(if exists (fn x => x=strict1_op) strict_symbols andalso
                 exists (fn x => x=strict2_op) strict_symbols 
              then (case subtract (constants_of rhs) SEM_COM of
                     [t] => Some ((lhs,Const(t)),(name,thm))
                   | _   => None)
              else None)
       | (Const("==",_ ) $ lhs $
            (Const ("Lifting.lift1", _) $ rhs)) 
           => (case subtract (constants_of rhs) SEM_COM of
                 [t] => Some ((lhs,Const(t)),(name,thm))
               | _   => None)
       | _ => None;
 
 
   fun core thy =
       mapfilter (find_thym_pair) (axioms_of thy);


(******************************************************************************)
(*                                                                            *)
(*  Generic Theorem Generator                                                 *)
(*                                                                            *)
(******************************************************************************)

(* basic lemmas *)

(* TO DO: Es sollten eher die Regeln der Bauart:
   |= is_def a
  -----------------
   |= is_def (f a) 

   oder 

   |= is_def a     |= is_def b
  -----------------------------
   |= is_def (f a b) 

   ins rewriting aufgenommen werden, oder ?

   Allgemeine Anmerkung: Es fehlt sowas wie iffD1 oder iffD2 
   auf OCL-Ebene, nicht?

   A = B   |= A
   ------------
       |=B
  
   ist allerdings trivial.
  
 *)



fun mRS S (name,thm) = mapfilter (fn a => Some(name,thm RS a) 
                                          handle THM _ => None)
                                 (S);

fun derive_sieve thms =
(* generates from a list of definitions several classes of rules
   and provides setup for automatic procedures ... *)
    let val trans  = flat (map (mRS [cp_lift1_fw,
                                     cp_lift2_fw]) thms)
        val undefs = flat (map (mRS [(*lift2_undef1_fw,*)
                                     lift2_undef1a_fw,lift2_undef2a_fw,
                                     lift1b_undef_fw (*, lift1_undef_fw *)]) thms)
        val is_defs= flat (map (mRS [lift1_strict_is_isdef_fw,
                                     lift1_strictify_is_isdef_fw,
                                     lift2_strict_is_isdef_fw,
                                     lift2_strictify_is_isdef_fw]) thms);
        val toggle             = ref(~1);
        fun toggle_name name() = (toggle:= ((!toggle+1) mod 2);
                                  name^BasisLibrary.Int.toString(!toggle))
    in  
        AddSIs (map (fn (_,t)=>t) trans);
        map (store_thm2 (K"OCL_cp")) trans;

        Addsimps (map (fn (_,t)=>t) undefs);
        map (store_thm2 (toggle_name"OCL_undef")) undefs;

        Addsimps (map (fn (_,t)=>t) is_defs); (* really ??? *)
        map (store_thm2 (toggle_name"OCL_is_def")) is_defs;
        
        (trans,undefs,is_defs) 
    end;

    (* superfluous now - has been integrated into OCL_Logic.thy *)



(******************************************************************************)
(*                                                                            *)
(*  The Lifter (LIFT_E in documentation                                       *)
(*                                                                            *)
(******************************************************************************)



fun convert thy subst (name,thm) =
    let fun lookup(s,t) = mapfilter (fn ((c1,Const(_,t')),(name,def)) => 
                                        if Type.raw_unify(t,t') 
                                        then Some c1
                                        else None)
                                    (Symtab.lookup_multi(subst,s))

        fun adoptTrueprop (A as(Const("op =",_) $ _ $ _)) = A 
           |adoptTrueprop X = (case fastype_of X of 
                                 Type("fun",[_,Type("Lifting.up",[Type("bool",[])])]) 
                                 => make_global_valid X
                               |  _ => X) handle _ => X

        fun conv (Free(s,_))    = [Free(s,dummyT)]
           |conv (Var((s,_),_)) = [Free(s,dummyT)]
           |conv (Const(s,t))   = (case lookup(s,t) of
                                    [] => [Const(s,dummyT)]
                                  | S  => S)
           |conv (Abs(s,_,t))   = map(fn x => Abs(s,dummyT,x)) (conv t)
           |conv (Const ("==>",t) $ A $ B) 
                                = flat(map(fn x => map(fn y => Const ("==>",t) $ x $ y)
                                          (conv B))(conv A)) (* optimization *)
           |conv ((A as Const ("Trueprop",t)) $ B) 
                                = map(fn x=> A $ adoptTrueprop x) (conv(B))
           |conv (s $ t)        = flat(map(fn x => map(fn y => x $ y)(conv t))(conv s))
           |conv x              = [x](* Bound variable *)

	val groundTVars  = map_term_types (Type.unvarifyT);

        val cc = map groundTVars (conv(term_of(cprop_of thm)));
        (* fun fst (x,y)        = x *)
        fun type_inference ct   = Some(fst(Sign.infer_types (sign_of thy) 
				          (K None) (K None) [] false ([ct],propT)))
                                  handle _ => None
        
        val errorfn_save = (!error_fn) 
        val _ = (error_fn:= (fn _ => ()))  (* type_inference: shut up *)                  
        val ccc = mapfilter type_inference cc
        val _ = (error_fn := errorfn_save) (* type_inference: speak up again *)
     in map (fn x => (name,cterm_of (sign_of thy) x)) ccc
     end;



(* Some example proof code for a theorem-lifting 



goalw thy  [plus_def,lift2_def]
"((X::'a Integer)+ Y)=Y+X";
br ext 1;
by(case_tac "DEF(X St)" 1);
by(case_tac "DEF(Y St)" 1);
by(ALLGOALS(asm_full_simp_tac (HOL_ss addsimps [not_DEF_X])));
by(ALLGOALS(asm_full_simp_tac (HOL_ss addsimps [DEF_X_up])));
auto();
(* br zadd_commute 1; *)
qed"plus_commute";

(* commute is very special and does not require congruence reasoning 
   in he datatype adaption layer. Therefor, in general, associativity
   theorems are much better test cases ...
 *)
 
goalw thy  [plus_def,lift2_def]
"((X::'a Integer)+ Y) + Z = X+(Y+Z)";
br ext 1;
by(case_tac "DEF(X St)" 1);
by(case_tac "DEF(Y St)" 1);
by(case_tac "DEF(Z St)" 1);
by(ALLGOALS(asm_full_simp_tac (HOL_ss addsimps [not_DEF_X])));
by(ALLGOALS(asm_full_simp_tac (HOL_ss addsimps [DEF_X_up])));
auto();
(* br zadd_assoc 1; - auto finds the trick *)
qed"plus_commute";




goalw OCL_Set.thy []  
"DEF(X) = (? c. UU ~: c & Rep_Set X = ( |. c .| ))";
br iffI 1;
by(forward_tac [exists_lift_Set] 1);
auto();
bd sym 1;
by(ALLGOALS(asm_full_simp_tac (simpset() addsimps [Rep_Set])));
br swap 1;
br not_DEF_Rep_Set 2;
ba 2;
by(ALLGOALS(asm_full_simp_tac (simpset())));
qed"DEF_X_Set_0";



goalw OCL_Set.thy []  
"DEF(X) = (EX c. UU ~: c & X = Abs_Set( |. c .| ))";
by(auto_tac(claset(),simpset() addsimps [DEF_Abs_Set]));
by(asm_full_simp_tac (simpset() addsimps [DEF_X_Set_0]) 1);
be exE 1; br exI 1; auto();
br (inj_Rep_Set RS injD) 1;
by(auto_tac (claset(),simpset() addsimps [Abs_Set_inverse]));
qed"DEF_X_Set";


goalw OCL_Set.thy []  
"!!x. [| (lift x ) : Set; (lift y ) : Set |] ==> lift(x Un y) : Set";
by(asm_full_simp_tac (HOL_ss addsimps [smashed_set_charn RS sym]) 1);
auto();
qed"smash_inv_Un";

goalw OCL_Set.thy []  
"!!x. [| (lift x ) : Set; (lift y ) : Set |] ==> lift(x Int y) : Set";
by(asm_full_simp_tac (HOL_ss addsimps [smashed_set_charn RS sym]) 1);
auto();
qed"smash_inv_Int";

goalw OCL_Set.thy []  
"!!x. [| (lift x ) : Set |] ==> lift((- x) - {UU}) : Set";
by(asm_full_simp_tac (HOL_ss addsimps [smashed_set_charn RS sym]) 1);
auto();
qed"smash_inv_minus";



val [] = goalw thy [thm"OCL_Set.union_def",lift2_def]
    "union X Y = union Y X";
    br ext 1;
    by(case_tac "DEF(X St)" 1);
    by(case_tac "DEF(Y St)" 1);

    by(ALLGOALS(asm_full_simp_tac (simpset() addsimps [thm"not_DEF_X"])));
    by(ALLGOALS(asm_full_simp_tac (HOL_ss addsimps [thm"DEF_X_Set"])));

    by(auto_tac (claset() addSIs congr_intros,
                 simpset() addsimps ([thm"Abs_Set_inverse"]) ));

    by(ALLGOALS(asm_full_simp_tac (HOL_ss addsimps congr_charns)));
    auto();

    br Un_commute 1;
    qed "OCL_union_commute";


val [] = goalw (theory"OCL") [thm"OCL_Set.union_def",lift2_def]
    "union X (union Y Z) = union (union X Y) Z";
    br ext 1;
    by(case_tac "DEF(X St)" 1);
    by(case_tac "DEF(Y St)" 1);
    by(case_tac "DEF(Z St)" 1);

    by(ALLGOALS(asm_full_simp_tac (simpset() addsimps [thm"not_DEF_X"])));
    by(ALLGOALS(asm_full_simp_tac (HOL_ss addsimps [thm"DEF_X_Set"])));

    by(auto_tac (claset()  addSIs congr_intros,
                 simpset() addsimps ([thm"Abs_Set_inverse",
                                      thm"DEF_Abs_Set"] @
                                      congr_charns) ));

    qed "OCL_union_assoc";


*)


val DEF_exhausts  = case config.collectiontype of
			config.unsmashed => [thm "DEF_X_up"]
		      | config.smashed   => [thm"DEF_X_up",
					     thm"DEF_X_Set",
					     thm"DEF_X_Sequence"];
val congr_inv_aux = case config.collectiontype of
			config.unsmashed => []
		      | config.smashed   => [thm"Abs_Set_inverse",
					     thm"DEF_Abs_Set",
					     thm"Abs_Sequence_inverse",
					     thm"DEF_Abs_Sequence"]

val congr_intros  = case config.collectiontype of
			 config.unsmashed => []
		       | config.smashed   => 
			 [((thm"inj_on_Abs_Set") RS inj_on_iff) RS iffD2,
                          ((thm"inj_on_Abs_Sequence") RS inj_on_iff) RS iffD2];
                       (* do not use as rewrites !!! *)
val congr_charns  = case config.collectiontype of
			 config.unsmashed => []
		       | config.smashed   => 
			 [(thm"smashed_set_charn") RS sym,
                          (thm"smashed_sequence_charn") RS sym];


fun eq_lifter_prover thmtab thms (name,goal) = 
    let val thyname  = hd(BasisLibrary.String.tokens(fn x => x = #".")(name))
        val core_thm = (* get_thm(theory thyname)(name) makes problems
                          since internal name convention does not always 
                          match access policies ... *)
                       the(Symtab.lookup(thmtab,name)) 
                       handle _ => refl (* just to prevent a crash here*)
        val v        = distinct(free_of(term_of goal))
        val fnm      = map (fn (X,_)=> "DEF("^X^" St)") v
        val case_tacs= map (fn X=> case_tac X 1) fnm
    in  Some(name,prove_goalw_cterm thms goal
                  (fn prems => [ cut_facts_tac prems 1,
                                 rtac ext 1]
                               @ case_tacs @
                               [ ALLGOALS(asm_full_simp_tac 
                                             (simpset() addsimps [thm"not_DEF_X"])),
                                 ALLGOALS(asm_full_simp_tac 
                                             (HOL_ss addsimps DEF_exhausts)),
(* SIMPLE
                                 auto_tac (claset(),simpset()),
 *)
(* WITH smasher ... *)
                                 auto_tac (claset()  addSIs congr_intros,
                                           simpset() addsimps (congr_inv_aux @
                                                               congr_charns)),
(* ... *)
                                 TRY(rtac core_thm 1)]))
        handle _ => None
    end;


fun lift root_thy core_E thms =
    let fun ran core_E   = map (fn ((_,y),_)=> y) core_E;
        val thmtab       = Symtab.make(filter (fn (x,y) => not(x ="")) thms);
        val the_logical_connectives = ["Trueprop","op =", "Lifting.DEF",
	                               "All", "==>"]
	val core_E_M     = filter (fn ((_,Const(s,_)),_) => 
	                               forall(fn x => not(x = s))
				             (the_logical_connectives))
			          (core_E)
        val the_ran_core_E_M        = map (fn (Const(s,_))=> s) (ran core_E_M)
        fun is_liftable (name,thm)  = 
                           let val cc  = constants_of(term_of(cprop_of thm)) 
                               val cc1 = subtract cc the_logical_connectives
                               val cc2 = subtract cc1 the_ran_core_E_M
                           in  not(null(cc1)) andalso null(cc2) end
        val liftables    = filter is_liftable thms;
    
        val subst        = Symtab.make_multi(map (fn((x,y as Const(s,t)),a)=>(s,((x,y),a))) core_E_M)
                           (* subst is the inversion of the core mapping.
                              Since the core of the morphism is *NOT*
                              injective in general, the inversion may be one-to-many ... *)
   
        val convertables = flat(map (convert root_thy subst) liftables)
        val defs         = lift0_def::lift1_def::lift2_def::
                              (map (fn (_,(_,t)) => t) core_E)

        val proven_lifts = mapfilter (eq_lifter_prover thmtab defs) convertables
        val _            = map (store_thm2 (K"OCL_")) proven_lifts

    in  (liftables,convertables,proven_lifts) end;


(******************************************************************************)
(*                                                                            *)
(*  Control                                                                   *)
(*                                                                            *)
(******************************************************************************)


val state = ref   {trace_morpher = true,
                   trans         = [] : (string * Thm.thm) list,
                   undefs        = [] : (string * Thm.thm) list,
                   is_defs       = [] : (string * Thm.thm) list,
                   thms          = [] : (string * Thm.thm) list,
                   liftables     = [] : (string * Thm.thm) list,
                   convertables  = [] : (string * Thm.cterm) list,
                   proven_lifts  = [] : (string * Thm.thm) list} 


fun thy_morpher root_thy =
    let val (E,holthys)
                   = BasisLibrary.List.partition(fn x => subthy(theory "Lifting",x))
                            (root_thy::(ancestors_of root_thy));
        val _      = (writeln("\n Theory-Morpher from HOL to OCL: ");
                      writeln(  " =============================== \n"));

        val core_E = flat (map core E);
        val _      = if not(#trace_morpher(!state)) then () 
                     else let val H = Syntax.simple_string_of_typ
                              val L = map (fn ((Const(s,t),Const(s',t')),_) => 
                                               (s^"::"^(H t),s'^"::"^(H t'))) core_E
                              val _ = (writeln("\n  The core of the morphism: ");
                                       writeln("  >>>>>>>>>>>>>>>>>>>>>>>>"));
                              val _ = app (fn (s,t) => writeln("\n      "^s^"\n |==> "^t)) L
                              val _ = writeln("\n  <<<<<<<<<<<<<<<<<<<<<<<< \n");
                          in  () end

        val (trans,undefs,is_defs)    
                   = derive_sieve (map (fn(_,x)=>x) core_E);
        val _      = (writeln("\n derive from generics:");
                      writeln(" ===================== \n");
                      writeln(" cp - rules: "^BasisLibrary.Int.toString(length trans));
                      writeln(" undef    - rules: "^BasisLibrary.Int.toString(length undefs));
                      writeln(" is_def   - rules: "^BasisLibrary.Int.toString(length is_defs)));

        val thms   = flat (map thms_of holthys);
        val (liftables,convertables,proven_lifts)      
                   = lift root_thy core_E thms;
        val _      = (writeln("\n theorem lifting:");
                      writeln(" ================== \n");
                      writeln(" HOL Theorems : "^BasisLibrary.Int.toString(length thms));
                      writeln(" Liftables    : "^BasisLibrary.Int.toString(length liftables));
                      writeln(" Convertables : "^BasisLibrary.Int.toString(length convertables));
                      writeln(" Proven Lifts : "^BasisLibrary.Int.toString(length proven_lifts)));
    in  state :=  {trace_morpher = #trace_morpher(!state),
                   trans         = trans,
                   undefs        = undefs,
                   is_defs       = is_defs,
                   thms          = thms,
                   liftables     = liftables,
                   convertables  = convertables,
                   proven_lifts  = proven_lifts}
    end;

    fun self_test () = 
        let val {trace_morpher, trans, undefs, is_defs, thms, 
             liftables, convertables, proven_lifts} = !state
        in  if length(trans) < 24         then error"Thy_morpher Check: trans"  else
            if length(undefs) < 46        then error"Thy_morpher Check: undefs"  else
            if length(is_defs)< 32        then error"Thy_morpher Check: is_def" else
            if length(liftables) < 550   then error"Thy_morpher Check: liftables"  else
            if length(convertables) < 736 then error"Thy_morpher Check: convertables"  else
            if length(proven_lifts) < 204 then error"Thy_morpher Check: proven_lifts"  else ()
        end;


end; (* struct *)
 
